#include "httpd.h"
#include "http_config.h"
#include "http_protocol.h"
#include "http_log.h"
#include "ap_config.h"

/* PostgreSQL access functions */
//#include "libpq-fe.h"

//#include "co_utils.h"
//#include "co_log.h"
#include "../include/co_defines.h"
#include "co_lib.h"
#include "co_defines_calls.h"

#include "co_mod.h"
#include "../comodules/generic_acl/co_generic_acl.h"

#include "co_broker.h"
#include "mod_coffice.h"
#include "../comodules/js_code/co_js_code.h"
#include "../comodules/profiler/co_profiler.h"

const char **
cou_getContainerInfo( co_context *ctx, const char* area_name, const char* class_name)
{
    PGresult *res;
    PGconn *conn;
    const char *sql_buf;
    const char **info; /* const char *info[2] */
    pool *p;

    conn=ctx->dbconn;
    p = ctx->request->pool;
    info = (const char **)ap_palloc(p, 2*sizeof(char*));

    cou_logError(ctx, CO_LOG_DEBUG,
	    "looking for container info for class %s::%s",
            area_name, class_name
	);

    sql_buf = ap_psprintf(p,
"SELECT container_class, container_column_name FROM co_%s_meta_class WHERE class_name = '%s'",
                area_name, class_name
	);

    cou_logError(ctx, CO_LOG_DEBUG,
	    "query='%s'", 
            sql_buf
	);

    res = PQexec(conn, sql_buf);

    if (PQresultStatus(res) != PGRES_TUPLES_OK)
    {
     cou_logError(ctx, CO_LOG_DEBUG,
       "TUPLES not ok");
            PQclear(res);
	    /* *err = -1; */
            return NULL;
    }

     cou_logError(ctx, CO_LOG_DEBUG,
       "got result");

    if(PQntuples(res) > 0)
    {
        info[0] = ap_pstrdup(p, PQgetvalue(res,0,0)); /* container_class */
        info[1] = ap_pstrdup(p, PQgetvalue(res,0,1)); /* container_column_name */
    }
    else
    {
        info[0] = NULL;
        info[1] = NULL;
    }
    return info;
}

void
cou_broker( const char* area_name, const char* class_name, const char* method_name, 
	    co_context *ctx, void *in_params )
{
	int (*action)(co_context *, ...);
//    const char *called_class_name = "";
	method_info m_i;
	int acl_check = 0;
	int err = 0;
	int id_permission = 0;
	int i;
	int id_profiler_entry = 0;
	bool sec_critical = ctx->critical; /* the same value throughout the call */
	pool *p = ctx->request->pool;
	req_params * in_params_req = (req_params *) in_params;

	// fill in m_i
	m_i.area_name = area_name;
	m_i.source_class_name = class_name;
	m_i.method_name = method_name;
	m_i.target_class_name = NULL;
	m_i.profiled = 0;
	m_i.acl = 0;

	cou_logError( ctx, CO_LOG_DEBUG, 
		"cou_broker call in area '%s': class='%s' method='%s'\n", area_name, class_name, method_name
	);

	// default values for acl checking
	m_i.check_par_area = area_name;
	m_i.check_par_class = class_name;
	m_i.check_par_id = 0;

	// now we try to read these values from params
	if(in_params_req == NULL)
	{
		void * ret =  cou_getParam(ctx,"par_area",CO_STRING, &err);
		if(err==0 && ret !=NULL)
		{
			m_i.check_par_area = (char *) ret;
		}

		ret =  cou_getParam(ctx,"par_class",CO_STRING, &err);
		if(err==0 && ret !=NULL)
		{
			m_i.check_par_class = (char *) ret;
		}

		ret =  cou_getParam(ctx,"par_id",CO_INTEGER, &err);
		if(err==0 && ret !=NULL)
		{
			m_i.check_par_id = *(int *)ret;
		}

	}
	else
	{
		if(in_params_req->magic_par_area == __PARAMS_MAGIC_WITH_PAR_AREA)
		{
			if(in_params_req->control->par_area==1)
			{
				m_i.check_par_area = in_params_req->values->par_area;
			}
		}
		else if(in_params_req->magic_par_area != __PARAMS_MAGIC_NO_PAR_AREA)
		{
			cou_logError( ctx, CO_LOG_DEBUG, 
				"cou_broker in_params with wrong magic number! call in area '%s': class='%s' method='%s'\n", area_name, class_name, method_name
						);
		}
		if(in_params_req->magic_par_class == __PARAMS_MAGIC_WITH_PAR_CLASS)
		{
			if(in_params_req->control->par_class==1)
			{
				m_i.check_par_class = in_params_req->values->par_class;
			}
		}
		else if(in_params_req->magic_par_class != __PARAMS_MAGIC_NO_PAR_CLASS)
		{
			cou_logError( ctx, CO_LOG_DEBUG, 
				"cou_broker in_params with wrong magic number! call in area '%s': class='%s' method='%s'\n", area_name, class_name, method_name
						);
		}

		if(in_params_req->magic_par_id == __PARAMS_MAGIC_WITH_PAR_ID)
		{
			if(in_params_req->control->par_id==1)
			{
				m_i.check_par_id = in_params_req->values->par_id;
			}
		}
		else if(in_params_req->magic_par_id != __PARAMS_MAGIC_NO_PAR_ID)
		{
			cou_logError( ctx, CO_LOG_DEBUG, 
				"cou_broker in_params with wrong magic number! call in area '%s': class='%s' method='%s'\n", area_name, class_name, method_name
						);
		}
	}
 

	for (i=0; i < SERVER_EVENT_COUNT; i++)
	{
		m_i.id_script[i] = 0;
		m_i.id_method[i] = 0;
	}

    /* find name of the class ( prehaps ancestor class ) which implements
	method */

	cou_logError( ctx, CO_LOG_DEBUG, 
		"cou_broker resolves call in area '%s': class='%s' method='%s'\n", area_name, class_name, method_name
	);
    
	err = 0;
	cox_gatherMethodInfo( ctx, &m_i, &err);

	if(err < 0)
	{
		cou_sendMessageX(ctx, 
			ap_psprintf(ctx->request->pool, "cou_broker: %s::%s::%s method can not be found in vtable!",
					area_name, class_name, method_name), CO_MSG_FORCE);
		ctx->error = CO_METHOD_NOT_FOUND;
		return ;
	}

	if(ctx->check_security && m_i.acl != 1)
	{
		// SOMETHING ILLEGAL IS GOIN ON!
		cou_sendMessageX(ctx, ap_psprintf(ctx->request->pool,
                              "cou_broker: Security warning - class: '%s' method: '%s' type: %d id_permission: %d", 
                               class_name, method_name, m_i.sec_type, ctx->id_permission),
                         CO_MSG_FORCE);
		ctx->error = CO_METHOD_FORBIDDEN;
		if (sec_critical) cou_error(ctx, "Operation forbidden");
		return;
	}


	cou_logError(ctx, CO_LOG_DEBUG,
			"cou_broker resolving result for call in area '%s': class='%s' method='%s'\n", area_name, m_i.target_class_name, method_name
	);


/*
	if((ctx->check_security) && (strcmp(method_name, "check_acl")!=0) && (strcmp(method_name, "check_acl_class")!=0) )
	{
		cou_logError(ctx, CO_LOG_DEBUG,
		"BEFORE cou_broker checkes acl in area '%s': class='%s' method='%s' id_permission:%d sec_type:%d \n", 
                                        area_name, class_name, method_name,ctx->id_permission, m_i.sec_type
	    );
		ctx->error = CO_NO_ERRORS;
		if(m_i.sec_type == __METHOD_SEC_TYPE_OBJECT)
		{
			cou_broker(area_name,class_name,"check_acl",ctx,in_params);
		}
		else if(m_i.sec_type == __METHOD_SEC_TYPE_CLASS)
		{
			cou_broker(area_name,class_name,"check_acl_class",ctx,in_params);
		}
		else
		{
			cou_sendMessageX(ctx, 
                	ap_psprintf(ctx->request->pool,  
                       "ERRORS wrong sec_type in area '%s': class='%s' method='%s' sec_type:%d \n", area_name, class_name, method_name,m_i.sec_type),
   	               CO_MSG_FORCE
			     );
  		    ctx->error = CO_METHOD_NOT_FOUND; ///TODO
	    	if (sec_critical) cou_error(ctx, "Wrong security type");
			return ;
		}  
        
		if (ctx->error != CO_NO_ERRORS)
		{
			acl_check = 0;
			cou_logError(ctx, CO_LOG_WARNING,
			"ERRORS in check_acl in area '%s': class='%s' method='%s' acl:%d \n", area_name, class_name, method_name,acl_check);
			if (sec_critical) cou_error(ctx, "ACL check failed");
		}
		else
		{
			acl_check = *(int *)cou_getReturnValue(ctx,CO_INTEGER, &err);
			cou_logError(ctx, CO_LOG_DEBUG,
				"AFTER cou_broker checkes acl in area '%s': class='%s' method='%s' acl:%d \n", area_name, class_name, method_name,acl_check
				);
		}

		// HERE ERROR HANDLING!
		if(acl_check != 1)
		{
			// SOMETHING ILLEGAL IS GOIN ON!
			cou_sendMessageX(ctx, ap_psprintf(ctx->request->pool, 
                                    "cou_broker: Trying to do something illegal, he? class: '%s' method: '%s' type: %d id_permission: %d", 
                                    class_name, method_name, m_i.sec_type, ctx->id_permission),
                          CO_MSG_FORCE);
			ctx->error = CO_METHOD_FORBIDDEN;
			if (sec_critical) cou_error(ctx, "Operation forbidden");
			return;
		}
    }
*/
    ctx->critical = true; /* always exit with critical security set to true */

   // if profiled => start profiling
   if (m_i.profiled != 0)
   {
     action = co_callByName( ctx, "profiler", "setStart");
     {
       __PARAM_STRUCT_CREATE_setStart(out_param2)
       __PARAM_STRUCT_SET(m_i.area_name, out_param2, par_area)
       __PARAM_STRUCT_SET(m_i.target_class_name, out_param2, par_class)
       __PARAM_STRUCT_SET(1, out_param2, par_id)
       __PARAM_STRUCT_SET(method_name, out_param2, par_method)
       action( ctx, out_param2);
     }
     id_profiler_entry = *((int *) cou_getReturnValue(ctx, CO_INTEGER, &err));
   }
   
   // if server onbefore event defined => execute such event
   if (m_i.id_script[SERVER_EVENT_ONBEFORE]) {
	if (in_params==NULL) {
	    void*(*param_builder)(co_context *);
	    param_builder = (void*(*)(co_context *))co_callByNameExt(ctx, m_i.target_class_name, "pbs", m_i.method_name);
	    if (param_builder) {
		in_params=param_builder(ctx);
	    }
	}
	cox_executeEventScript( ctx, &m_i, SERVER_EVENT_ONBEFORE, in_params, &err);
   }
   else if (m_i.id_method[SERVER_EVENT_ONBEFORE]) cox_executeEventMethod( ctx, &m_i, SERVER_EVENT_ONBEFORE, &err);

   // REQUIRES ERROR HANDLING!

   action = co_callByName( ctx, m_i.target_class_name, method_name );
   if( action ) {
	if(action( ctx, in_params ) != 0) {
    	    cou_sendMessage(ctx, 
	        ap_psprintf(ctx->request->pool, "cou_broker: method %s::%s::%s (implemented in %s) returned with error", 
	           area_name, class_name, method_name, m_i.target_class_name)
		);
	    ctx->error = CO_METHOD_FAILED;
	}
   }
   else {
        cou_sendMessage(ctx, 
	            ap_psprintf(ctx->request->pool, "cou_broker: %s::%s::%s method can not be correctly evaluated, probably module '%s' is not loaded!", 
	                   area_name, class_name, method_name, m_i.target_class_name)
		     );
   }
   if (ctx->error != CO_NO_ERRORS)
    {
	if (sec_critical) cou_error(ctx, "Method call failed");
    }

   if (m_i.id_script[SERVER_EVENT_ONAFTER]) {
	if (in_params==NULL) {
	    void*(*param_builder)(co_context *);
	    param_builder = (void*(*)(co_context *))co_callByNameExt(ctx, m_i.target_class_name, "pbs", m_i.method_name);
	    if (param_builder) {
		in_params=param_builder(ctx);
	    }
	}
	cox_executeEventScript( ctx, &m_i, SERVER_EVENT_ONAFTER, in_params, &err);
   }
   else if (m_i.id_method[SERVER_EVENT_ONAFTER]) cox_executeEventMethod( ctx, &m_i, SERVER_EVENT_ONAFTER, &err);

   // if profiled => finish profiling
   if (m_i.profiled != 0)
   {
     action = co_callByName( ctx, "profiler", "setEnd");
     {
       __PARAM_STRUCT_CREATE_setEnd(out_param2)
       __PARAM_STRUCT_SET(m_i.area_name, out_param2, par_area)
       __PARAM_STRUCT_SET(class_name, out_param2, par_class)
       __PARAM_STRUCT_SET(1, out_param2, par_id)
       __PARAM_STRUCT_SET(id_profiler_entry, out_param2, id_entry)
       action( ctx, out_param2);
     }
   }

   if (ctx->error != CO_NO_ERRORS)
    {
	if (sec_critical) cou_error(ctx, "Event call failed");
    }


   cou_logError( ctx, CO_LOG_DEBUG, "cou_broker: exiting" );
}

//#define CO_SQL_BUF_LEN 255
//static char sql_buf[CO_SQL_BUF_LEN];

void cox_executeEventScript( co_context *ctx, method_info *m_i, co_server_event s_e, void *in_params, int *err)
{
   pool * p = ctx->request->pool;
   cou_logError(ctx, CO_LOG_DEBUG,"executeEventScript called");
   {
      __PARAM_STRUCT_CREATE_evaluate(out_param2)
      __PARAM_STRUCT_SET(m_i->area_name,      out_param2, par_area)
      __PARAM_STRUCT_SET("js_code",           out_param2, par_class)
      __PARAM_STRUCT_SET(m_i->id_script[s_e], out_param2, par_id)
      __PARAM_STRUCT_SET("",                  out_param2, prolog)
      __PARAM_STRUCT_SET(0,                   out_param2, asp)
      __PARAM_STRUCT_SET(in_params,           out_param2, in_params)
      __PARAM_STRUCT_SET(m_i->method_name,    out_param2, method_name)
      __PARAM_STRUCT_SET(m_i->target_class_name,    out_param2, target_name)
      cou_broker(m_i->area_name, "js_code", "evaluate", ctx, out_param2);
   }

    return;
}

void cox_executeEventMethod( co_context *ctx, method_info *m_i, co_server_event s_e, int *err)
{
    PGresult *res;
    PGconn *conn;
    const char* sql_buf;
    const char* class_name;
    const char* method_name;

    conn = ctx->dbconn;

    cou_logError( ctx, CO_LOG_DEBUG, "cou_executeEventMethod: entering" );


    sql_buf = ap_psprintf(ctx->request->pool,
        "select * from co_%s_meta_method where id = %d;", m_i->area_name, m_i->id_method[s_e]);

    res = PQexec(conn, sql_buf);

    class_name = ap_pstrdup(ctx->request->pool, PQgetvalue(res, 0, PQfnumber(res,"class_name")));
    method_name = ap_pstrdup(ctx->request->pool, PQgetvalue(res, 0, PQfnumber(res,"method_name")));

    cou_logError(ctx, CO_LOG_DEBUG,"clearing");
    PQclear(res);

    cou_logError(ctx, CO_LOG_DEBUG,"cou_executeEventMethod: after clearing");

// would be fine to create new context

    {
      co_context *new_ctx;
      
      new_ctx = ap_palloc( ctx->request->pool, sizeof(co_context));
      *new_ctx = *ctx;

/*
      new_ctx->request = ctx->request;
      new_ctx->dbconn = ctx->dbconn;
      new_ctx->id_user = ctx->id_user;
      new_ctx->id_role = ctx->id_role;
      new_ctx->id_session = ctx->id_session;
      new_ctx->jscnx = NULL;
      new_ctx->putJS = false;
      new_ctx->result_type = CO_NO_TYPE;
      new_ctx->debug_level = 0;
*/
      
      cou_broker( m_i->area_name, class_name, method_name, new_ctx, NULL);
      
    }


/*
    return;
*/
}

/* cox_getTargetClassName cannot be called recurrently */
const char *cox_getTargetClassName( co_context *ctx, const char* area_name, const char* class_name, const char* method_name, int *err)
{
    PGresult *res;
    PGconn *conn;
    char *target_name;
    const char* sql_buf;
    
//    *err = 1;

    conn=ctx->dbconn;
    
     cou_logError(ctx, CO_LOG_DEBUG,
       "looking for target class area='%s' class='%s' method='%s'", 
               area_name, class_name, method_name
	       );

   sql_buf = ap_psprintf(ctx->request->pool,
        "select * from co_%s_meta_vtable where class_name = '%s' and method_name = '%s';",
                area_name, class_name, method_name );

     cou_logError(ctx, CO_LOG_DEBUG,
       "query='%s'", 
               sql_buf
	       );

    res = PQexec(conn, sql_buf);

    if (PQresultStatus(res) != PGRES_TUPLES_OK)
    {
     cou_logError(ctx, CO_LOG_DEBUG,
       "TUPLES not ok");
            PQclear(res);
	    *err = -1;
            return NULL;
    }

     cou_logError(ctx, CO_LOG_DEBUG,
       "got result");

    //target_name = ap_pstrdup(ctx->request->pool, PQgetvalue(res, 0, 0));
    if(PQntuples(res) > 0)
    {
       target_name = ap_pstrdup(ctx->request->pool, PQgetvalue(res, 0, PQfnumber(res,"target_class_name")));
       if(strcmp(method_name, "check_acl")!=0)
       { 
          ctx->id_permission = atoi(PQgetvalue(res, 0, PQfnumber(res,"id_permission")));
       }
       cou_logError(ctx, CO_LOG_DEBUG,
                     "got target='%s' id_permission=%d", target_name, ctx->id_permission);
    }
    else
    {
      *err = -2;
      return NULL;
    }        
//    ctx->id_permission = atoi(PQgetvalue(res, 0, PQfnumber(res,"id_permission")));
     cou_logError(ctx, CO_LOG_DEBUG,
       "clearing");
    PQclear(res);
    return target_name;
}

void cox_gatherMethodInfo(co_context *ctx, method_info *m_i, int *err)
{
    PGresult *res;
    PGconn *conn;
    const char *sql_buf;

    conn=ctx->dbconn;

    cou_logError(ctx, CO_LOG_DEBUG,
       "looking for target class area='%s' class='%s' method='%s'",
               m_i->area_name, m_i->source_class_name, m_i->method_name
	       );

    sql_buf = ap_psprintf(ctx->request->pool,
        "select * , CASE WHEN sec_type=1 OR force_sec_type=1 THEN co_%s_%s_check_acl_class(id_permission, %d) " 
        " ELSE co_%s_%s_check_acl(%d, id_permission, %d) END AS acl "
        " FROM co_%s_meta_vtable where class_name = '%s' and method_name = '%s'",
          m_i->check_par_area,  m_i->check_par_class, ctx->id_role, 
          m_i->check_par_area,  m_i->check_par_class, m_i->check_par_id, ctx->id_role, 
          m_i->area_name, m_i->source_class_name, m_i->method_name );

     cou_logError(ctx, CO_LOG_DEBUG,
       "query='%s'", 
               sql_buf
	       );

    res = PQexec(conn, sql_buf);

    if (PQresultStatus(res) != PGRES_TUPLES_OK)
    {
     cou_logError(ctx, CO_LOG_DEBUG,
       "TUPLES not ok");
            PQclear(res);
	    *err = -1;
            return;
    }

     cou_logError(ctx, CO_LOG_DEBUG,
       "got result");

    if(PQntuples(res) > 0)
    {
       m_i->target_class_name = ap_pstrdup(ctx->request->pool, PQgetvalue(res, 0, PQfnumber(res,"target_class_name")));
       if(strcmp(m_i->method_name, "check_acl")!=0 && strcmp(m_i->method_name, "check_acl_class")!=0)
       {
//TODO: we can optimise this code by using index not names of fields
          ctx->id_permission = atoi(PQgetvalue(res, 0, PQfnumber(res,"id_permission")));
          m_i->id_script[SERVER_EVENT_ONBEFORE] = atoi(PQgetvalue(res, 0, PQfnumber(res,"id_onbefore_script")));
          m_i->id_method[SERVER_EVENT_ONBEFORE] = atoi(PQgetvalue(res, 0, PQfnumber(res,"id_onbefore_method")));
          m_i->id_script[SERVER_EVENT_ONAFTER] = atoi(PQgetvalue(res, 0, PQfnumber(res,"id_onafter_script")));
	  m_i->id_method[SERVER_EVENT_ONAFTER] = atoi(PQgetvalue(res, 0, PQfnumber(res,"id_onafter_method")));
          m_i->profiled = atoi(PQgetvalue(res, 0, PQfnumber(res,"profiled")));
          m_i->sec_type = atoi(PQgetvalue(res, 0, PQfnumber(res,"sec_type"))); 
          if(!m_i->sec_type)
          {
              m_i->sec_type = atoi(PQgetvalue(res, 0, PQfnumber(res,"force_sec_type"))); 
          }
		m_i->acl = atoi(PQgetvalue(res, 0, PQfnumber(res,"acl"))); 


       }

       cou_logError(ctx, CO_LOG_DEBUG,
                     "got target='%s' id_permission=%d acl=%d", m_i->target_class_name, ctx->id_permission, m_i->acl);
    }
    else
    {
      *err = -2;
      return;
    }

    cou_logError(ctx, CO_LOG_DEBUG,"clearing");
    PQclear(res);
    return;

}