#include "httpd.h"
#include "http_protocol.h"

/* PostgreSQL access functions */
#include "libpq-fe.h"

#include "jsapi.h"
#include "jsstr.h" /* structure of JSString */

#include "config.h"

#define INSIDE_CO_JS 1

#include "co_lib.h"
#include "co_mod.h"
#include "co_js.h"

/* mod coffice functions: */
#include "co_session.h"

#include "co_result.h"
#include "mod_coffice.h"
#include "co_recordset.h"
#include "co_array.h"
#include "co_string.h"
#include "co_defines.h"

/*********************/
/* UTILITY FUNCTIONS */
/*********************/
/* borrowed from co_js.c */
extern
co_context*
cjs_getCOContext(JSContext *cx);

/* borrowed from co_broker.c */
extern
const char *cox_getTargetClassName( co_context *cnx, 
    const char* area_name, const char* class_name, const char* method_name, int *err_ptr );

/* requires class_name being a name of a real submodule */
static
void *
cox_getNewParam( co_context *cnx, const char* class_name )
{
    array_header* lib_array;
    coffice_libinfo* elts;
    coffice_libinfo* elt;
    int i;

    lib_array = cox_getCOfficeConfig(cnx)->loaded_libs;

    elts = (coffice_libinfo*)lib_array->elts;
    for (i = 0; i < lib_array->nelts; i++) {
        elt = &elts[i];
        if (strcmp(elt->name, class_name) == 0) {
	    return elt->new_param;
        }
   }
   return NULL; /* class_name not found */
}

static
void *
cox_getSetParam( co_context *cnx, const char* class_name )
{
    array_header* lib_array;
    coffice_libinfo* elts;
    coffice_libinfo* elt;
    int i;

    lib_array = cox_getCOfficeConfig(cnx)->loaded_libs;

    elts = (coffice_libinfo*)lib_array->elts;
    for (i = 0; i < lib_array->nelts; i++) {
        elt = &elts[i];
        if (strcmp(elt->name, class_name) == 0) {
	    return elt->set_param;
        }
   }
   return NULL; /* class_name not found */
}

static
void *
cox_getGetParam( co_context *cnx, const char* class_name )
{
    array_header* lib_array;
    coffice_libinfo* elts;
    coffice_libinfo* elt;
    int i;

    lib_array = cox_getCOfficeConfig(cnx)->loaded_libs;

    elts = (coffice_libinfo*)lib_array->elts;
    for (i = 0; i < lib_array->nelts; i++) {
        elt = &elts[i];
        if (strcmp(elt->name, class_name) == 0) {
	    return elt->get_param;
        }
   }
   return NULL; /* class_name not found */
}


/***********************/
/** CLASS DEFINITIONS **/
/***********************/

static JSClass comethod_class;

typedef struct {
  const char *strArea;
  const char *strClass;
  const char *strMethod;
  const char *strTargetClass;
  void *ptrParam;
} COMethod_data;

/* allocates memory for private data */
static COMethod_data *
comethod_init(JSContext *cx, JSObject *obj)
{
    COMethod_data *data;
    data = (COMethod_data*)JS_malloc(cx, sizeof(COMethod_data) );
//    data = ap_palloc(cjs_getCOContext(cx)->request->pool, sizeof(COMethod_data) );

    if (!data) return NULL;

    if (!JS_SetPrivate(cx, obj, data)) {
/*
        JS_ReportErrorNumber(cx, JSFile_GetErrorMessage, NULL,
            JSFILEMSG_CANNOT_SET_PRIVATE_FILE, file->path);
*/
//        JS_free(cx, data);
        return NULL;
    } else
        return data;
};

/* TODO: protect strings from GC */
static JSBool
COMethod(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
    JSString *strArea;
    JSString *strClass;
    JSString *strMethod;
    COMethod_data *data;
    const char* par_area; /* save cnx->par_area */
    jsval keep_area, keep_class, keep_method;

    if (argc >= 3) {
        strArea = JS_ValueToString(cx, argv[0]);
        if (!strArea)
            return JS_FALSE;
	keep_area = STRING_TO_JSVAL(strArea);

        strClass = JS_ValueToString(cx, argv[1]);
        if (!strClass)
            return JS_FALSE;
	keep_class = STRING_TO_JSVAL(strClass);

        strMethod = JS_ValueToString(cx, argv[2]);
        if (!strMethod)
            return JS_FALSE;
	keep_method = STRING_TO_JSVAL(strMethod);


	if ( !(data = comethod_init(cx, obj ))) return JS_FALSE;

	data->strArea = JS_GetStringBytes(strArea);
	data->strClass = JS_GetStringBytes(strClass);
	data->strMethod = JS_GetStringBytes(strMethod);

	{ /* allocate structure for parameters */
//	    void* param_ptr;
	    co_context *cnx;
	    int err=0;

	    /* new_param is a pointer to  the param allocator */
	    void * (*new_param)( co_context *, const char*);

	    cnx = cjs_getCOContext(cx);

	    /* resolve class_name -> new_param pointer */
	    data->strTargetClass = cox_getTargetClassName( cnx, data->strArea, 
		data->strClass, data->strMethod, &err );
	    cou_logError( cnx, CO_LOG_DEBUG, "Got target: %s", data->strTargetClass);
	    if(err != CO_NO_ERRORS) { /* class/method not found in vtable */
                JS_ReportError(cx,"COMethod: no implementation for: %s::%s::%s\n",
		    data->strArea, data->strClass, data->strMethod
		    );
    		return JS_FALSE;
	    }

	    if ( !(new_param = cox_getNewParam( cnx, data->strTargetClass ) ))
        	return JS_FALSE;

	    if ( !(data->ptrParam = new_param( cnx, data->strMethod) ) )
        	return JS_FALSE;

	}

    } else {
/*        str = cx->runtime->emptyString;*/
        return JS_FALSE;
    }
/*
    if (!cx->fp->constructing) {
        *rval = STRING_TO_JSVAL(strMethod);
        return JS_TRUE;
    }
*/
/*    OBJ_SET_SLOT(cx, obj, JSSLOT_PRIVATE, STRING_TO_JSVAL(str)); */

    return JS_TRUE;
}


static void
comethod_finalize(JSContext *cx, JSObject *obj)
{
    COMethod_data *data = JS_GetInstancePrivate(cx, obj, &comethod_class, NULL);

    cou_logError( cjs_getCOContext(cx), CO_LOG_DEBUG, "Finalizing COMethod" );

//    if ((data!=NULL)&&(data->strArea!=NULL)) {
    if (data) {
	cou_logError( cjs_getCOContext(cx), CO_LOG_DEBUG, "destroying data" );
	data->strArea = NULL;
	data->strClass = NULL;
	data->strMethod = NULL;
	data->strTargetClass = NULL;
	data->ptrParam = NULL;
	JS_free(cx, data);
	cou_logError( cjs_getCOContext(cx), CO_LOG_DEBUG, "data destroyed" );
    }
};

static JSClass comethod_class = {
    "COMethod", JSCLASS_HAS_PRIVATE,
    JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub,
    JS_EnumerateStub,   JS_ResolveStub,     JS_ConvertStub,  comethod_finalize,
    JSCLASS_NO_OPTIONAL_MEMBERS
};

static JSBool
comethod_setParam(JSContext *cx, JSObject *obj, uintN argc, jsval *argv,
             jsval *rval)
{
    COMethod_data *data = JS_GetInstancePrivate(cx, obj, &comethod_class, NULL);
    int (*set_param)( co_context *, void *, const char*, const char*, void *);
    co_context *cnx;
    jsval param_value;
    JSString *param_name;

    if ( !(param_name=JS_ValueToString(cx, argv[0])) )
            return JS_FALSE;
    /* root as you go */
    argv[0] = STRING_TO_JSVAL(param_name);

    param_value = argv[1];

    cnx = cjs_getCOContext(cx);

    cou_logError( cnx, CO_LOG_DEBUG, "setParam: finding ");

    if ( !(set_param = cox_getSetParam( cnx, data->strTargetClass )) )
            return JS_FALSE;

    cou_logError( cnx, CO_LOG_DEBUG, "setParam: found set_param, setting ");

    if (JSVAL_IS_INT(param_value)) {
	int int_value;

	cou_logError( cnx, CO_LOG_DEBUG, "setParam: value is int");

	JS_ValueToInt32( cx, param_value, &int_value );
	set_param( cnx,
	    data->ptrParam, 
	    data->strMethod,
	    JS_GetStringBytes( param_name ) ,
	    (void *)&int_value);
	return JS_TRUE;
    } else if (JSVAL_IS_STRING(param_value)) {
	JSString* str_value;

	cou_logError( cnx, CO_LOG_DEBUG, "setParam: value is string");

	if ( !(str_value=JS_ValueToString(cx, param_value)) )
            return JS_FALSE;
	set_param( cnx,
	    data->ptrParam, 
	    data->strMethod,
	    JS_GetStringBytes( param_name ) ,
	    JS_GetStringBytes( str_value ) );
	return JS_TRUE;
    } else if (JSVAL_IS_OBJECT(param_value)) {
	JSObject* obj_value;
	jsint length;
	jsval el_value;
	co_array_of_pointers * c_array;
	jsint index;

        c_array = coa_Create(cnx);
  
	cou_logError( cnx, CO_LOG_DEBUG, "setParam: the value is an object");

	if ( !(obj_value=JSVAL_TO_OBJECT(param_value)) )
            return JS_FALSE;
	if (!JS_GetArrayLength(cx, obj_value, &length ))
            JS_ReportError(cx,"COMethod::setParam: argument is an object, but length cannot be determined." );

	for (index=0; index<length; index++) {
	    if (!JS_GetElement(cx, obj_value, index, &el_value))
        	JS_ReportError(cx,"COMethod::setParam: can not retrieve element for index=%d", index );
	    if (!JSVAL_IS_STRING(el_value))
        	JS_ReportError(cx,"COMethod::setParam: element is not string for index=%d", index );
	    coa_AppendElement(cnx, c_array, JS_GetStringBytes(JSVAL_TO_STRING(el_value)));
	}
	set_param( cnx,
	    data->ptrParam, 
	    data->strMethod,
	    JS_GetStringBytes( param_name ) ,
	    c_array );
	return JS_TRUE;

    }

    JS_ReportError(cx,"COMethod::setParam: unknown type" );

    return JS_FALSE;
};

static JSBool
comethod_call(JSContext *cx, JSObject *obj, uintN argc, jsval *argv,
             jsval *rval)
{
    sigjmp_buf sigenv_tmp;
    const char* par_area; /* save cnx->par_area */
    COMethod_data *data = JS_GetInstancePrivate(cx, obj, &comethod_class, NULL);
    co_context *cnx = cjs_getCOContext(cx);

    par_area = cnx->par_area;

    cnx->par_area = data->strArea; /* switch action queue */

	sigenv_tmp[0] = cnx->sigenv[0];
    if ( sigsetjmp(cnx->sigenv, CO_ERR_SAVE_SIGNALS) == 0) {
		cou_broker( 
			data->strArea, 
			data->strClass, 
			data->strMethod, 
			cnx, data->ptrParam );
    }
    else {
		/* error */
		/* restore error context */
		cnx->sigenv[0] = sigenv_tmp[0];
		/* report error */
		JS_ReportError(cx,"COMethod::call: forced error while calling %s::%s::%s inherited from %s",
    	    data->strArea, 
	    	data->strClass, 
	 		data->strMethod, 
	    	data->strTargetClass
		);
#if CO_USES_EXCEPTIONS
		JS_SetPendingException(cx, JSVAL_TRUE);
#endif
        return JS_FALSE;
    }
	/* restore error context */
	cnx->sigenv[0] = sigenv_tmp[0];

    cnx->par_area = par_area; /* restore previous action queue */

    if (cnx->error != CO_NO_ERRORS) {
#if CO_USES_EXCEPTIONS
		JS_SetPendingException(cx, JSVAL_TRUE);
#endif
        return JS_FALSE;
    }

    cou_logError( cnx, CO_LOG_DEBUG, "co_js_broker: returns to JS");
        return JS_TRUE;
};

static JSFunctionSpec comethod_methods[] = {

    {"setParam",       comethod_setParam,       2,0,0 },

    {"call",           comethod_call,           0,0,0 },
    {0,0,0,0,0}
};

JSObject*
comethod_Init(JSContext *cx)
{
    return JS_InitClass(cx, JS_GetGlobalObject(cx), NULL, &comethod_class, COMethod, 3, NULL,comethod_methods,NULL,NULL);
}

/**********************/
/***** Recordset ******/
/**********************/

static JSClass recordset_class;

static JSBool
Recordset(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
    co_context* cocnx;

    cocnx = cjs_getCOContext(cx);

    cou_logError( cocnx, CO_LOG_DEBUG, "Recordset constructor: begins");

    /* check if really recordset was returned */
    if (cocnx->result_type != CO_RECORDSET) {
	return JS_FALSE;
    }

    /* take responsibility for the query result (recordset) */
    if (!JS_SetPrivate(cx, obj, cocnx->result)) {
	return JS_FALSE;
    }
    /* protect result form other calls */
    cocnx->result = NULL;
    cocnx->result_type = CO_NO_TYPE;
/*    cocnx->dbconn->result = NULL; is equivalent to:
    cox_clearConnectionResult( cocnx );
*/
    cou_logError( cocnx, CO_LOG_DEBUG, "Recordset constructor: returns");
    
    return JS_TRUE;
}


static void
recordset_finalize(JSContext *cx, JSObject *obj)
{
    co_recordset *data;
    co_context * cnx =  cjs_getCOContext(cx);

    cou_logError(cnx, CO_LOG_DEBUG, "Finalizing Recordset: Start" );
    data = JS_GetInstancePrivate(cx, obj, &recordset_class, NULL);
    if (data) {
		cou_logError( cjs_getCOContext(cx), CO_LOG_DEBUG, "clearing result" );
		cor_Clear(cnx, data);
		/* remove reference to cleared data */
 		if (!JS_SetPrivate(cx, obj, NULL)) {
			JS_ReportError(cx,"Cannot remove reference to cleared recorset!");
			return;
		}
    }
    cou_logError( cnx, CO_LOG_DEBUG, "Finalizing Recordset: End" );
}

static JSClass recordset_class = {
    "Recordset", JSCLASS_HAS_PRIVATE,
    JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub,
    JS_EnumerateStub,   JS_ResolveStub,     JS_ConvertStub,  recordset_finalize,
    JSCLASS_NO_OPTIONAL_MEMBERS
};

static JSBool
recordset_getValue(JSContext *cx, JSObject *obj, uintN argc, jsval *argv,
             jsval *rval)
{
    JSString *str;
    co_context * cnx =  cjs_getCOContext(cx);
    co_recordset *data = JS_GetInstancePrivate(cx, obj, &recordset_class, NULL);
    const char * ret_val;
    int32 tuple_num;
    int32 field_num;
    char MESGString[MESGSIZE];

    cou_logError( cnx, CO_LOG_DEBUG, "Recordset.getValue" );

    if (!data) {
	return JS_FALSE;
    }

    /* 2 arguments required */
    if (argc < 2) {
		JS_ReportError(cx,"Recordset.getValue requires 2 arguments");
		return JS_FALSE;
    }

    tuple_num = JSVAL_TO_INT(argv[0]);
    field_num = JSVAL_TO_INT(argv[1]);
    ret_val = cor_GetValue(cnx,data, tuple_num, field_num);

    __SEND_MESG_DEBUG("JS getValue: %s",ret_val)

    if(ret_val==NULL)
    {
	    JS_ReportError(cx,"Error in getting value from recordset column num: %d , tuple num: %d", field_num, tuple_num);
            return JS_FALSE;
    }

    str = JS_NewStringCopyZ( cx, ret_val);
    *rval = STRING_TO_JSVAL(str) ;
    return JS_TRUE;
}

static JSBool
recordset_getValueByName(JSContext *cx, JSObject *obj, uintN argc, jsval *argv,
             jsval *rval)
{
    JSString *str;
    JSString *param_name;
    const char* param_str;
    const char* param_str2;
    int32 param_num;
    int32 tuple_num;
    co_context * cnx =  cjs_getCOContext(cx);
    co_recordset *data = JS_GetInstancePrivate(cx, obj, &recordset_class, NULL);

    cou_logError(cnx, CO_LOG_DEBUG, "Recordset.getValueByName" );

    if (!data) {
	return JS_FALSE;
    }

    /* 2 arguments required */
    if (argc < 2) {
	JS_ReportError(cx,"Recordset.getValueByName requires 2 arguments");
	return JS_FALSE;
    }

    if ( !(param_name=JS_ValueToString(cx, argv[1])) ) {
	    JS_ReportError(cx,"Missing name in Recordset.getValueByName");
            return JS_FALSE;
    }
    /* root as you go */
    argv[1] = STRING_TO_JSVAL(param_name);

    param_str = JS_GetStringBytes(param_name);
    param_num = cor_GetFieldNum(cnx, data, param_str);

    if (param_num<0) {
	    JS_ReportError(cx,"Column name %s not found in Recordset", param_str);
            return JS_FALSE;
    }

    tuple_num = JSVAL_TO_INT(argv[0]);
    if(!(param_str2=cor_GetValue(cnx,data, tuple_num, param_num)))
    {
	    JS_ReportError(cx,"Error in getting value from recordset Column name: %s , tuple num: %d", param_str, tuple_num);
            return JS_FALSE;
    }

    str = JS_NewStringCopyZ( cx, param_str2);
    *rval = STRING_TO_JSVAL(str) ;
    return JS_TRUE;
}

static JSBool
recordset_getTupleCount(JSContext *cx, JSObject *obj, uintN argc, jsval *argv,
             jsval *rval)
{
    co_context * cnx =  cjs_getCOContext(cx);
    co_recordset *data = JS_GetInstancePrivate(cx, obj, &recordset_class, NULL);
    int32 NTuples;

    cou_logError(cnx, CO_LOG_DEBUG, "Recordset.getTupleCount" );
    if (!data) {
	return JS_FALSE;
    }
   
    NTuples = cor_GetTupleCount(cnx,data);
    if (NTuples<0) {
	    JS_ReportError(cx,"Error in getting number of tuples");
            return JS_FALSE;
    }

    *rval = INT_TO_JSVAL(NTuples);
    return JS_TRUE;
}

static JSBool
recordset_getFieldCount(JSContext *cx, JSObject *obj, uintN argc, jsval *argv,
             jsval *rval)
{
    co_context * cnx =  cjs_getCOContext(cx);
    co_recordset *data = JS_GetInstancePrivate(cx, obj, &recordset_class, NULL);
    int32 NFields;   
 
    cou_logError(cnx, CO_LOG_DEBUG, "Recordset.getFieldCount: Start" );
    if (!data) {
	return JS_FALSE;
    }

    NFields = cor_GetFieldCount(cnx,data);
    if (NFields<0) {
	    JS_ReportError(cx,"Error in getting number of fields");
            return JS_FALSE;
    }

    *rval = INT_TO_JSVAL(NFields);
    cou_logError(cnx, CO_LOG_DEBUG, "Recordset.getFieldCount: End" );
    return JS_TRUE;
}

static JSBool
recordset_getFieldName(JSContext *cx, JSObject *obj, uintN argc, jsval *argv,
             jsval *rval)
{
    JSString *str;
    JSString *param_name;
    int32 param_num;
    int32 param_int;
    const char * param_str;
    co_context * cnx =  cjs_getCOContext(cx);
    co_recordset *data = JS_GetInstancePrivate(cx, obj, &recordset_class, NULL);

    cou_logError(cnx, CO_LOG_DEBUG, "Recordset.getFieldName" );
    if (!data) {
	return JS_FALSE;
    }
    
    /* 2 arguments required */
    if (argc < 1) {
	JS_ReportError(cx,"Recordset.getFieldName requires 1 argument");
	return JS_FALSE;
    }

    param_int = JSVAL_TO_INT(argv[0]);
    if(!(param_str= cor_GetFieldName(cnx,data, param_int)))
    {
	JS_ReportError(cx,"Recordset.getFieldName field not found, for number: %d", param_int);
	return JS_FALSE;
    }

    str = JS_NewStringCopyZ(cx,param_str);
    *rval = STRING_TO_JSVAL(str) ;
    return JS_TRUE;
}


static JSBool
recordset_getFieldNum(JSContext *cx, JSObject *obj, uintN argc, jsval *argv,
             jsval *rval)
{
    JSString *str;
    JSString *param_name;
    int32 param_num;
    const char * param_str;
    co_context * cnx =  cjs_getCOContext(cx);
    co_recordset *data = JS_GetInstancePrivate(cx, obj, &recordset_class, NULL);
    

    cou_logError(cnx, CO_LOG_DEBUG, "Recordset.getFieldName" );
    if (!data) {
	return JS_FALSE;
    }

    /* 2 arguments required */
    if (argc < 1) {
	JS_ReportError(cx,"Recordset.getFieldName requires 1 argument");
	return JS_FALSE;
    }

    if ( !(param_name=JS_ValueToString(cx, argv[0])) ) {
	    JS_ReportError(cx,"Missing name in Recordset.getFieldName");
            return JS_FALSE;
    }
 
    argv[0] = STRING_TO_JSVAL(param_name);

    param_str = JS_GetStringBytes(param_name);
    param_num = cor_GetFieldNum(cnx, data, param_str);

    if (param_num<0) {
	    JS_ReportError(cx,"Column name %s not found in Recordset", param_str);
            return JS_FALSE;
    }

    *rval = INT_TO_JSVAL(param_num);
    return JS_TRUE;
}



static JSFunctionSpec recordset_methods[] = {

    {"getValue",       recordset_getValue,       2,0,0 },

    {"getValueByName", recordset_getValueByName, 2,0,0 },

    {"getTupleCount",  recordset_getTupleCount,  0,0,0 },

    {"getFieldCount",  recordset_getFieldCount,  0,0,0 },
      {"getFieldName",  recordset_getFieldName,    1,0,0 },
      {"getFieldNum",  recordset_getFieldNum,    1,0,0 },
      {0,0,0,0,0}
};

JSObject*
recordset_Init(JSContext *cx)
{
    return JS_InitClass(cx, JS_GetGlobalObject(cx), NULL, &recordset_class, Recordset, 0, NULL,recordset_methods,NULL,NULL);
}
/**********************/
/****** Request *******/
/**********************/
static JSBool
Request(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
    return JS_TRUE;
}

static JSClass request_class = {
    "Request", 0,
    JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub,
    JS_EnumerateStub,   JS_ResolveStub,     JS_ConvertStub,  JS_FinalizeStub,
    JSCLASS_NO_OPTIONAL_MEMBERS
};

static JSBool
request_QueryString(JSContext *cx, JSObject *obj, uintN argc, jsval *argv,
             jsval *rval)
{
    int err=0;
    const char *param_name;
    const char *param_value;
    JSString *str;

    if (argc==0) {
	param_value = cjs_getCOContext(cx)->method_args;
    }
    else {
	param_name = JS_GetStringBytes( JS_ValueToString(cx, argv[0]) );
	param_value = (char*)cou_getParam( cjs_getCOContext(cx), param_name, CO_STRING, &err);
    }
    str = JS_NewStringCopyZ( cx, param_value );
    *rval = STRING_TO_JSVAL(str) ;

    return JS_TRUE;
}

static JSFunctionSpec request_static_methods[] = {
    {"QueryString",  request_QueryString,  1,0,0 },
    {0,0,0,0,0}
};

JSObject*
request_Init(JSContext *cx)
{
    return JS_InitClass(cx, JS_GetGlobalObject(cx), NULL, &request_class, Request, 0, NULL, NULL, NULL, request_static_methods);
}

/**********************/
/****** Params *******/
/**********************/
static JSBool
Params(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
    return JS_TRUE;
}

static JSClass params_class = {
    "Params", 0,
    JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub,
    JS_EnumerateStub,   JS_ResolveStub,     JS_ConvertStub,  JS_FinalizeStub,
    JSCLASS_NO_OPTIONAL_MEMBERS
};

static JSBool
params_getParam(JSContext *cx, JSObject *obj, uintN argc, jsval *argv,
             jsval *rval)
{
    int err=0;
    const char *param_name;
    const char *method_name;
    const char *target_class;
    void *ptr_param_value;
    co_context *cnx = cjs_getCOContext(cx);
    void * (*get_param)( co_context *, void *, const char*, const char*, co_value_type *);
    co_value_type param_type;

    if (argc==0) {
	JS_ReportError(cx,"Missing param name in Params::getParam");
    	return JS_FALSE;
    }

    method_name = cjs_getCOMethodName(cx);
    target_class = cjs_getCOTargetName(cx);
/*
cox_getTargetClassName( cnx, cnx->par_area, 
		cnx->par_class, method_name, &err );
*/
    if (err) {
	JS_ReportError(cx,"Target class not found in Params::getParam");
        return JS_FALSE;
    }

    if ( !(get_param = cox_getGetParam( cnx, target_class )) ) {
	JS_ReportError(cx,"Cannot find param getter in Params::getParam");
        return JS_FALSE;
    }

    param_name = JS_GetStringBytes( JS_ValueToString(cx, argv[0]) );

        cou_logError( cnx, CO_LOG_DEBUG, "get_param looking for: '%s' in method '%s'", param_name, method_name);

    ptr_param_value = get_param( cnx,
	    cjs_getCOInParams(cx), 
	    method_name,
	    param_name, &param_type);

    if (ptr_param_value==NULL) {
	JS_ReportError(cx,"Cannot get param value in Params::getParam");
        return JS_FALSE;
    }

    if ((param_type==CO_STRING)||(param_type==CO_STRING_SQL_NAME)) {
	JSString *str;
	str = JS_NewStringCopyZ( cx, *(char**)ptr_param_value );
	*rval = STRING_TO_JSVAL(str) ;
	return JS_TRUE;
    }
    else if (param_type==CO_INTEGER) {
	jsdouble d;
	d = *(int*)ptr_param_value;
        return JS_NewNumberValue(cx, d, rval);
    }
    else {
	JS_ReportError(cx,"Unknown type of param value in Params::getParam");
        return JS_FALSE;
    }
}

static JSBool
params_setParam(JSContext *cx, JSObject *obj, uintN argc, jsval *argv,
             jsval *rval)
{
    int err=0;
    const char *param_name;
    const char *method_name;
    const char *target_class;
    jsval param_value;
    co_context *cnx = cjs_getCOContext(cx);
    int (*set_param)( co_context *, void *, const char*, const char*, void *);
    co_value_type param_type;

    if (argc<2) {
	JS_ReportError(cx,"Missing param name or value in Params::setParam");
    	return JS_FALSE;
    }

    method_name = cjs_getCOMethodName(cx);
    target_class = cjs_getCOTargetName(cx);
/*
cox_getTargetClassName( cnx, cnx->par_area, 
		cnx->par_class, method_name, &err );
*/
    if (err) {
	JS_ReportError(cx,"Target class not found in Params::setParam");
        return JS_FALSE;
    }

    if ( !(set_param = cox_getSetParam( cnx, target_class )) ) {
	JS_ReportError(cx,"Cannot find param getter in Params::setParam");
        return JS_FALSE;
    }

    param_name = JS_GetStringBytes( JS_ValueToString(cx, argv[0]) );

        cou_logError( cnx, CO_LOG_DEBUG, "set_param looking for: '%s' in method '%s'", param_name, method_name);

    param_value = argv[1];

    if (JSVAL_IS_INT(param_value)) {
	int int_value;

	cou_logError( cnx, CO_LOG_DEBUG, "setParam: value is int");

	JS_ValueToInt32( cx, param_value, &int_value );
	set_param( cnx,
	    cjs_getCOInParams(cx), 
	    method_name,
	    param_name, (void *)&int_value);
	return JS_TRUE;
    } else if (JSVAL_IS_STRING(param_value)) {
	JSString* str_value;

	cou_logError( cnx, CO_LOG_DEBUG, "setParam: value is string");

	if ( !(str_value=JS_ValueToString(cx, param_value)) )
            return JS_FALSE;
	set_param( cnx,
	    cjs_getCOInParams(cx), 
	    method_name,
	    param_name, JS_GetStringBytes( str_value ));
	return JS_TRUE;
    } else if (JSVAL_IS_OBJECT(param_value)) {
	JSObject* obj_value;
	jsint length;
	jsval el_value;
	co_array_of_pointers * c_array;
	jsint index;

        c_array = coa_Create(cnx);
  
	cou_logError( cnx, CO_LOG_DEBUG, "setParam: the value is an object");

	if ( !(obj_value=JSVAL_TO_OBJECT(param_value)) )
            return JS_FALSE;
	if (!JS_GetArrayLength(cx, obj_value, &length ))
            JS_ReportError(cx,"Param::setParam: argument is an object, but length cannot be determined." );

	for (index=0; index<length; index++) {
	    if (!JS_GetElement(cx, obj_value, index, &el_value))
        	JS_ReportError(cx,"Param::setParam: can not retrieve element for index=%d", index );
	    if (!JSVAL_IS_STRING(el_value))
        	JS_ReportError(cx,"Param::setParam: element is not string for index=%d", index );
	    coa_AppendElement(cnx, c_array, JS_GetStringBytes(JSVAL_TO_STRING(el_value)));
	}
	set_param( cnx,
	    cjs_getCOInParams(cx), 
	    method_name,
	    param_name, c_array);
	return JS_TRUE;

    }

    JS_ReportError(cx,"Param::setParam: unknown type" );

    return JS_FALSE;

}

static JSFunctionSpec params_static_methods[] = {
    {"getParam",  params_getParam,  1,0,0 },
    {"setParam",  params_setParam,  2,0,0 },
    {0,0,0,0,0}
};

JSObject*
params_Init(JSContext *cx)
{
    return JS_InitClass(cx, JS_GetGlobalObject(cx), NULL, &params_class, Params, 0, NULL, NULL, NULL, params_static_methods);
}

/**********************/
/****** COServer ******/
/**********************/
static JSBool
COServer(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
    return JS_TRUE;
}

static JSClass coserver_class = {
    "COServer", 0,
    JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub,
    JS_EnumerateStub,   JS_ResolveStub,     JS_ConvertStub,  JS_FinalizeStub,
    JSCLASS_NO_OPTIONAL_MEMBERS
};

static JSBool
coserver_getUser(JSContext *cx, JSObject *obj, uintN argc, jsval *argv,
             jsval *rval)
{
    return JS_NewNumberValue(cx, cjs_getCOContext(cx)->id_user, rval);
}

static JSBool
coserver_getRole(JSContext *cx, JSObject *obj, uintN argc, jsval *argv,
             jsval *rval)
{
    return JS_NewNumberValue(cx, cjs_getCOContext(cx)->id_role, rval);
}

static JSBool
coserver_getSession(JSContext *cx, JSObject *obj, uintN argc, jsval *argv,
             jsval *rval)
{
    co_context *cnx;
    cnx = cjs_getCOContext(cx);
    
    if (cnx->id_session == 0)
	cou_assignNewSession( cnx );
        
    return JS_NewNumberValue(cx, cnx->id_session, rval);
}

static JSBool
coserver_getAreaName(JSContext *cx, JSObject *obj, uintN argc, jsval *argv,
             jsval *rval)
{
    JSString *str = JS_NewStringCopyZ( cx, cjs_getCOContext(cx)->par_area );
    *rval = STRING_TO_JSVAL(str) ;
    return JS_TRUE;
}

static JSBool
coserver_getClassName(JSContext *cx, JSObject *obj, uintN argc, jsval *argv,
             jsval *rval)
{
    JSString *str = JS_NewStringCopyZ( cx, cjs_getCOContext(cx)->par_class );
    *rval = STRING_TO_JSVAL(str) ;
    return JS_TRUE;
}

static JSBool
coserver_getImagePath(JSContext *cx, JSObject *obj, uintN argc, jsval *argv,
             jsval *rval)
{
    JSString *str = JS_NewStringCopyZ( cx, cou_getCOImagePath() );
    *rval = STRING_TO_JSVAL(str) ;
    return JS_TRUE;
}

static JSBool
coserver_useWritesBuffer(JSContext *cx, JSObject *obj, uintN argc, jsval *argv,
             jsval *rval)
{
    co_js_context* cocx;
    jschar *chars;
    cocx = (co_js_context*)JS_GetContextPrivate(cx);

    chars = (jschar *) JS_malloc(cx, sizeof(jschar));
    chars[0] = 0;
    cocx->writes_buffer = chars;
    cocx->buffer_length = 0;
    return JS_TRUE;
}

static JSBool
coserver_getWritesBuffer(JSContext *cx, JSObject *obj, uintN argc, jsval *argv,
             jsval *rval)
{
    JSString* str;
    co_js_context* cocx;
    jschar *chars;
    cocx = (co_js_context*)JS_GetContextPrivate(cx);

    chars = (jschar *) (cocx->writes_buffer);

    /* create new String containing buffer */
    str = JS_NewUCString(cx, chars, cocx->buffer_length);
    if (!str)
	return JS_FALSE;

    *rval = STRING_TO_JSVAL(str) ;
    return JS_TRUE;
}

static JSBool
coserver_useWritesDirect(JSContext *cx, JSObject *obj, uintN argc, jsval *argv,
             jsval *rval)
{
    co_js_context* cocx;
    cocx = (co_js_context*)JS_GetContextPrivate(cx);

    cocx->writes_buffer = NULL;
    cocx->buffer_length = 0;

    return JS_TRUE;
}
///////////////////////////////
static JSBool
coserver_beginJSCreation(JSContext *cx, JSObject *obj, uintN argc, jsval *argv,
             jsval *rval)
{
    co_context *cnx;
    co_js_context* cocx;
    cocx = (co_js_context*)JS_GetContextPrivate(cx);
    cnx = cjs_getCOContext(cx);

	if (cocx->withinJSCreation) {
		/* do nothing */
	    return JS_TRUE;
	}
	cocx->withinJSCreation = 1;
	cocx->saved_client_mode=cnx->client_mode;
	cocx->saved_send_data=cnx->send_data;
	
	cnx->client_mode=CO_CLIENT_JS;
	cnx->send_data=1;

    return JS_TRUE;
}

static JSBool
coserver_endJSCreation(JSContext *cx, JSObject *obj, uintN argc, jsval *argv,
             jsval *rval)
{
    co_context *cnx;
    co_js_context* cocx;
    cocx = (co_js_context*)JS_GetContextPrivate(cx);
    cnx = cjs_getCOContext(cx);

	if (!cocx->withinJSCreation) {
		/* do nothing */
	    return JS_TRUE;
	}
	/* mark end */
	cocx->withinJSCreation = 0;
	/* restore js mode of server */
	cnx->client_mode = cocx->saved_client_mode;
	cnx->send_data = cocx->saved_send_data;
	
    return JS_TRUE;
}

static JSBool
coserver_sendJSBuffer(JSContext *cx, JSObject *obj, uintN argc, jsval *argv,
             jsval *rval)
{
    co_context *cnx;
    cnx = cjs_getCOContext(cx);
	cos_SendString(cnx, cnx->js_response);
    return JS_TRUE;
}

static JSBool
coserver_clearJSBuffer(JSContext *cx, JSObject *obj, uintN argc, jsval *argv,
             jsval *rval)
{
    co_context *cnx;
    cnx = cjs_getCOContext(cx);
	cnx->js_response = cos_CreateEmpty(cnx);
    return JS_TRUE;
}

static JSFunctionSpec coserver_static_methods[] = {

    {"getUserId",		coserver_getUser,	0,0,0 },

    {"getRoleId",		coserver_getRole,	0,0,0 },

    {"getSessionId",	coserver_getSession,	0,0,0 },

    {"getAreaName",	coserver_getAreaName,	0,0,0 },

    {"getClassName",	coserver_getClassName,	0,0,0 },

    {"getImagePath",	coserver_getImagePath,	0,0,0 },

    {"useWritesBuffer",	coserver_useWritesBuffer,	0,0,0 },

    {"getWritesBuffer",	coserver_getWritesBuffer,	0,0,0 },

    {"useWritesDirect",	coserver_useWritesDirect,	0,0,0 },

    {"beginJSCreation",	coserver_beginJSCreation,	0,0,0 },

    {"endJSCreation",	coserver_endJSCreation,	0,0,0 },

    {"sendJSBuffer",	coserver_sendJSBuffer,	0,0,0 },

    {"clearJSBuffer",	coserver_clearJSBuffer,	0,0,0 },
    {0,0,0,0,0}
};

JSObject*
coserver_Init(JSContext *cx)
{
    return JS_InitClass(cx, JS_GetGlobalObject(cx), NULL, &coserver_class, COServer, 0, NULL, NULL, NULL, coserver_static_methods);
}

/**********************/
/** GLOBAL FUNCTIONS **/
/**********************/

static JSBool
    co_js_apwrite(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
    {
        JSString* str;

	size_t length, buffer_length;
	jschar *chars, *writes_buffer;
	co_js_context* cocx;

        if ( !(str=JS_ValueToString(cx, argv[0])) )
            return JS_FALSE;

	cocx = (co_js_context*)JS_GetContextPrivate(cx);

	writes_buffer = (jschar *)((cocx)->writes_buffer);

	if (writes_buffer) {
	    /* see also: str_concat in String implementation */
	    buffer_length = ((co_js_context*)JS_GetContextPrivate(cx))->buffer_length;
	    /* allocate new buffer */
	    length = buffer_length + str->length;
	    chars = (jschar *) JS_malloc(cx, (length + 1) * sizeof(jschar));
	    if (!chars)
    		return JS_FALSE;
	    /* copy both parts */
	    js_strncpy(chars, writes_buffer, buffer_length);
    	    js_strncpy(chars + buffer_length, str->chars, str->length);
    	    chars[length] = 0;
	    /* update buffer */
	    cocx->writes_buffer = chars;
	    cocx->buffer_length = length;
	}
	else {
	    ap_rputs( JS_GetStringBytes( str ), cjs_getCOContext(cx)->request );
	}
        return JS_TRUE;
    }

static JSBool
    co_js_broker(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
    {
        JSString* area_name;
        JSString* class_name;
        JSString* method_name;
	int in_params; 

	co_context *cnx = cjs_getCOContext(cx);
    
        if ( !(area_name=JS_ValueToString(cx, argv[0])) )
            return JS_FALSE;

        if ( !(class_name=JS_ValueToString(cx, argv[1])) )
            return JS_FALSE;

        if ( !(method_name=JS_ValueToString(cx, argv[2])) )
            return JS_FALSE;

	if ( JSVAL_IS_INT(argv[3]) )
    	    JS_ValueToInt32( cx, argv[3], &in_params );
	else
            return JS_FALSE;

	cou_broker( 
	    JS_GetStringBytes( area_name ), 
	    JS_GetStringBytes( class_name ), 
	    JS_GetStringBytes( method_name ), 
	    cnx, (void*)in_params ); /* TODO: PARAMETERS */

	if (cnx->error != CO_NO_ERRORS) {
	    JS_ReportError(cx,"Error in co_broker");
    	    return JS_FALSE;
	}

	cou_logError( cnx, CO_LOG_DEBUG, "co_js_broker: returns");
        return JS_TRUE;
    }
    
static JSBool
    co_js_getretAny(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
    {
	int* retval_ptr;
	char* retval_sptr;
	int co_err;
	jsdouble d;
//	jsval tmpval;
	JSString *str;
	JSObject *newobj;

	co_context* cocnx;

	d = 0; /* use it as 'defualt number value' */
	cocnx = cjs_getCOContext(cx);
	cou_logError( cocnx, CO_LOG_DEBUG, "co_js_getretAny called");

	switch(cocnx->result_type) {
	case CO_INTEGER:
	    retval_ptr = (int*)cou_getReturnValue( cocnx, CO_INTEGER, &co_err );
	    if (retval_ptr) {
		cou_logError( cocnx, CO_LOG_DEBUG, "co_js_getretAny: got Integer value");
		d = *retval_ptr;
	    }
    	    return JS_NewNumberValue(cx, d, rval);
	    break;
	case CO_STRING:
	case CO_STRING_SQL_VALUE:
	case CO_STRING_SQL_NAME:
	    retval_sptr = (char*)cou_getReturnValue( cocnx, CO_STRING, &co_err );
	    if (retval_sptr) {
		cou_logError( cocnx, CO_LOG_DEBUG, "co_js_getretAny: got String value");
		str = JS_NewStringCopyZ( cx, retval_sptr);
		*rval = STRING_TO_JSVAL(str) ;
	    } else {
		cou_logError( cocnx, CO_LOG_DEBUG, "co_js_getretAny: got NULL as String value; returning empty String");
		*rval = JS_GetEmptyStringValue( cx );
	    }
	    break;
	case CO_RECORDSET:
	    cou_logError( cocnx, CO_LOG_DEBUG, "co_js_getretAny: got Recordset value");
	    newobj = JS_ConstructObject(cx, &recordset_class, NULL, JS_GetGlobalObject(cx)); 
					  
	    if (!newobj) {
    		JS_ReportError(cx,"co_js_getretAny: could not construct Recordset object" );
		return JS_FALSE;
	    }
	    cou_logError( cocnx, CO_LOG_DEBUG, "co_js_getretAny: got Recordset object");
	    *rval = OBJECT_TO_JSVAL(newobj) ;
	    cou_logError( cocnx, CO_LOG_DEBUG, "co_js_getretAny: converted Recordset object");
	    break;
	case CO_NO_TYPE:
	    cou_logError( cocnx, CO_LOG_DEBUG, "co_js_getretAny: undefined (NO_TYPE) return value");
	    JS_ConvertValue( cx, DOUBLE_TO_JSVAL(d), JSTYPE_VOID, rval);
	    break;
	default:
    	    JS_ReportError(cx,"co_getretAny: return value type could not be returned" );
	    return JS_FALSE;
	}
	cou_logError( cocnx, CO_LOG_DEBUG, "co_js_getretAny: returns JS_TRUE");
        return JS_TRUE;
    }

static JSBool
    co_js_getretInteger(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
    {
	int* retval_ptr;
	int co_err;
	jsdouble d;

	co_context* cocnx;

	cocnx = cjs_getCOContext(cx);
	cou_logError( cocnx, CO_LOG_DEBUG, "co_js_getretInteger: called");

	d = 0;
	retval_ptr = (int*)cou_getReturnValue( cocnx, CO_INTEGER, &co_err );
	if (retval_ptr) {
	    cou_logError( cocnx, CO_LOG_DEBUG, "co_js_getretInteger: got value");
	    d = *retval_ptr;
	}
	cou_logError( cocnx, CO_LOG_DEBUG, "co_js_getretInteger: returns");
        return JS_NewNumberValue(cx, d, rval);
    }

static JSBool
    co_js_getretString(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
    {
	const char* retval_ptr;
	int co_err;
	JSString *str;

	co_context* cocnx;

	cocnx = cjs_getCOContext(cx);
	cou_logError( cocnx, CO_LOG_DEBUG, "co_js_getretString: called");

	retval_ptr = (char*)cou_getReturnValue( cocnx, CO_STRING, &co_err );
	if (retval_ptr) {
	    cou_logError( cocnx, CO_LOG_DEBUG, "co_js_getretString: got value");
	} else retval_ptr = "";
	str = JS_NewStringCopyZ( cx, retval_ptr);
	cou_logError( cocnx, CO_LOG_DEBUG, "co_js_getretString: returns");
	*rval = STRING_TO_JSVAL(str) ;
        return JS_TRUE;
    }

static JSBool
    co_js_getretField(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
    {
	int retval;
	int co_err = CO_NO_ERRORS;
	jsdouble d;
	int i,j;

	co_context* cocnx;

	cocnx = cjs_getCOContext(cx);
	cou_logError( cocnx, CO_LOG_DEBUG, "co_js_getretField: called");

	d = 0;
	JS_ValueToInt32( cx, argv[0], &i );
	JS_ValueToInt32( cx, argv[1], &j );
	//cou_logError( cocnx, CO_LOG_DEBUG, "co_js_getretField: i=%d, j=%d", i, j );
	retval = atoi(cou_getReturnValueRSvalue( cocnx, i, j, &co_err ));
	if (co_err == CO_NO_ERRORS) {
	    cou_logError( cocnx, CO_LOG_DEBUG, "co_js_getretField: got value");
	    d = retval;
	}
	cou_logError( cocnx, CO_LOG_DEBUG, "co_js_getretField: returns");
        return JS_NewNumberValue(cx, d, rval);
    }


static JSBool
    co_js_getret_RSntuples(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
    {
	int retval;
	int co_err = CO_NO_ERRORS;
	jsdouble d;

	co_context* cocnx;

	cocnx = cjs_getCOContext(cx);
	cou_logError( cocnx, CO_LOG_DEBUG, "co_js_getret_RSntuples: called");

	d = 0;
	retval = cou_getReturnValueRSntuples( cocnx, &co_err );
	if (co_err == CO_NO_ERRORS) {
	    cou_logError( cocnx, CO_LOG_DEBUG, "co_js_getret_RSntuples: got value");
	    d = retval;
	}
	cou_logError( cocnx, CO_LOG_DEBUG, "co_js_getret_RSntuples: returns");
        return JS_NewNumberValue(cx, d, rval);
    }

static JSBool
    co_js_getret_RSnfields(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
    {
	int retval;
	int co_err = CO_NO_ERRORS;
	jsdouble d;

	co_context* cocnx;

	cocnx = cjs_getCOContext(cx);
	cou_logError( cocnx, CO_LOG_DEBUG, "co_js_getret_RSnfields: called");

	d = 0;
	retval = cou_getReturnValueRSnfields( cocnx, &co_err );
	if (co_err == CO_NO_ERRORS) {
	    cou_logError( cocnx, CO_LOG_DEBUG, "co_js_getret_RSnfields: got value");
	    d = retval;
	}
	cou_logError( cocnx, CO_LOG_DEBUG, "co_js_getret_RSnfields: returns");
        return JS_NewNumberValue(cx, d, rval);
    }

static JSBool
    co_js_getret_RSvalue(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
    {
	const char* retval;
	int co_err = CO_NO_ERRORS;
	JSString *str;
	int i,j;

	co_context* cocnx;

	cocnx = cjs_getCOContext(cx);
	cou_logError( cocnx, CO_LOG_DEBUG, "co_js_getret_RSvalue: called");

	JS_ValueToInt32( cx, argv[0], &i );
	JS_ValueToInt32( cx, argv[1], &j );
	retval = cou_getReturnValueRSvalue( cocnx, i, j, &co_err );
	if (co_err == CO_NO_ERRORS) {
	    cou_logError( cocnx, CO_LOG_DEBUG, "co_js_getret_RSvalue: got value");
	    str = JS_NewStringCopyZ( cx, retval);
	}
	else {
	    str = JS_NewStringCopyZ( cx, "");
	}
	cou_logError( cocnx, CO_LOG_DEBUG, "co_js_getret_RSvalue: returns");
	*rval = STRING_TO_JSVAL(str) ;
        return JS_TRUE;
    }


    /*
     * Use a JSFunctionSpec array terminated with a null name to define a
     * bunch of native functions.
     */
JSFunctionSpec cjs_functions[] = {
    /*    name          native          nargs    */
    {"ap_write",	co_js_apwrite,         	1},
    {"co_broker",	co_js_broker,		4},
    {"co_getretAny",	co_js_getretAny,	0},
    {"co_getretInteger",co_js_getretInteger,  0},
    {"co_getretString", co_js_getretString,   0},
    {"co_getretField",	co_js_getretField,  2},
    {"co_getretRSntuples", co_js_getret_RSntuples,  0},
    {"co_getretRSnfields", co_js_getret_RSnfields,  0},
    {"co_getretRSvalue", co_js_getret_RSvalue,  2},
    {0}
};

/****************************************/
/*********** INIT ALL CO OBJECTS ********/
/****************************************/
void
cjs_InitCOClasses(JSContext *cx)
{
        if (!comethod_Init(cx)) {
            cou_error(cjs_getCOContext(cx), "JS_InitClass failed: COMethod");
        }

        if (!recordset_Init(cx)) {
            cou_error(cjs_getCOContext(cx), "JS_InitClass failed: Recordset");
        }

        if (!request_Init(cx)) {
            cou_error(cjs_getCOContext(cx), "JS_InitClass failed: Request");
        }

        if (!params_Init(cx)) {
            cou_error(cjs_getCOContext(cx), "JS_InitClass failed: Params");
        }

        if (!coserver_Init(cx)) {
            cou_error(cjs_getCOContext(cx), "JS_InitClass failed: COServer");
        }
}
