#include "httpd.h"
#include "http_config.h"
#include "http_protocol.h"
#include "http_log.h"
#include "ap_config.h"

/* PostgreSQL access functions */
#include "libpq-fe.h"

//#include "co_utils.h"
//#include "co_log.h"
#include "co_lib.h"
#include "co_defines.h"
#include "co_mod.h"
#include "../comodules/generic_acl/co_generic_acl.h"

#include "co_recordset.h"
#include "co_array.h"
#include "co_string.h"
#include "mod_coffice.h"

typedef struct
{
   PGresult * pg_result;
   int cleared;
   void * next;
} co_recordset_add;

struct co_recordset
{
   int magic;
   char * name;

   PGresult * pg_result;
   int cleared;
   int NTuples;
   int NFields;
   int type;

   co_recordset_add * next;
   co_recordset_add * last;

   co_array_of_pointers * recs;
   co_array_of_pointers * fnames;

};

//#undef __SEND_MESG_DEBUG
//#define __SEND_MESG_DEBUG(x...) 

#define FUN_COMMON(fun_name) \
  char MESGString[MESGSIZE]; \
  pool * p = cnx->request->pool; \
  __SEND_MESG_DEBUG("entering co_recordset::" # fun_name) \
  __ASSERT(cor_IsRecordset(cnx,corec), "ERROR: co_recordset::" # fun_name " trying to execute for incorrect corec") \
  __SEND_MESG_DEBUG("co_recordset::" # fun_name " for corec.name = %s", corec->name)

#define FUN_COMMON_NULL(fun_name) \
  char MESGString[MESGSIZE]; \
  pool * p = cnx->request->pool; \
  __SEND_MESG_DEBUG("entering co_recordset::" # fun_name) \
  __ASSERT_NULL(cor_IsRecordset(cnx,corec), "ERROR: co_recordset::" # fun_name " trying to execute for incorect corec") \
  __SEND_MESG_DEBUG("co_recordset::" # fun_name " for corec.name = %s", corec->name)


#define FUN_COMMON2(fun_name) \
  char MESGString[MESGSIZE]; \
  pool * p = cnx->request->pool; \
  __SEND_MESG_DEBUG("co_recordset::" # fun_name)

#define FUN_COMMON_JS(fun_name) \
  char MESGString[MESGSIZE]; \
  pool * p = cnx->request->pool; \
  __SEND_MESG_DEBUG(fun_name)

////////////////////////////////////////////////////////////////////////////////////
// CONSTRUCTORS

co_recordset * rawCreate(co_context * cnx)
{
	co_recordset * ret;
	FUN_COMMON2("rawCreate")

	ret = ap_pcalloc(cnx->request->pool, sizeof(co_recordset));
	ret->type = COREC_EMPTY;
	ret->magic = COREC_MAGIC_NUMBER;
	ret->fnames = NULL;
	
  return ret;
}

co_recordset * cor_CreateEmpty(co_context * cnx)
{
  FUN_COMMON2("cor_CreateEmpty")

  return rawCreate(cnx);
}

co_recordset * cor_CreateEmptyNamed(co_context * cnx, const char * name)
{
  co_recordset * ret;
  FUN_COMMON2("cor_CreateEmptyNamed")

  ret = rawCreate(cnx);
  ret->name = ap_pstrdup(cnx->request->pool, name);
  return ret;
}

co_recordset * cor_CreateOnPGResult(co_context * cnx, PGresult * res)
{
  co_recordset * ret;
  FUN_COMMON2("cor_CreateOnPGResult")

  __ASSERT_NULL( (res != (PGresult *)NULL) , "co_recordset::CreateOnPGResult -- res is NULL")

  ret = rawCreate(cnx);
  
  ret->type = COREC_OLD_TYPE;
  ret->pg_result = res;
  ret->NTuples = PQntuples(res);

  return ret;
}

PGresult * rawCreateResultOnSQLString(co_context * cnx, const char * SQLString)
{
  PGresult * res;
  FUN_COMMON2("rawCreateResultOnSQLString")
  __ASSERT_NULL( (SQLString!=NULL) , "co_recordset::rawCreateResultOnSQLString -- SQLString is NULL")

  res = PQexec(cnx->dbconn, SQLString);
  if (PQresultStatus(res) != PGRES_TUPLES_OK) 
  { 
    char *error_message = ap_pstrdup(cnx->request->pool,PQresultErrorMessage(res));

    cou_logError(cnx, CO_LOG_ERR, "ERROR:%s", error_message);
   // dirty_res = 0; 
    PQclear(res); 
    cou_logError(cnx, CO_LOG_ERR, "QUERY:\n%s", SQLString);
    cou_error(cnx, error_message); 
    return NULL;
  }
  return res;
}


co_recordset * cor_CreateOnSQLString(co_context * cnx, const char * SQLString)
{
	co_recordset * ret;
	PGresult * res;

	FUN_COMMON2("cor_CreateOnSQLString")
	__ASSERT_NULL( (SQLString!=NULL) , "co_recordset::cor_CreateResultOnSQLString -- SQLString is NULL")

	ret = rawCreate(cnx);
	__ASSERT_NULL(ret != NULL, "Error allocating space for co_recordset")

	ret->name = ap_pstrdup(cnx->request->pool, SQLString);

	res = rawCreateResultOnSQLString(cnx, SQLString);

	if (res == NULL) 
	{ 
		ret->type = COREC_EMPTY;
	}
	else
	{
		ret->type = COREC_OLD_TYPE;
		ret->pg_result = res;
		ret->NTuples = PQntuples(res);
	}
	return ret;
}

//////////////////////////////////////////////////////////////////////
//TEST METHODS

int cor_IsRecordset(co_context * cnx, const co_recordset * corec)
{
   if(corec==NULL) return 0;
   if(corec->magic != COREC_MAGIC_NUMBER) return 0;
   return 1;
} 

int cor_IsEmpty(co_context * cnx, const co_recordset * corec)
{
   FUN_COMMON("cor_IsEmpty")
   if(corec->type == COREC_EMPTY) return 1;
   return 0;
} 

/////////////////////////////////////////////////////////////////////
// GET METHODS

int cor_GetTupleCount(co_context * cnx, const co_recordset * corec)
{
   FUN_COMMON("cor_GetTupleCount")
   
   if(corec->type == COREC_OLD_TYPE)
   {
      return PQntuples(corec->pg_result);
   };
 
   if(corec->type == COREC_EXTENDED_TYPE)
   {
      return corec->NTuples;
   };
 
   if(corec->type == COREC_ARRAY_TYPE)
   {
      return corec->NTuples;
   };
   
  __SEND_MESG_FORCE("OPERATION ON INVALID RECORDSET")
  __SEND_MESG_DEBUG("corec->type = %d", corec->type)
  return -1;

}; 


int cor_GetFieldCount(co_context * cnx, const co_recordset * corec)
{
   FUN_COMMON("cor_GetFieldCount")

   if(corec->type == COREC_OLD_TYPE)
   {
      return PQnfields(corec->pg_result);
   };
 
   if(corec->type == COREC_EXTENDED_TYPE)
   {
      return PQnfields(corec->pg_result);
   };
 
   if(corec->type == COREC_ARRAY_TYPE)
   {
      return corec->NFields;
   };
  
  __SEND_MESG_FORCE("OPERATION ON INVALID RECORDSET")
  __SEND_MESG_DEBUG("corec->type = %d", corec->type)
  return -1;

}; 

int cor_GetFieldNum(co_context * cnx, const co_recordset * corec, const char *field_name)
{
	FUN_COMMON("cor_GetFieldNum")

	if(corec->type == COREC_ARRAY_TYPE || corec->fnames != NULL)
	{
		int i=0;
		int NFields = cor_GetFieldCount(cnx, corec);
		__ASSERT(NFields>-1, "Error getting number of fields")
		
		for(i=0; i<NFields; i++)
		{
			char * fname;  
			fname = (char *) coa_GetElement(cnx, corec->fnames, i);
			__ASSERT(fname != NULL, "HARD CORE ERROR in getting fname from co_array corec->type:%d", corec->type)

			if(strcmp(fname, field_name)==0)
			{
				return i;
			}
		}
		__SEND_MESG_DEBUG("FIELD NOT FOUND")
		return -1;
	}
 
   if(corec->type == COREC_OLD_TYPE)
   {
      return PQfnumber(corec->pg_result, field_name);
   };

   if(corec->type == COREC_EXTENDED_TYPE)
   {
      return PQfnumber(corec->pg_result, field_name);
   };

  __SEND_MESG_FORCE("OPERATION ON INVALID RECORDSET")
  return -1;

}; 

const char * cor_GetFieldName(co_context * cnx, const co_recordset * corec, int field_num)
{
   FUN_COMMON_NULL("cor_GetFieldName")

   if(corec->type == COREC_ARRAY_TYPE || corec->fnames != NULL)
   {
      int NFields = cor_GetFieldCount(cnx, corec);
      char *fname;
      if(field_num <0)         return NULL;
      if(field_num >= NFields) return NULL; 
      fname =  (char *) coa_GetElement(cnx, corec->fnames, field_num);
      return fname;
   };
 
   if(corec->type == COREC_OLD_TYPE)
   {
      return PQfname(corec->pg_result, field_num);
   };

   if(corec->type == COREC_EXTENDED_TYPE)
   {
      return PQfname(corec->pg_result, field_num);
   };

  __SEND_MESG_FORCE("OPERATION ON INVALID RECORDSET")
  return NULL;

}; 

char * rawGetValue(co_context * cnx, const co_recordset * corec, int tuple_num, int field_num)
{
   FUN_COMMON_NULL("rawGetValue")

   __SEND_MESG_DEBUG("getting val for type: %d tn: %d fn: %d", corec->type, tuple_num, field_num)

   if(corec->type == COREC_OLD_TYPE)
   {
      return PQgetvalue(corec->pg_result, tuple_num, field_num);
   }
   if(corec->type == COREC_EXTENDED_TYPE)
   {
      int prev_max = 0;
      int local_max = PQntuples(corec->pg_result);
      co_recordset_add * next = corec->next;
      co_recordset_add * prev_next = next;

      __SEND_MESG_DEBUG("getting val for EXTENDED tn: %d fn: %d", tuple_num, field_num)

      if(tuple_num < local_max) 
      {
          return PQgetvalue(corec->pg_result, tuple_num, field_num);
      }

      while(local_max <= tuple_num && next != NULL)
      {
         prev_max = local_max;
         local_max += PQntuples(next->pg_result);
         prev_next = next;
         next = next->next; 
         __SEND_MESG_DEBUG("local_max: %d prev_max: %d", local_max, prev_max)
      }
      if(prev_next != NULL)
      {
         __SEND_MESG_DEBUG("getting val tn: %d fn: %d", tuple_num-prev_max, field_num)
         return PQgetvalue(prev_next->pg_result, tuple_num-prev_max, field_num);
      }
      else
      {
         __SEND_MESG_FORCE("ERROR IN CO_RECORDSET STRUCTURE!")
         return NULL;
      }
   }

   if(corec->type == COREC_ARRAY_TYPE)
   {
      int NFields = cor_GetFieldCount(cnx, corec);
      int NTuples = cor_GetTupleCount(cnx, corec);
      
      __SEND_MESG_DEBUG("Getting value from COREC_ARRAY_TYPE")

      __ASSERT_NULL((field_num>=0),        "field_num<0 (field_num: %d) in getting value for corec: %s", field_num, corec->name)
      __ASSERT_NULL((field_num < NFields), "field_num>=NFields (field_num: %d) in getting value for corec: %s", field_num, corec->name) 
      __ASSERT_NULL((tuple_num >=0),       "tuple_num<0 (tuple_num: %d) in getting value for corec: %s", tuple_num, corec->name)
      __ASSERT_NULL((tuple_num < NTuples), "tuple_num>=NTuples (tuple_num: %d) in getting value for corec: %s", tuple_num, corec->name)

      {
        co_array_of_pointers * rec = (co_array_of_pointers *) coa_GetElement(cnx, corec->recs, tuple_num);
        return (char *) coa_GetElement(cnx, rec, field_num);
      }
   }
 
  __SEND_MESG_FORCE("OPERATION ON INVALID RECORDSET")
  return NULL;
}
 

const char * cor_GetValue(co_context * cnx, const co_recordset * corec, int tuple_num, int field_num)
{
 
  FUN_COMMON_NULL("cor_GetValue")

  return rawGetValue(cnx, corec, tuple_num, field_num);
}; 

const char * cor_GetValueByName(co_context * cnx, const co_recordset * corec, int tuple_num, const char* field_name)
{
   int field_num = 0;

   FUN_COMMON_NULL("cor_GetValueByName")

   field_num=cor_GetFieldNum(cnx, corec, field_name);
   return rawGetValue(cnx, corec, tuple_num, field_num);
}; 

char * cor_GetCopyOfValueByName(co_context * cnx, const co_recordset * corec, int tuple_num, const char* field_name)
{
   int field_num = 0;
   char * ret; 

   FUN_COMMON_NULL("cor_GetCopyOfValueByName")

   field_num=cor_GetFieldNum(cnx, corec, field_name);
   ret =  rawGetValue(cnx, corec, tuple_num, field_num);
   if(ret != NULL)
   {
     return ap_pstrdup(cnx->request->pool, ret);
   }
   return NULL;
}; 

char * cor_GetCopyOfValue(co_context * cnx, const co_recordset * corec, int tuple_num, int field_num)
{
   char * ret;

   FUN_COMMON_NULL("cor_GetCopyOfValue")

   ret =  rawGetValue(cnx, corec, tuple_num, field_num);
   if(ret != NULL)
   {
     return ap_pstrdup(cnx->request->pool, ret);
   }
   return NULL;
}; 

co_array_of_pointers * cor_GetCopyOfRecord(co_context * cnx, const co_recordset * corec, int tuple_num)
{
   co_array_of_pointers * ret;

   FUN_COMMON_NULL("cor_GetCopyOfRecord")
   __ASSERT_NULL(corec->type == COREC_ARRAY_TYPE, "Wrong corec type: %d", corec->type)

   __ASSERT_NULL(tuple_num >= 0 ,                           "in cor_GetCopyOfRecord, tuple_num is <0, tuple_num: %d", tuple_num)
   __ASSERT_NULL(tuple_num < cor_GetTupleCount(cnx, corec), "in cor_GetCopyOfRecord, tuple_num is >= NTuple, tuple_num: %d", tuple_num)

   ret =  coa_Copy(cnx, coa_GetElement(cnx, corec->recs, tuple_num));
   return ret;
}; 

co_array_of_pointers * cor_GetCopyOfColumnByName(co_context * cnx, const co_recordset * corec, const char* field_name)
{
	co_array_of_pointers * ret;
	int NTuples;
	int i;
	int FNum;
	int NElements=0;

	FUN_COMMON_NULL("cor_GetCopyOfColumnByName")
	NTuples = cor_GetTupleCount(cnx, corec);

	FNum = cor_GetFieldNum(cnx, corec, field_name);
	__ASSERT_NULL(FNum > -1, "Trying to copy column for not existing field name, field_name: %s", field_name)

	__SEND_MESG_DEBUG("Number of field: %d", FNum)

	ret = coa_CreateTyped(cnx, COARRAY_TYPE_STRING);

	__SEND_MESG_DEBUG("ret created")

	for(i=0; i<NTuples; i++)
	{
		char * val = cor_GetValue(cnx, corec, i, FNum);
		__SEND_MESG_DEBUG("we have val: %s", val);
		NElements = coa_AppendCopyOfElement(cnx, ret, val);
		__ASSERT_NULL(NElements==i+1, "HARD CORE ERROR in co_array_of_pointers");
	}

   return ret;
}; 



//////////////////////////////////////////////////////////////////////////////
// CLEAR METHODS

int cor_Clear(co_context * cnx, co_recordset * corec)
{
  FUN_COMMON("cor_Clear")
 
   corec->NTuples=0;

   if(corec->type == COREC_OLD_TYPE)
   {
      if(corec->cleared==0)
      {
         PQclear(corec->pg_result);
         corec->cleared = 1;
         return 1;
      } 
      else
      {
#ifdef CO_NEVER
         __SEND_MESG_FORCE("method: cor_Clear -- trying to clear cleared recordset")
         return -1;
#else /* i.e. always */
         return 1;
#endif      
      }
   };

   if(corec->type == COREC_EXTENDED_TYPE)
   {
      co_recordset_add * next = corec->next;
      if(corec->cleared==0)
      {
        PQclear(corec->pg_result);
        corec->cleared = 1;
      }
#ifdef CO_NEVER
      else
      {
         __SEND_MESG_FORCE("method: cor_Clear -- trying to clear cleared recordset")
         return -1;
      }
#endif
      while(next)
      {
        if(next->cleared==0)
        {
           PQclear(next->pg_result);
           next->cleared = 1;
           next = next->next;
        }
        else
        {
           __SEND_MESG_FORCE("method: cor_Clear -- trying to clear cleared recordset")
           return -1;
        }
      }
      return 1;
   };
  
   if(corec->type == COREC_ARRAY_TYPE)
   {
     return 1;
   }
 
  __SEND_MESG_FORCE("OPERATION ON INVALID RECORDSET")
  return -1;

}; 

int cor_IsCleared(co_context * cnx, const co_recordset * corec)
{
	return corec->cleared;
}
///////////////////////////////////////////////////////////////////////////////////////
// SET METHODS

int  cor_AppendRowset(co_context * cnx, co_recordset * corec, PGresult * res)
{
  co_recordset_add * new;

  FUN_COMMON("cor_AppendRowset")
  __ASSERT( (res!=NULL) , "in cor_AppendRowset res is NULL")

  __ASSERT(((corec->type == COREC_EXTENDED_TYPE) || (corec->type == COREC_OLD_TYPE) || (corec->type == COREC_EMPTY)),
           "in cor_AppendRowset wrong corec type, type: %d", corec->type)

  new = ap_pcalloc(cnx->request->pool, sizeof(co_recordset_add));

  new->pg_result = res;
  if(corec->last != NULL) corec->last->next = new;
  if(corec->next == NULL) corec->next = new;
  corec->last = new;
  corec->NTuples += PQntuples(res);
  corec->type = COREC_EXTENDED_TYPE;
  return 1;
}

int  cor_AppendOnSQLString(co_context * cnx, co_recordset * corec, const char * SQLString)
{
  PGresult * res;
  FUN_COMMON("cor_AppendOnSQLString")
  __ASSERT( (SQLString!=NULL) , "in cor_AppendOnSQLString SQLString is NULL")

  __ASSERT(corec->type == COREC_EXTENDED_TYPE || corec->type == COREC_OLD_TYPE || corec->type == COREC_EMPTY,
           "in cor_AppendOnSQLString wrong corec type, type: %d", corec->type)

  res = rawCreateResultOnSQLString(cnx, SQLString);
  __SEND_MESG_DEBUG("We have results");
  if(res != NULL)
  {
    return cor_AppendRowset(cnx, corec, res); 
  }
  return -1;
}

int  cor_AppendRecord(co_context * cnx, co_recordset * corec, const co_recordset * source, int tuple_num)
{
  int FCount;

  FUN_COMMON("cor_AppendRecord")
  __ASSERT( (cor_IsRecordset(cnx, source)) ,"in cor_AppendRecord source is not recodset")

  __ASSERT( ((corec->type == COREC_ARRAY_TYPE) || (corec->type == COREC_EMPTY)),
           "in cor_AppendRecord wrong corec type, type: %d", corec->type)

   FCount =  cor_GetFieldCount(cnx, source);

  if(corec->type == COREC_EMPTY)
  {
    int i = 0;
    __SEND_MESG_DEBUG("Appending record to recordset -- first time")

    corec->recs = coa_CreateTyped(cnx, COARRAY_TYPE_OBJECT);
    corec->fnames = coa_Create(cnx);
    for(i=0; i<FCount; i++)
    {
       coa_AppendElement(cnx, corec->fnames, ap_pstrdup(cnx->request->pool, cor_GetFieldName(cnx, source, i)));
    }
    corec->NFields = FCount;
    corec->type = COREC_ARRAY_TYPE;
  }

  if(corec->type == COREC_ARRAY_TYPE)
  {  
    __SEND_MESG_DEBUG("Appending record to recordset -- adding record")

    __ASSERT( (corec->NFields == FCount) ,"WRONG NUMBER OF FIELDS IN cor_AppendRecord")

    {
      int i=0;
      co_array_of_pointers * new_rec = coa_Create(cnx); 
      int last_num = coa_AppendElement(cnx, corec->recs, new_rec);
      for(i=0; i<FCount; i++)
      {
         coa_AppendElement(cnx, new_rec, cor_GetCopyOfValue(cnx, source, tuple_num, i));
      }
      corec->NTuples++;
      return 1;
    }
 }
  return -1;
}

int  cor_AppendRecordset(co_context * cnx, co_recordset * corec, const co_recordset * source)
{
  int RCount;
  int i;

  FUN_COMMON("cor_AppendRecordset")
  __ASSERT( (cor_IsRecordset(cnx, source)) ,"in cor_AppendRecordset source is not recodset")

  __ASSERT( ((corec->type == COREC_ARRAY_TYPE) || (corec->type == COREC_EMPTY)),
           "in cor_AppendRecordset wrong corec type, type: %d", corec->type)

  RCount =  cor_GetTupleCount(cnx, source);
  for(i=0; i<RCount; i++)
  {
    if (cor_AppendRecord(cnx, corec, source, i)<0) return -1;
  }
  return 1;
}

int  cor_SetFieldName(co_context * cnx, co_recordset * corec, int field_num, char * new_name)
{
  int FCount;

  FUN_COMMON("cor_setFieldName")
  __ASSERT( new_name != NULL, "cor_setFieldName: new_name is NULL")

  __ASSERT( corec->type != COREC_EMPTY,
           "in cor_setFieldName wrong corec type, type: %d", corec->type)

   FCount =  cor_GetFieldCount(cnx, corec);

  __ASSERT( (field_num>=0) && (field_num<FCount),
           "in cor_setFieldName wrong field_num: %d", field_num)


  if((corec->type == COREC_OLD_TYPE || corec->type == COREC_EXTENDED_TYPE) && corec->fnames==NULL)
  {
    int i=0;
    corec->fnames = coa_Create(cnx);
    for(i=0; i<FCount; i++)
    {
       coa_AppendElement(cnx, corec->fnames, ap_pstrdup(p, PQfname(corec->pg_result, i)));
    }
    corec->NFields = FCount;
    coa_SetElement(cnx, corec->fnames, field_num, new_name);
    return 1;
  }

  if(corec->type == COREC_ARRAY_TYPE)
  {  
     coa_SetElement(cnx, corec->fnames, field_num, new_name);
     return 1;
   }
 
  return -1;
}





///////////////////////////////////////////////////////////////////////////////////////
// JS creation from co_recordset

int
cou_createJSHeadFromRecordset
(
    co_context *cnx,
    const co_recordset *corec
)
{
    int i,j;
    int nFields=0;
    const char *js_out=0;
 
    FUN_COMMON_JS("cou_createJSHeadFromRecordset") 
    __ASSERT( (cor_IsRecordset(cnx, corec)) , "in cou_createJSHeadFromRecordset corec is not recordset")

    nFields = cor_GetFieldCount(cnx, corec);

    js_out = ap_psprintf(p, "var o = new Object();\no.d = new Object();\n");

    for (i = 0; i < nFields; i++)
    {
	js_out=ap_pstrcat(p, js_out, "o.d[\"",cor_GetFieldName(cnx, corec, i), "\"] = new Array();\n", NULL);
    };
 
    cou_jsputs(cnx,js_out);
    return 1;

};


int
cou_createJSBodyFromRecordset
(
    co_context *cnx,
    const co_recordset *corec
)
{
    int i,j;
    int nFields=0;
    int nTuples=0;
    const char *js_out="";
    char ** field_names;
    char buf[JSBUFSIZE] = "";
    int req_size =0;
    int old_size = 0;
    
    FUN_COMMON_JS("cou_createJSBodyFromRecordset") 
    __ASSERT(cor_IsRecordset(cnx, corec), "in cou_createJSBodyFromRecordset corec is not recordset")

  
    nFields = cor_GetFieldCount(cnx, corec);
    field_names = ap_pcalloc(p, sizeof(char *) * nFields);
    nTuples = cor_GetTupleCount(cnx, corec);

   for (j = 0; j < nFields; j++)
   {
       field_names[j] = cor_GetFieldName(cnx, corec, j); 
   }

    for (i = 0; i < nTuples; i++)
    {
        for (j = 0; j < nFields; j++)
	{
            old_size =  strlen(buf); 
            req_size = snprintf(buf + strlen(buf), JSBUFSIZE-old_size, 
                      "o.d[\"%s\"][%d] = \"%s\";\n", field_names[j], i, cou_buildEncodedJS(cnx,cor_GetValue(cnx, corec, i, j)));

            // here we check whether our buffer was big enough
            if(req_size>=JSBUFSIZE-old_size)
            {
               buf[old_size]=0;
               cou_jsputs(cnx, ap_pstrdup(p, buf));
               snprintf(buf, JSBUFSIZE, 
                      "o.d.%s[%d] = \"%s\";\n", field_names[j], i, cou_buildEncodedJS(cnx,cor_GetValue(cnx, corec, i, j)));
            }
	}
    }
    cou_jsputs(cnx, ap_pstrdup(p, buf));
    return 1;
};

int
cou_createJSCallback
(
    co_context *cnx,
    const char *callback
)
{
    const char *js_out="";

    FUN_COMMON_JS("cou_createJSCallback") 
  
    cou_jsputs(cnx, ap_pstrcat(p, callback, "o);\n", NULL));
    return 1;

};

int
cou_createJSFromRecordset
(
    co_context *cnx,
    const co_recordset *corec,
    const char *callback
)
{
    const char * js_out = "";

    FUN_COMMON_JS("cou_createJSFromRecordset") 

    __ASSERT(cor_IsRecordset(cnx, corec), "in cou_createJSFromRecordset corec is not recordset")
  
     cou_createJSHeadFromRecordset(cnx, corec);
     cou_createJSBodyFromRecordset(cnx, corec);
     cou_createJSCallback(cnx, callback);
     return 1;
}


/********************************************/
/* Linux To Windows Recordset Serialization */
/********************************************/

#include "co_string.h"

#define XML_INT "<s:datatype dt:type='int' dt:maxLength='4' rs:precision='10' rs:fixedlength='true' rs:maybenull='false'/>"
#define XML_STR "<s:datatype dt:type='string' dt:maxLength='8000' rs:maybenull='false'/>"

char *
cor_GetADOXMLSerialization(co_context *cnx, const co_recordset *corec)
{
	int nFields, nTuples;
	int tpl, fld;
	co_string *costr;
	nFields = cor_GetFieldCount(cnx, corec);
	nTuples = cor_GetTupleCount(cnx, corec);
	costr = cos_CreateEmpty(cnx);
	cos_AppendString(cnx, costr, 
"<xml xmlns:s='uuid:BDC6E3F0-6DA3-11d1-A2A3-00AA00C14882'
        xmlns:dt='uuid:C2F41010-65B3-11d1-A29F-00AA00C14882'
        xmlns:rs='urn:schemas-microsoft-com:rowset'
        xmlns:z='#RowsetSchema'>
<s:Schema id='RowsetSchema'>");
	cos_AppendString(cnx, costr, "
        <s:ElementType name='row' content='eltOnly'>");
	for (fld=0; fld<nFields; fld++) {
		cos_AppendFormatted(cnx, costr, "
                <s:attribute type='%s'/>", cor_GetFieldName( cnx, corec, fld ) );
	}
	cos_AppendString(cnx, costr, "
                <s:extends type='rs:rowbase'/>
        </s:ElementType>");
	for (fld=0; fld<nFields; fld++) {
		cos_AppendFormatted(cnx, costr, "
        <s:AttributeType name='%s' rs:number='%d' rs:writeunknown='true'>
                %s
        </s:AttributeType>", cor_GetFieldName( cnx, corec, fld ), fld+1,
		    (PQftype(corec->pg_result, fld)==23 ? XML_INT : XML_STR) );
	}
	cos_AppendString(cnx, costr, "
</s:Schema>
<rs:data>");
	for (tpl=0; tpl<nTuples; tpl++) {
		cos_AppendString(cnx, costr, "
        <z:row");
		for (fld=0; fld<nFields; fld++) {
			cos_AppendFormatted(cnx, costr, " %s='%s'", 
			    cor_GetFieldName( cnx, corec, fld ),
			    cor_GetValue( cnx, corec, tpl, fld ) );
		}
		cos_AppendString(cnx, costr, "/>");
	}
	cos_AppendString(cnx, costr, "
</rs:data>
</xml>"
	);
	return cos_GetSimplifiedString(cnx, costr);
}
