#include "co_lib.h"
#undef JMP_BUF
#include "libpq-fe.h"
#include "libpq-int.h"
#include "co_recordset.h"

void cox_clearConnectionResult( co_context *ctx )
{
    ctx->dbconn->result = NULL;
}

void *
cou_getReturnValue( co_context *ctx,
        co_param_type result_type,
        int *err
    )
{
    cou_logError( ctx, CO_LOG_DEBUG, 
            "cou_getReturnValue: called" );

    if (ctx->result_type != result_type) {
	*err = CO_ERR_TYPE_MISMATCH;
	return NULL;
    }
    
    //cou_logError( ctx, CO_LOG_DEBUG, "result: %s", ctx->result);

    return ctx->result;
}

int
cou_getReturnValueField( co_context *ctx,
        co_param_type result_type, int i, int j,
        int *err
    )
{
    cou_logError( ctx, CO_LOG_DEBUG, 
            "cou_getReturnValue: called" );

    if (ctx->result_type != result_type) {
	*err = CO_ERR_TYPE_MISMATCH;
	return 0;
    }

    if (ctx->result_type != CO_RECORDSET) {
	*err = CO_ERR_TYPE_MISMATCH;
	return 0;
    }
/*    
    return ctx->result;
*/
    return atoi(cor_GetValue(ctx,ctx->result, i, j));
}


void
cou_setReturnValue( co_context *ctx,
	void *result_value,
        co_param_type result_type
    )
{

    cou_logError( ctx, CO_LOG_DEBUG, 
            "cou_setReturnValue: called" );


    if ( ctx->result_type == CO_RECORDSET )
    { 
	 if(ctx->result != result_value)
         {
            if (!cor_IsCleared(ctx,ctx->result)) cor_Clear(ctx,ctx->result);
         }
    }

    cou_logError( ctx, CO_LOG_DEBUG, 
            "cou_setReturnValue: PQ checked" );

    if (result_type == CO_INTEGER) {
	/* allocate space, copy value */
	ctx->result = ap_palloc( ctx->request->pool, sizeof(int) );
	cou_logError( ctx, CO_LOG_DEBUG, 
            "cou_setReturnValue: sizeof(int) allocated" );

	*(int*)(ctx->result) = *(int*)result_value;
	cou_logError( ctx, CO_LOG_DEBUG, 
            "cou_setReturnValue: int value copied" );
    }
    if ((result_type == CO_RECORDSET)||(result_type == CO_STRING)) {
	/* just store pointer */
	ctx->result = result_value;
    }
    ctx->result_type = result_type;
    cou_logError( ctx, CO_LOG_DEBUG, 
            "cou_setReturnValue: type set" );

    if (result_type == CO_INTEGER) {
	cou_logError( ctx, CO_LOG_DEBUG, "result: CO_INTEGER %d", *(int*)(ctx->result));
    }
    else if (result_type == CO_STRING) {
	cou_logError( ctx, CO_LOG_DEBUG, "result: CO_STRING %s", (const char*)ctx->result);
    }
    else if (result_type == CO_RECORDSET) {
	cou_logError( ctx, CO_LOG_DEBUG, "result: CO_RECORDSET");
    }
}


/* ------------------------------------------------ */
/* interface to recordset */

int
cou_getReturnValueRSntuples( co_context *ctx,
        int *err
    )
{
    cou_logError( ctx, CO_LOG_DEBUG, 
            "cou_getReturnValueRSntuples: called" );

    if (ctx->result_type != CO_RECORDSET) {
	*err = CO_ERR_TYPE_MISMATCH;
	return 0;
    }
    return cor_GetTupleCount(ctx,ctx->result);
}

int
cou_getReturnValueRSnfields( co_context *ctx,
        int *err
    )
{
    cou_logError( ctx, CO_LOG_DEBUG, 
            "cou_getReturnValueRSnfields: called" );

    if (ctx->result_type != CO_RECORDSET) {
	*err = CO_ERR_TYPE_MISMATCH;
	return 0;
    }
    return cor_GetFieldCount(ctx,ctx->result);
}

const char*
cou_getReturnValueRSvalue( co_context *ctx,
        int i, int j,
        int *err
    )
{
    cou_logError( ctx, CO_LOG_DEBUG, 
            "cou_getReturnValueRSvalue: called" );

    if (ctx->result_type != CO_RECORDSET) {
	*err = CO_ERR_TYPE_MISMATCH;
	return 0;
    }

    return cor_GetValue(ctx,ctx->result, i, j);
}
