#include <stdio.h>
#include <string.h>

#include "httpd.h"
#include "http_protocol.h"

#if 0
#include "co_context.h"
#include "co_log.h"     /* logError and constants */
#include "co_utils.h"   /* sendMessage  etc       */
#endif
#include "co_lib.h"     /* all above and cou_error*/
#include "co_defines.h" /* macros                 */

#include "co_string.h"

struct co_string_elem;

struct co_string
{
   int magic;
   size_t length; /* length of the complete str; if length==0, do not try to read str */
   struct co_string_elem* first; /* NULL <=> it is empty */
   struct co_string_elem* last; /* NULL <=> it is empty */
};

typedef struct co_string_elem
{
   int magic;
   const char* str; /* assumption: data pointed by str is never modified outside the co_string API */
   size_t length; /* length of the str; never 0 (we do not create elem for empty string) */
   struct co_string* parent;
   struct co_string_elem* next; /* NULL <=> it is the last element */
} co_string_elem;

co_string *
cos_CreateEmpty(co_context * cnx)
{
  co_string *costr;
  costr = ap_pcalloc(cnx->request->pool, sizeof(co_string));
  costr->magic = COSTR_MAGIC_NUMBER;
  costr->length = 0;
  costr->first = NULL;
  costr->last = NULL;
  return costr;
}

int
cos_AppendString(co_context * cnx, co_string * costr, const char * str)
{
  size_t length;
  char MESGString[MESGSIZE];
  __ASSERT(str, "Appending NULL to co_string")
  length = strlen(str);
  if (length==0) return 0; /* nothing to do */
  if(costr->last) {
    __ASSERT(costr->first, "co_string has last but not first")
    __ASSERT(costr->last->parent == costr, "Last part of co_string does not belong to this co_string")
    __ASSERT(costr->last->next == NULL, "Last part of co_string is not last")
    costr->last->next = ap_pcalloc(cnx->request->pool, sizeof(co_string_elem));;
    costr->last = costr->last->next;
    costr->last->str = str;
    costr->last->length = length;
    costr->last->next = NULL;
  }
  else {
    __ASSERT(costr->first == NULL, "co_string has first but not last")
    __ASSERT(costr->length == 0, "co_string has no parts but length != 0")
    costr->first = ap_pcalloc(cnx->request->pool, sizeof(co_string_elem));
    costr->first->str = str;
    costr->first->length = length;
    costr->first->next = NULL;
    costr->last = costr->first; /* it is the only one element */
  }
  costr->length += length;
  costr->last->parent = costr;
  costr->last->magic = COSTRELM_MAGIC_NUMBER;
  return 0;
}

#define cos_AppendFormatted(cnx, costr, format, args... ) \
    cos_AppendString(cnx, costr, ap_psprintf(cnx->request->pool, format, ## args ) )


char *
cos_GetSimplifiedString(co_context * cnx, const co_string * costr)
{
  char *ret;
  char *ptr;
  co_string_elem* elem;
  char MESGString[MESGSIZE];
  __ASSERT_NULL(costr, "NULL passed as co_string")
  /* allocate area for complete string and ending NULL */
  ret = ap_pcalloc(cnx->request->pool, (costr->length+1)*sizeof(char));
  if (costr->length==0) { /* return empty string */
    *ret = '\0';
    return ret;
  }
  elem = costr->first;
  ptr = ret;
  while (elem) {
    ptr += snprintf( ptr, elem->length+1, "%s", elem->str);
    cou_logError(cnx, CO_DEBUG, "elem->str='%s'", elem->str);
    cou_logError(cnx, CO_DEBUG, "full str='%s'", ret);
    elem = elem->next;
  }
  __ASSERT_NULL((ptr-ret)==costr->length, "co_string length is different than sum of elems lengths")
  *ptr = '\0';
  return ret;
}

int
cos_SendString(co_context * cnx, const co_string * costr)
{
  co_string_elem* elem;
  request_rec *r;
  static char MESGString[MESGSIZE];
  __ASSERT(costr, "NULL passed as co_string")
  if (costr->length==0) { /* nothing to do */
    return 0;
  }
  r = cnx->request;
  elem = costr->first;
  while (elem) {
    ap_rputs( elem->str, r );
    elem = elem->next;
  }
  return 0;
}