#include "httpd.h"
#include "http_protocol.h"
#include "http_log.h"

#include <string.h>
#include <stdlib.h>

#include "libpq-fe.h"
#include "../include/co_utils.h"
#include "co_convutils.h"
#include "co_log.h"
#include "co_mod.h"
#include "co_recordset.h"
#include "co_array.h"
#include "co_string.h"
#include "co_output.h"
#include "co_defines.h"

#include "mod_coffice.h"


void *
co_str2str(request_rec *r, char* str, int* err)
{
    return (void*)str;
}

void *
co_str2int(request_rec *r, char* str, int* err)
{
    int *int_ptr;
    int_ptr = ap_palloc( r->pool, sizeof(int) );
    *int_ptr = atoi(str);
    return (void*)int_ptr;    
}

void *
co_str2SQLValue(request_rec *r, char* str, int* err)
{
    // we assume, that output length is <= input
    char *return_val = ap_palloc( r->pool, 1+strlen(str)*2 );
    memset(return_val, 0, 1+strlen(str)*2);	
    cou_SQLEncode(str,return_val);
    return (void*)return_val;
}


void *
co_str2SQLName(request_rec *r, char* str, int* err)
{
    /* SQL names can not be larger than 50 characters*/
    if(strlen(str)>50) *err = CO_GET_SQL_NAME_FAILED;
    return (void*)str;
}

void *
co_ptr2strArr(request_rec *r, char* str, int* err)
{
    return (void*)str;
}


/*
// converts string to HTML
void *
co_str2HTML(request_rec *r, char* str, int* err)
{
    char *return_val = ap_palloc( r->pool, 1+strlen(str)*2 );
    memset(return_val, 0, 1+strlen(str)*2);	
    cou_HTMLEncode(str,return_val);
    return (void*)return_val;
}
*/

typedef void * (*co_converter)(request_rec *r, char*, int*);

static 
co_converter
co_converters[] =
{
    /* converter for every type enumerated in co_param_type */
    co_str2str,
    co_str2int,
    0,  // this is CO_RECORDSET for later something :-)
    co_str2SQLValue,
    co_str2SQLName,
    co_ptr2strArr
};
/*
co_converter
cou_getConverter(co_param_type t)
{
   return co_converters[t];
}
*/
//ap_log_error(APLOG_MARK, APLOG_DEBUG, cmd->server, "cmd_comodulepath: ...");
void *
cou_getParam( co_context *cnx,
	const char *param_name,
	const co_param_type param_type,
	int *err
    )
{
    request_rec *r;
    size_t name_len;
    const char* in;
    const char *result_str;
    char *result_buf;
    
    r = cnx->request;
	cou_logError( cnx, CO_LOG_DEBUG, "cou_getParam: called for '%s'", param_name );

	if(cnx->method_args == NULL)
	{
		cou_logError( cnx, CO_LOG_DEBUG, "cou_getParam: no parameters in (args)" );
		*err = CO_NO_PARAMETERS_FOUND;
		return NULL;
	}

    in = cnx->method_args;
	cou_logError( cnx, CO_LOG_DEBUG, "cou_getParam: in (args)=%s", in );

    while ((*in) && (*in != '&')) in++;
    if (*in=='\0') 
	{
		/* no parameters */
		cou_logError( cnx, CO_LOG_DEBUG, "cou_getParam: no parameters in (args)" );
		*err = CO_NO_PARAMETERS_FOUND;
		return NULL;
    }

    /* in points to the first '&' */
    *err = CO_PARAMETER_NOT_FOUND;
    name_len = strlen(param_name);
    result_str = strstr( in, ap_pstrcat( r->pool, "&", param_name, "=", NULL ) );
    /* check if found */
	if (!result_str) {
		cou_logError( cnx, CO_LOG_DEBUG, "cou_getParam: Parameter '%s' not found using GET", param_name );

		/* try with POST */
		if ((r->method_number == M_POST)&&(cnx->post_args)) {
	    	if ((strncmp(cnx->post_args, param_name, name_len )==0)&&(*(cnx->post_args+name_len)=='=')) {
				/* begins with our param, skip name and '=' */
				result_str = cnx->post_args+name_len+1;
	 		}
			else {
				/* hidden in the middle */
				result_str = strstr( cnx->post_args, ap_pstrcat( r->pool, "&", param_name, "=", NULL ) );
				if (result_str) result_str += name_len +2; /* skip param name and seperators */
			}
		}
		if (!result_str) 
			return NULL;	
	}
	else {
		result_str += name_len +2; /* skip param name and seperators */
    }

    in = result_str;
    /* find next separating '&' */
    while ((*in) && (*in != '&')) in++;
		    //strncpy( result_buf, result_str, in - result_str );
    result_buf = ap_pstrndup( r->pool, result_str, in - result_str);
//    if (r->method_number == M_POST) 
	cou_decode_string_post(result_buf);
//    else
//	cou_decode_string(result_buf);

	cou_logError( cnx, CO_LOG_DEBUG, "cou_getParam: decoded result='%s'", result_buf );

    /* on our side everything is correct; converters can set their err codes */
    *err = CO_NO_ERRORS;

    return co_converters[param_type](r, result_buf, err);
};


co_array_of_pointers *
cou_getParamArray( co_context *cnx,
	const char *param_name,
	const char *delim,
	int intTabOutLenMax,
 	int *err
    )
{
    request_rec *r;
    size_t name_len;
    const char* in;
    const char *result_str;
    co_array_of_pointers *result_arr;
/*    char **scan_result_arr; */
    char *result_buf;
    int i;
    char MESGString[MESGSIZE];

    r = cnx->request;
	cou_logError( cnx, CO_LOG_DEBUG, "cou_getParamArray: called for '%s'", param_name );

//    in = r->args;
    in = cnx->method_args;
	cou_logError( cnx, CO_LOG_DEBUG, "cou_getParamArray input: method_args='%s'", in );
//    ap_log_error(APLOG_MARK, APLOG_DEBUG, r->server, 
//	    ap_pstrcat( r->pool, "cou_getParam: in (args)=", in, NULL )
//	);
    while ((*in) && (*in != '&')) in++;
    if (*in=='\0') {
	/* no parameters */
	*err = CO_NO_PARAMETERS_FOUND;
	return NULL ;
    }

    /* in points to the first '&' */
    *err = CO_PARAMETER_NOT_FOUND;
    result_str = strstr( in, ap_pstrcat( r->pool, "&", param_name, "=", NULL ) );
    /* check if found */
	if (!result_str) {
		cou_logError( cnx, CO_LOG_DEBUG, "cou_getParamArray: Parameter '%s' not found", param_name);
		return NULL;	
	}

    name_len = strlen(param_name);
    /* skip param name and seperators */
    result_str = result_str + name_len +2;
    in = result_str;
    /* find next separating '&' */
    while ((*in) && (*in != '&')) in++;
		    //strncpy( result_buf, result_str, in - result_str );
    result_buf = ap_pstrndup( r->pool, result_str, in - result_str);

//    result_arr = ap_palloc( r->pool, intTabOutLenMax * sizeof(char*) );
    result_arr = coa_Create(cnx);
    intTabOutLenMax = coa_AppendSplit(cnx, result_arr, result_buf, "=");

//    intTabOutLenMax = cou_split( result_buf, result_arr, delim, intTabOutLenMax );
    if (intTabOutLenMax<0) {
	*err = CO_BUFFER_TOO_SMALL;
	return NULL; /* buffer too small */
    }

    for( i=0; i< intTabOutLenMax; i++)
    {
		cou_logError( cnx, CO_LOG_DEBUG, "cou_getParamArray: result[%d]='%s'", i, (char *) coa_GetElement(cnx, result_arr, i) );
    }

    for( i=0; i< intTabOutLenMax; i++)
    {
		cou_decode_string( (char *) coa_GetElement(cnx, result_arr, i));
		cou_logError( cnx, CO_LOG_DEBUG, "cou_getParamArray: decoded result[%d]='%s'", i, (char *) coa_GetElement(cnx, result_arr, i) );
    }

	//cou_logError( cnx, CO_LOG_DEBUG, "cou_getParamArray: decoded result[0]='%s'", (char *) coa_GetElement(cnx, result_arr, 0) );

    /* on our side everything is correct; converters can set their err codes */
    *err = CO_NO_ERRORS;

    return result_arr;
};
#if 0
/* Following 2 functions write to buffer of size n; */
/*   and ensure that last char is '\0' */
/*   return 1 if encoded string was cut (buffer too small) */
/* this function changes all occurences of " to \" */
int
cou_encodeJS
(
    char *dest,
    size_t n,
    const char *source
)
{
    /* find last address where we can write non-zero value */
    char* limit = dest + n - 2;

    while ((*source) && (dest<limit)) {
	if (*source=='"') { *dest++='\\'; *dest='"'; }
	else *dest=*source;
	dest++; source++;
    }
    if (dest>limit) { *limit = '\0'; return 1; } /* possible when last read was '"' */
    if (*--source) { *limit = '\0'; return 1; } /* did not read to end of source */
    return 0; /* ok */
}

/*this function changes all occurences of ' to '' */
int
cou_encodeSQL
(
    char *dest,
    size_t n,
    const char *source
)
{
    char* limit = dest + n - 2;

    while ((*source) && (dest<limit)) {
	if (*source=='\'') { *dest++='\''; *dest='\''; }
	if (*source=='\\') { *dest++='\\'; *dest='\\'; }
	else *dest=*source;
	dest++; source++;
    }
    if (dest>limit) { *limit = '\0'; return 1; }
    if (*--source) { *limit = '\0'; return 1; }
    return 0;
}

/*this function changes all occurences of ' to &#39; */
int
cou_encodeHTMLApostrophe
(
    char *dest,
    size_t n,
    const char *source
)
{
    char* limit = dest + n - 2;

    while ((*source) && (dest<limit)) {
	if (*source=='\'') { *dest++='&'; *dest++='#'; *dest++='3'; *dest++='9'; *dest=';'; }
	else *dest=*source;
	dest++; source++;
    }
    if (dest>limit) { *limit = '\0'; return 1; }
    if (*--source) { *limit = '\0'; return 1; }
    return 0;
}
#endif

/* cou_buildEncoded* functions allocate memory for encoded strings */
const char*
cou_buildEncodedHTMLForScript
(
    co_context *cnx,
    const char *source
)
{
    size_t len=0;
    const char *src_ptr;
    char *dst_ptr;
    char c;
    char *dest;
    /* find length of the destination buffer, */
    /* including "'" replaced with "&#39;" */
    src_ptr = ( source = ap_escape_html( cnx->request->pool, source ) );
    while( (c = *src_ptr++) ){ len++; if (c=='\'') len +=4; }
    len++; /* ending NULL */
    dest = (char*)ap_palloc( cnx->request->pool, len );

    dst_ptr = dest;
    while ( (c = *source++) ) {
	if (c=='\'') { *dst_ptr++='&';*dst_ptr++='#';*dst_ptr++='3';*dst_ptr++='9';*dst_ptr++=';'; }
	else *dst_ptr++ = c;
    }
    *dst_ptr = '\0';
    return dest;
}

const char*
cou_buildEncodedSQL
(
    co_context *cnx,
    const char *source
)
{
    size_t len=0;
    const char *src_ptr;
    char *dst_ptr;
    char c;
    char *dest;
	if (!source) return "";
    /* find length of the destination buffer, */
    /* including doubled (or escaped) "'" and escaped "\" */
    src_ptr = source;
    while( (c = *src_ptr++) ){ len++; if ((c=='\'')||(c=='\\')) len++; }
    len++; /* ending NULL */
    dest = (char*)ap_palloc( cnx->request->pool, len );

    dst_ptr = dest;
    while( (c = *source++) ){
	if ((c=='\'')||(c=='\\')) *dst_ptr++='\\';
	*dst_ptr++ = c;
    }
    *dst_ptr = '\0';
    return dest;
}

/* before double quote mark insert two backslashes */
/* used for building arrays in SQL queries */
const char*
cou_buildEncodedSQLArrayField
(
    co_context *cnx,
    const char *source
)
{
    size_t len=0;
    const char *src_ptr;
    char *dst_ptr;
    char c;
    char *dest;
    /* find length of the destination buffer, including doubled "'" */
    src_ptr = source;
    while( (c = *src_ptr++) ){ 
	len++; 
	if (c=='\\') len += 3; /* escape backslash, but then again escape each
	    backslash, so we have... 4 backslashes :) */
	else if (c=='"') len += 2; 
	else if (c=='\'') len++; 
    }
    len++; /* ending NULL */
    dest = (char*)ap_palloc( cnx->request->pool, len );

    dst_ptr = dest;
    while( (c = *source++) ){
	if (c=='\\') { *dst_ptr++='\\';*dst_ptr++='\\';*dst_ptr++='\\'; }
	else if (c=='"') { *dst_ptr++='\\';*dst_ptr++='\\'; }
	else if (c=='\'') *dst_ptr++='\\';
	*dst_ptr++ = c;
    }
    *dst_ptr = '\0';
    return dest;
}

/* cou_buildEncodedJS - so far just copy of cou_buildEncodedSQL */
/* escapes "'", "\" and special chars (\n, \r) */
const char*
cou_buildEncodedJS
(
    co_context *cnx,
    const char *source
)
{
    size_t len=0;
    const char *src_ptr;
    char *dst_ptr;
    char c;
    char *dest;
    /* find length of the destination buffer, */
    /* including escaped "'" and escaped "\" */
    src_ptr = source;
    while( (c = *src_ptr++) ){ len++; if ((c=='\'')||(c=='\"')||(c=='\\')||c=='\n'||c=='\r') len++; } //"
    len++; /* ending NULL */
    dest = (char*)ap_palloc( cnx->request->pool, len );

    dst_ptr = dest;
    while( (c = *source++) ){
	if ((c=='\'')||(c=='\"')||(c=='\\')) { *dst_ptr++='\\'; *dst_ptr++ = c; } //"
	else if (c=='\n') { *dst_ptr++='\\'; *dst_ptr++='n'; }
	else if (c=='\r') { *dst_ptr++='\\'; *dst_ptr++='r'; }
	else *dst_ptr++ = c;
    }
    *dst_ptr = '\0';
    return dest;
}

// replaces \n with <BR>, will be doing a bit more
const char*
cou_buildHTMLEncodedTextDocument
(
    co_context *cnx,
    const char *source
)
{
    size_t len=0;
    const char *src_ptr;
    char *dst_ptr;
    char c;
    char *dest;
    /* find length of the destination buffer, */
    /* including escaped "'" and escaped "\" */
    src_ptr = source;
    while( (c = *src_ptr++) ){ 
	len++;
	if (c=='\n') { len +=3;} 
	else if (c=='\t')  { len +=47;} 
	else if (c==' ')  { len +=5;} 
    }
    len++; /* ending NULL */
    dest = (char*)ap_palloc( cnx->request->pool, len );

    dst_ptr = dest;
    while( (c = *source++) ){
	if (c=='\n') { *dst_ptr++='<'; *dst_ptr++='B'; *dst_ptr++='R'; *dst_ptr++='>'; }
	else if (c=='\t') { strcpy(dst_ptr, "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"); dst_ptr += 48; }
	else if (c==' ') { *dst_ptr++='&'; *dst_ptr++='n'; *dst_ptr++='b'; *dst_ptr++='s'; *dst_ptr++='p'; *dst_ptr++=';'; }
	else *dst_ptr++ = c;
    }
    *dst_ptr = '\0';
    return dest;
}

const char*
cou_buildTEXTAREAEncodedTextDocument
(
    co_context *cnx,
    const char *source
)
{
    size_t len=0;
    const char *src_ptr;
    char *dst_ptr;
    char c;
    char *dest;
    /* find length of the destination buffer, */
    /* including escaped "'" and escaped "\" */
    src_ptr = source;
    while( (c = *src_ptr++) ){ 
		len++;
		if (c=='<') { len +=3;} // &lt;
		else if (c=='&')  { len +=4;} // &amp;
    }
    len++; /* ending NULL */
    dest = (char*)ap_palloc( cnx->request->pool, len );

    dst_ptr = dest;
    while( (c = *source++) ){
		if (c=='<') { *dst_ptr++='&'; *dst_ptr++='l'; *dst_ptr++='t'; *dst_ptr++=';'; }
		else if (c=='&') { *dst_ptr++='&'; *dst_ptr++='a'; *dst_ptr++='m'; *dst_ptr++='p'; *dst_ptr++=';'; }
		else *dst_ptr++ = c;
    }
    *dst_ptr = '\0';
    return dest;
}

const char *
cou_createJSHeadFromRsX
(
    co_context *cnx,
    PGresult *res
)
{
    int i,j;
    int nFields=0;
    const char *js_out=0;
    pool *p;
 
    p=cnx->request->pool;
  
    nFields = PQnfields(res);

    js_out = ap_psprintf(p, "var o = new Object();\no.d = new Object();\n");

    for (i = 0; i < nFields; i++)
    {
	js_out=ap_pstrcat(p, js_out, "o.d.",PQfname(res, i), " = new Array();\n", NULL);
    };
 
    return(js_out);

};


const char *
cou_createJSBodyFromRsX
(
    co_context *cnx,
    PGresult *res
)
{
    int i,j;
    int nFields=0;
    int nTuples=0;
    const char *js_out="";
    pool *p;
 
    p=cnx->request->pool;
  
    nFields = PQnfields(res);
    nTuples = PQntuples(res);

    for (i = 0; i < nTuples; i++)
    {
        for (j = 0; j < nFields; j++)
	{
            js_out = ap_pstrcat(p, js_out, 
                        ap_psprintf(p, "o.d.%s[%d] = \"%s\";\n",PQfname(res, j), i, 
                                    cou_buildEncodedJS(cnx,PQgetvalue(res, i, j))),
                                NULL 
                               );
	}
    }
    return(js_out);
};

const char *
cou_createJSBodyFromRsRowX
(
    co_context *cnx,
    PGresult *res,
    int row_num,     // this is number of row to be taken from rs
    int in_js_num   // this is index in javascript table
)
{
    int i,j;
    int nFields=0;
    int nTuples=0;
    const char *js_out="";
    pool *p;
 
    p=cnx->request->pool;
  
    nFields = PQnfields(res);
    nTuples = PQntuples(res);

    for (j = 0; j < nFields; j++)
    {
      js_out = ap_pstrcat(p, js_out, 
            ap_psprintf(p, "o.d.%s[%d] = \"%s\";\n",PQfname(res, j), in_js_num, 
                                    cou_buildEncodedJS(cnx,PQgetvalue(res, row_num, j))),
                                NULL 
                               );
    }
    return(js_out);
};


const char *
cou_createJSCallbackFromRsX
(
    co_context *cnx,
    const char *callback
)
{
    const char *js_out="";
    pool *p;
 
    p=cnx->request->pool;
  
    js_out=ap_pstrcat(p, callback, "o);\n", NULL);
    return(js_out);

};

const char *
cou_createJSFromRsX
(
    co_context *cnx,
    PGresult *res,
    const char *callback
)
{
    const char * js_out = "";
    pool *p;

    p=cnx->request->pool;
  
    js_out=ap_pstrcat(p, 
                 cou_createJSHeadFromRsX(cnx, res),
                 cou_createJSBodyFromRsX(cnx, res),
                 cou_createJSCallbackFromRsX(cnx, callback), 
                 NULL);
    return(js_out);

};


/* cou_spaceTrim deletes spaces in front and at the end */

char*
cou_spaceTrim
(
    co_context *cnx,
    const char *source
)
{
    const char *src_ptr;
    const char *end_ptr;
    char c;
    char *dest;
    int start_text=0;
    int end_text=0;

    src_ptr = source;
    start_text = 0;
    while ( (c = *src_ptr) && (c==' ')) { start_text++; src_ptr++;}
    end_ptr = source + (strlen(source)-1);
    end_text = strlen(source);
    while ( (c = *end_ptr--) && (c==' ') && (end_text>=0) ) { end_text--;}
    dest = (char*)ap_pstrndup( cnx->request->pool, src_ptr, end_text-start_text);
    return dest;
}

/* possible debug levels:
 0 - ignore messages
 1 - show (as) important messages
 2 - show (when client wants) all
*/
int 
cou_sendMessageX
(
  co_context *cnx, 
  const char * message, int debug_level
)
{
	if (debug_level == CO_MSG_FORCE) {
		cou_logError( cnx, CO_LOG_INFO, "%s", message );
	} else {
		cou_logError( cnx, CO_LOG_DEBUG, "%s", message );
	}
    if (cnx->debug_level >= debug_level)
    {
	if (cnx->client_mode == CO_CLIENT_JS)
	{
            cou_msgprintf( cnx, "alert('%s');", cou_buildEncodedJS(cnx, message));
  
	 //   ap_rprintf( cnx->request, "alert('%s');", cou_buildEncodedJS(cnx, message));
	}
	else if ((cnx->client_mode == CO_CLIENT_HTML)||(cnx->client_mode == CO_CLIENT_ASP))
	{
    cou_logError( cnx, CO_LOG_DEBUG, "ctx->html_mode=%d", cnx->html_mode );
	    cou_putsBODY(cnx, "<H1>");
	    cou_putsBODY(cnx, cou_buildHTMLEncodedTextDocument( cnx, ap_escape_html( cnx->request->pool, message)) );
	    cou_putsBODY(cnx, "</H1>");
	}
    }	
	errno = 0;
    return 1;
}

int 
cou_sendMessage
(
  co_context *cnx, 
  const char * message
)
{
    return cou_sendMessageX(cnx, message, 2);
}
				       
