/* 
**  mod_coffice.c -- Apache coffice module
**  [Autogenerated via ``apxs -n coffice -g'']
**
*/ 

#include "httpd.h"
#include "http_config.h"
#include "http_main.h"
#include "http_protocol.h"
#include "http_log.h"
#include "ap_config.h"

/* PostgreSQL access functions */
#include "libpq-fe.h"

#include "co_utils.h"
#include "co_log.h"
#include "co_lib.h"
#include "co_defines_calls.h"
#include "co_defines.h"

#include "co_mod.h" /* include before co_js.h */
#include "co_js.h"
#include "co_action_get.h"
#include "co_action.h"

#include "../comodules/js_code/co_js_code.h"
#include "co_recordset.h"
#include "co_public_web.h"
#include "co_string.h"
#include "co_output.h"

#include "mod_coffice.h"

/* mapping files into memory */
#include <stdlib.h>
#include <unistd.h>
#include <sys/mman.h>
#include <fcntl.h>

/* password encryption */
#include <crypt.h>

/* memory allocation stats */
#include <malloc.h>

/* Dynamic linking of submodules */
#include <dlfcn.h>
#define CO_SUBMODULE_LIMIT 20    /* limits number of submodules that can be loaded */
#define CO_MULTIREQUEST_LEN 16 /* max number of calls handled in a single request */

#include <setjmp.h>	/* error handling */

#define CO_NAME_BUF_LEN 32

module coffice_module;

/* module error handling - not to be set while handling requests;
 * it is intended to be set during module loading/activation */
/* typedef enum { false=0, true=1 } bool; */
static bool coffice_error = false;
static const char* coffice_error_msg = NULL;
static int coffice_usejs = false;
/* connection string (per proccess) - set as PGConn command */
static const char *coffice_pgconn = NULL;
static const char *coffice_module_path = NULL;
static const char *coffice_root_path = NULL;
static const char *coffice_data_path = NULL;
static const char *coffice_html_lib_path = NULL;
static const char *coffice_js_lib_path = NULL;
static const char *coffice_image_path = NULL;
static PGconn *conn = NULL;


static void
send_error_info( request_rec *r, const char *error_msg )
{
    ap_send_http_header(r);
    ap_rputs(DOCTYPE_HTML_3_2, r);
    ap_rputs("<HTML>\n", r);
    ap_rputs("<BODY>\n", r);
    ap_rprintf( r, "<H1>%s</H1>\n", ap_escape_html( r->pool, error_msg) );
    ap_rputs("</BODY>\n", r);
    ap_rputs("</HTML>\n", r);
}

void
cou_putJS( co_context *ctx, const char* script)
{
    cou_logError( ctx, CO_LOG_DEBUG, "cou_putJS");
    if (ctx->client_mode==CO_CLIENT_JS)
    {
	cos_AppendString(ctx, ctx->js_response, script);
	//ap_rputs( script, ctx->request );
    };
}

const char *
cou_getCORootPath( co_context* cnx)
{
    return coffice_root_path;
}

const char *
cou_getCODataPath()
{
    return coffice_data_path;
}

const char *
cou_getCOHtmlLibPath()
{
    return coffice_html_lib_path;
}

const char *
cou_getCOJsLibPath()
{
    return coffice_js_lib_path;
}

const char *
cou_getCOImagePath()
{
    return coffice_image_path;
}

coffice_server_conf *
cox_getCOfficeConfig( co_context *ctx )
{
    return (coffice_server_conf *)ap_get_module_config(ctx->request->server->module_config,
                                                &coffice_module);
}

static
void
cox_beginTrans( co_context *ctx )
{
    PGconn *conn = ctx->dbconn;
    PGresult *res;

    res = PQexec(conn, "BEGIN");
    if (PQresultStatus(res) != PGRES_COMMAND_OK)
    {
        PQclear(res);
    }

    PQclear(res);
}

static
void
cox_commitTrans( co_context *ctx )
{
    PGconn *conn = ctx->dbconn;
    PGresult *res;

    res = PQexec(conn, "END");
    if (PQresultStatus(res) != PGRES_COMMAND_OK)
    {
        PQclear(res);
    }

    PQclear(res);
}

static
void
cox_abortTrans( co_context *ctx )
{
    PGconn *conn = ctx->dbconn;
    PGresult *res;

    res = PQexec(conn, "ABORT");
    if (PQresultStatus(res) != PGRES_COMMAND_OK)
    {
        PQclear(res);
    }

    PQclear(res);
}

int (*
co_callByNameExt( co_context *ctx, const char* submodule_name , const char* prefix, const char* method_name)
)(co_context *, ... )
{
    request_rec *r;
   int i;
   int nelt; /* number of elements */
   array_header* lib_array;
   void *dlh;
   const char *name;
   const char *err_msg;
   coffice_libinfo* elts;
   coffice_libinfo* elt;

   int (*action )(co_context *, ... );

    r = ctx->request;
    lib_array = ((coffice_server_conf *)ap_get_module_config(r->server->module_config,
                                       &coffice_module))->loaded_libs;

    elts = (coffice_libinfo*)lib_array->elts;
    for (i = 0; i < lib_array->nelts; i++) {
   	elt = &elts[i];
        dlh = elt->dlh;
        name = elt->name;
        if ((dlh) && (strcmp(name, submodule_name) == 0)) {
	    action = ap_os_dso_sym( dlh,
			ap_pstrcat(r->pool, prefix, "_", method_name, NULL)
		    );
	    err_msg = dlerror();
	    if (err_msg) cou_logError(ctx, CO_LOG_ERR, "SYMBOL RESOLVING ERROR:\n%s", err_msg );
    	    return action;
    	}

   }
    cou_logError(ctx, CO_LOG_WARNING, "CO submodule '%s' not found in server configuration.", submodule_name );
   return NULL;
};


static
int send_file(co_context *ctx, char *p_path)
{
	FILE *fh;
	int check_security;

	request_rec *r = ctx->request;

	if (strlen(r->path_info)>1)
	{
		ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, "file_path to send:%s",
		ap_pstrcat(r->pool, coffice_root_path, p_path, NULL));
		fh = ap_pfopen(r->pool, ap_pstrcat(r->pool, coffice_root_path, p_path, NULL), "r");
    }
    else 
	{
		/* empty, slash ? */
		check_security = ctx->check_security;
		ctx->check_security = 0;
		ctx->par_area = "sec";
		cou_broker("sec","start_page","getUserStartPage",ctx,NULL);
		ctx->check_security = check_security;

		return OK;
    }
    if (!fh) return DECLINED;
	r->content_type = "application/octet-stream";
	r->content_type = "text/html";
	ap_send_http_header(r);
    ap_send_fd( fh, r ); /* returns total of bytes sent */
    ap_pfclose(r->pool,fh);

    return OK;
}

/* /co/html/ handler: just use single set of arguments */
static
int
co_send_html( co_context *ctx, const char **area_name_ptr, const char **module_name_ptr, const char **method_name_ptr, const char **method_arg_ptr )
{
    request_rec *r;

    /* IFRAME content creation */
    r = ctx->request;

    ctx->par_class = *module_name_ptr;
    ctx->method_args = *method_arg_ptr;
/*
    ap_send_http_header(r);
    ap_rputs(DOCTYPE_HTML_4_0T, r);
    ap_rputs("<HTML>\n", r);
*/
    if (strcmp("getFile",*method_name_ptr)!=0)
	cou_putsHTML( ctx, ""); /* for compatibility */

    ctx->putJS = false;

	{
	    if ( sigsetjmp(ctx->sigenv, CO_ERR_SAVE_SIGNALS) == 0) {
		cox_beginTrans( ctx );
		cou_broker( *area_name_ptr, *module_name_ptr, *method_name_ptr, ctx, NULL);
		cou_logError( ctx, CO_LOG_DEBUG, "broker returned" );
		cox_commitTrans( ctx );
	    }
	    else {
		cox_abortTrans( ctx );
		/* ap_rputs( "<BODY><H1>Critical error - transaction aborted! Check log.</H1></BODY>", r ); */
		ap_rputs( "Error - transaction aborted!", r );
	    }
	}

    cou_logError( ctx, CO_LOG_DEBUG, "Transaction finished" );
    /* ctx->putJS is still false */
    cou_finishHTML( ctx );

    cou_logError( ctx, CO_LOG_DEBUG, "co_send_html returns" );

    return OK;
}

/* asp parsed to script */
#define CO_OUTBUF_LEN 60000
static char out_buf[CO_OUTBUF_LEN];
/* TODO: allocate required memory from request pool */
extern
int co_ap_parse_template(char *strScript, const char *strSource, size_t sizeDest);

int
co_send_activeFromFile( co_context *ctx, const char *filename )
{
    const char *file_name;
    int fd;
    const char *script;
    void *map;
    struct stat stat_buf;
    size_t length;
    request_rec *r;
    pool *p; 
    r = ctx->request;
    p = r->pool;
 
/*
    ap_send_http_header(r);
    ap_rputs(DOCTYPE_HTML_4_0T, r);
*/
    //file_name = ap_pstrcat( r->pool, cou_getCORootPath(ctx), filename, NULL);
    file_name = filename;
    fd = open( file_name, O_RDONLY);
    if (fd==-1) return HTTP_NOT_FOUND; /* can't open file */
    if (fstat(fd, &stat_buf)==-1) return HTTP_NOT_FOUND; /* can't stat file */

    /* par_area and par_class are required for webserver.js */
    cou_putsHTML( ctx, "<HEAD>\n");
    ap_rputs("<meta http-equiv=\"Content-Type\" content=\"text/html; charset=iso-8859-2\">\n",r);

    ap_rputs("<META HTTP-EQUIV=\"Expires\" CONTENT=\"0\">\n",r);
    ap_rputs("<SCRIPT>\n", r);

    ap_rprintf( r, "var  par_class='%s';\n", ctx->par_class );
    ap_rprintf( r, "var  par_area = '%s';\n", ctx->par_area );

    ap_rputs("</SCRIPT></HEAD>\n", r);

    length = stat_buf.st_size+1;
    co_ap_parse_template(out_buf, (map=mmap( 0, length, PROT_READ, MAP_SHARED, fd, 0)), CO_OUTBUF_LEN);
    out_buf[CO_OUTBUF_LEN-1]='\0';
    munmap(map, length);
    close(fd);

    {
	__PARAM_STRUCT_CREATE_evaluate(in_params)
	__PARAM_STRUCT_SET(ctx->par_area,       in_params, par_area)
	__PARAM_STRUCT_SET("js_code",           in_params, par_class)
	__PARAM_STRUCT_SET(-2,                  in_params, par_id)
	__PARAM_STRUCT_SET(out_buf,             in_params, direct_code)
	__PARAM_STRUCT_SET(1,                   in_params, asp)
	cou_broker( ctx->par_area, "js_code", "evaluate",
            ctx, in_params );
    }

    return OK;
}

static
int 
co_send_active( co_context *ctx, const char **area_name_ptr, const char **class_name_ptr, const char **method_name_ptr, const char **method_arg_ptr )
{
    const char *file_name;
    request_rec *r;
    int result;

    r = ctx->request;

    ctx->par_class = *class_name_ptr;
    ctx->par_area = *area_name_ptr;
    ctx->putJS = false;

    file_name = ap_psprintf( r->pool, "%s/%s/%s.asp", cou_getCORootPath(ctx), *class_name_ptr, *method_name_ptr);

    //ap_rputs("<BODY>\n", r);
    if ( sigsetjmp(ctx->sigenv, CO_ERR_SAVE_SIGNALS) == 0) {
	cox_beginTrans( ctx );
	result = co_send_activeFromFile( ctx, file_name );
	cox_commitTrans( ctx );
    }
    else {
	result = OK;
	cox_abortTrans( ctx );
	/* ap_rputs( "Critical error - transaction aborted! Check log.", r ); */
	ap_rputs( "Error - transaction aborted!", r );
    }
    //ap_rputs("</BODY>\n", r);
	if (result==HTTP_NOT_FOUND) {
		ap_rputs("<HTML>\n<BODY>\n<H1>Error: file ", r);
		ap_rputs(file_name, r);
		ap_rputs(" not found</BODY>\n</HTML>", r);
		result = OK;
	}

    return result;
}

static
int 
co_send_javascript( co_context *ctx, const char **area_names, const char **module_names, const char **method_names, const char **method_args )
{
    request_rec *r;

    /* IFRAME content creation */
    r = ctx->request;
/*
    ap_send_http_header(r);
    ap_rputs(DOCTYPE_HTML_4_0T, r);
    ap_rputs("<HTML>\n", r);
*/
    cou_putsHTML( ctx, "<HEAD>"); /* begin HTML */
    ap_rputs("<META HTTP-EQUIV=\"Expires\" CONTENT=\"0\">\n",r);
    ap_rputs("<meta http-equiv=\"Content-Type\" content=\"text/html; charset=iso-8859-2\">\n",r);
    ap_rputs("</HEAD>\n", r);

    ap_rputs("<BODY>\n", r);
    {
        int *RequestN;
        int *item_id;
	int err;

        RequestN = (int*)cou_getParam( ctx,
                "RequestN",
                CO_INTEGER,
                &err
            );
	/* TODO: check err, print err info */
	ap_rprintf( r, "<BR>USER: %s\n", r->connection->user );
	ap_rprintf( r, "<BR>id_session: %d\n", ctx->id_session );
	ap_rprintf( r, "<BR>id_user: %d\n", ctx->id_user );
	ap_rprintf( r, "<BR>id_role: %d\n", ctx->id_role );
	if (RequestN)
	    ap_rprintf( r, "<BR>RequestN: %d\n", *RequestN );
#if defined(CO_DEBUG)
	if (r->args)
	    ap_rprintf( r, "<BR>request-&gt;args: %s\n",ap_escape_html( r->pool, r->args ) );
#endif
	ap_rputs("<BR><textarea id='action_code' rows=24 cols=50></textarea>\n", r);
	ap_rputs("<SCRIPT>\n", r);
	ap_rputs("function action() { with(parent) {\n", r);
	ctx->putJS = true;
	/* always send waiting actions */
	cox_sendWaitingActionsX( ctx );
	{
      	cou_logError(ctx,CO_LOG_DEBUG,"co_send_javascript he");

	    if ( sigsetjmp(ctx->sigenv, CO_ERR_SAVE_SIGNALS) == 0) {
		int i=0;
		cox_beginTrans( ctx );
		while((area_names[i]!=NULL)&&(module_names[i]!=NULL)&&(method_names[i]!=NULL)) {
		    ctx->par_class = module_names[i];
		    ctx->method_args = method_args[i];
		    cou_broker( area_names[i], module_names[i], method_names[i], ctx, NULL);
			if ( (ctx->result_type == CO_RECORDSET) && (ctx->result != NULL) )
			{
				if (!cor_IsCleared(ctx,ctx->result)) cor_Clear(ctx,ctx->result);
			}
		    i++;
		}
		cox_commitTrans( ctx );
		cos_SendString(ctx, ctx->js_response); /* only if successful */
	    }
	    else {
		cox_abortTrans( ctx );
		/* ap_rputs( "alert('Critical error - transaction aborted! Check log.');", r ); */
		ap_rputs( "alert('Error - transaction aborted!');", r );
	    }
	}
	cos_SendString(ctx, ctx->messages); /* only if successful */
	/* send new action only when succes/failure of the transaction is decided */
	//cox_sendLocalActions( ctx );
	ctx->putJS = false;
	ap_rputs("  }\n};\n", r);

	ap_rputs("document.getElementById('action_code').appendChild( document.createTextNode( String(action).replace(/\\n/g,'\\r') ));\n", r);
	if (RequestN) {
	    ap_rprintf( r, "parent.DoIFrame[%d] = action;\n", *RequestN );
	    ap_rprintf( r, "parent.DoneIFrame[%d]=1;\n", *RequestN );
	    ap_rputs("parent.DoIF();\n", r);
	}
	else { /* it is not "valid request", try to fire action at once */
	    ap_rputs("action();\n", r);
	}
	ap_rputs("</SCRIPT>\n", r);
    }
    ap_rputs("</BODY>\n", r);
    ap_rputs("</HTML>\n", r);
    return OK;
}

typedef
int 
(*call_handler)( co_context *, const char **, const char **, const char **, const char ** );

typedef enum { CALL_JS=0, CALL_HTML=1, CALL_ASP=2 } call_type;
typedef struct url_type {
    const char *branch;
    int length;
} url_type ;

static url_type handled_urls[] = {
    {"/javascript/", 12},
    {"/html/", 6},
    {"/asp/", 5},
    {NULL, 0}
};

#define CLIENT_BLOCK_LENTH 2048

static
int 
co_handle_url_call(co_context *ctx, call_type call_id, call_handler handler)
{
    char *str_ptr = NULL;
    char *str_area = NULL;
    char *str_scan_ptr = NULL;
    char area_name_buf[CO_NAME_BUF_LEN];
    char *args; /* args stripped of method name */
    char *arg_ptr; /* pointer used to scan args */

    const char* areas[CO_MULTIREQUEST_LEN+1];
    const char* classes[CO_MULTIREQUEST_LEN+1];
    const char* methods[CO_MULTIREQUEST_LEN+1];
    const char* method_args[CO_MULTIREQUEST_LEN+1];

    request_rec *r;
	char *local_args;
	char *local_path_info;

    r = ctx->request;
	local_args = r->args;
	local_path_info = r->path_info;

if (r->method_number == M_POST) {
    /* Content-Length is required */
    int i;
    char client_block[CLIENT_BLOCK_LENTH+1];
//    char *post_part="&";
    char *buf;
    char *buf_start;
    int buf_len;
    long len;
    long post_len;
    pool *local_subpool;

    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, "POST Content-Length=%s",
	    ap_table_get(r->headers_in, "Content-Length")
    );
//    ap_setup_client_block( r, REQUEST_CHUNKED_DECHUNK );
    ap_setup_client_block( r, REQUEST_CHUNKED_ERROR );
    post_len = r->remaining;

    local_subpool = ap_make_sub_pool(r->pool);

    buf = (char*)ap_palloc(local_subpool, post_len+1);
    buf_start = buf;
    buf_len = CLIENT_BLOCK_LENTH;
    while ((len = ap_get_client_block( r, buf, buf_len ))>0) {
	buf += len;
	if ((buf_start + post_len - buf) < buf_len) {
	    buf_len = buf_start + post_len - buf;
	}
    }
/*
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, "Bytes in local subpool=%d",
	ap_bytes_in_pool(local_subpool)
    );
*/
/*
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, "Bytes in destroyed subpool=%d",
	ap_bytes_in_pool(p)
    );
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, "Bytes in request pool=%d",
	ap_bytes_in_pool(r->pool)
    );
*/
    if (len<0) {
	send_error_info( r,"Error while reading POST client block" );
	return OK;
    }
    *(buf_start + post_len) = '\0';
    ctx->post_args = buf_start;

    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, "post=%s",ctx->post_args );
#define CO_PARTS_COUNT 8
	/* try to find query_line1=... */
	/* rewrites local_path_info and local_args */
	if (strncmp(buf_start, "query_line1=", strlen("query_line1="))==0) {
		/* parts follow... */
		int part_offset = strlen("&query_line1=");
		char *result_ptr = buf_start;
		char *source_ptr = buf_start;
		char *next_source_ptr = buf_start;
/*
		static const char *tag_array[] = { "&query_line2=", "&query_line3=",
				"&query_line4=", "&query_line5=", "&query_line6=",
				"&query_line7=", "&query_line8="
				};
*/
		int n;
		source_ptr--; /* just to allow for the same separator length as other parts */
		source_ptr += part_offset;
/*		
		ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, "ok 1" );
		for (n=0; n<CO_PARTS_COUNT-1; n++) {
			next_source_ptr = strstr( source_ptr, tag_array[n]);
			memmove( result_ptr, source_ptr, (next_source_ptr-source_ptr) );
			result_ptr += (next_source_ptr-source_ptr);
			source_ptr = next_source_ptr + part_offset;
		    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, "ok loop %d", n );
		}
		ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, "ok 2" );
*/
		/* copy last part - to the end; including \0 */
		memmove( result_ptr, source_ptr, (buf_start + post_len - source_ptr) );
        cou_decode_string_post(buf_start);
		ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, "ok 3" );
		ctx->post_args = NULL;
		ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, "post decoded=%s", buf_start );
		/* find '?'; set 0 there, set local_args to the next char */

		local_path_info = buf_start;
		local_args = index( buf_start, '?' );
		*local_args = '\0';
		local_args++;
	} /* ... parts found */
} else ctx->post_args = NULL;

	    str_area = (local_path_info)+handled_urls[call_id].length;

	    /* expect area name: alphanumeric section */
	    str_ptr = str_area;
	    while (ap_isalnum(*str_ptr)) str_ptr++; /* for now just skip it */
	    if (*str_ptr == '/') { /* next section is class_name */
		/* point to class_name */
		str_ptr++;

		// TODO: check CO_NAME_BUF_LEN
		/* previous increment used to add place for ending '\0' */
		if (str_ptr - str_area >= CO_NAME_BUF_LEN) {
		    send_error_info( r, "Area name is too long!" );
		    return OK;
		}
		ap_cpystrn(area_name_buf, str_area, str_ptr - str_area );

		/* now assume it is component name */
		if(!local_args) {
		    send_error_info( r, "No method arguments" );
		    return OK;
		}
		args = ap_pstrdup(r->pool, local_args);
		arg_ptr = args; /* two pointers to the duplicate of request args */
		if (*str_ptr != '*') {
		    while (ap_isalnum(*arg_ptr)) arg_ptr++;
		    if (*arg_ptr == '&') {
	    		// there are args for the procedure
	    		(*arg_ptr++) = '\0';
	    		// args points to method name, arg_ptr to the encoded arguments
		    } else if (*arg_ptr) {
	    		// malformed co request
			send_error_info( r, "Bad method name" );
			return OK;
		    };
		    areas[0] = area_name_buf;
		    classes[0] = str_ptr;
		    methods[0] = args;
		    method_args[0] = local_args; /* pass everything as method args */
		    areas[1] = NULL;
		    classes[1] = NULL;
		    methods[1] = NULL;
		    method_args[1] = NULL;
		}
		else {
		    int i=0;
		    int loop=1;
		    /* ASSUMES: arg_ptr, args are two pointers to the duplicate of
		       the beginning of new section of request args */
		    /* SECTION SEPARATORS: && */
		    while(*arg_ptr) {
			/* find fragment of the form: class/method& */
			/* we accept underscore in class name */
			while (ap_isalnum(*arg_ptr) || (*arg_ptr=='_')) arg_ptr++;
			if (*arg_ptr == '/') {
	    		    // there are args for the procedure
	    		    (*arg_ptr++) = '\0';
	    		    // args points to component name, arg_ptr to the encoded arguments
			} else if (*arg_ptr) {
	    		    // malformed co request
			    send_error_info( r, "Bad component name: only alphanumeric and underscore accepted" );
			    return OK;
			};
			areas[i] = area_name_buf;
			classes[i] = args;
			methods[i] = arg_ptr; /* should be method name */
			while (ap_isalnum(*arg_ptr)) arg_ptr++;
			if (*arg_ptr == '&') {
	    		    *arg_ptr = '\0'; /* mark end of method name */
			    methods[i] = ap_pstrdup(r->pool, methods[i]); /* should be method name, use copy */
	    		    // args points to component name, arg_ptr to the encoded arguments
	    		    *arg_ptr = '&'; /* needed by getParam */
			    method_args[i] = arg_ptr; /* getParam input */

    			} else if (*arg_ptr) {
	    		    // malformed co request
			    send_error_info( r, "Bad method name" );
			    return OK;
			};
			while ((*arg_ptr)&&!((*arg_ptr=='&')&&(*(arg_ptr+1)=='&'))) arg_ptr++;
			if (*arg_ptr) {
			    /* double '&' found; skip them */
			    ++arg_ptr;
			    args = ++arg_ptr;
			}
			i++;
			if (i>CO_MULTIREQUEST_LEN) break; /* tables are full */
		    } /* we got to the end of request->args copy */

		    /* mark table ends */
		    areas[i] = NULL;
		    classes[i] = NULL;
		    methods[i] = NULL;
		    method_args[i] = NULL;
		}
		ctx->par_area = area_name_buf;
		ctx->method_args = local_args; /* we need args to get id_session */
		{
		    int *int_ptr;
		    int err;

		    int_ptr = (int*)cou_getParam(ctx,"id_session",CO_INTEGER, &err);
		    if (int_ptr) ctx->id_session = *int_ptr;
		    int_ptr = (int*)cou_getParam(ctx,"debug_level",CO_INTEGER, &err);
		    if (int_ptr) ctx->debug_level = *int_ptr;
		    //ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, "OK");
		}
		return handler( ctx, areas, classes, methods, method_args );

	    } /* '/' after area name else bad area name */
	    else {
		send_error_info( r, "Missing area name or class name" );
		return OK;
	    }

}

static
int send_pgdata(co_context *ctx)
{
	int			nFields;
	int			i,
				j;
	PGresult   *res;
	request_rec *r;
	r = ctx->request;
	ctx->client_mode = CO_CLIENT_NO_MODE;
        ctx->error = CO_NO_ERRORS;
        ctx->html_mode = CO_OUTSIDE;

        // this is very important place! here we switch on security! 
        ctx->check_security = 1;
	ctx->critical = true;

	if ((ctx->id_user==0)||(ctx->id_role==0))
	{
		int ret = 0;
		ctx->client_mode = CO_CLIENT_HTML;
		ctx->send_data   = 1;
		ctx->par_area = "usr";
		// we need to add one & becouse we do not have method name
	    ctx->method_args = ap_pstrcat(r->pool, "&", r->args, NULL);

		if ( sigsetjmp(ctx->sigenv, CO_ERR_SAVE_SIGNALS) == 0)
		{
			int i=0;
			cox_beginTrans( ctx );
			ret = co_public_web_map(ctx);
			cox_commitTrans( ctx );
			cos_SendString(ctx, ctx->js_response); /* only if successful */
	    }
	    else 
		{
			cox_abortTrans( ctx );
			ret = OK;
			/* ap_rputs( "alert('Critical error - transaction aborted! Check log.');", r ); */
			ap_rputs( "alert('Error - transaction aborted!');", r );
	    }
		return ret;
	}

	if ( strncmp( r->path_info, "/script/", 6) == 0 )
	{ 
	    char client_block[CLIENT_BLOCK_LENTH+1];
	    ap_send_http_header(r);
	    /* HTML creation */
    	    ap_rputs(DOCTYPE_HTML_3_2, r);
	    ap_rputs("<HTML>\n", r);
    	    ap_rputs("<BODY>\n", r);
	    ap_rputs( ap_array_pstrcat(r->pool, ap_table_elts(r->headers_in), ';'), r);
	    ap_rputs("<BR>\n", r);
    	    ap_rputs(r->the_request, r);
    	    ap_rputs("<BR>\n", r);
	    /*
	    ap_setup_client_block( r, REQUEST_CHUNKED_PASS );
	    */
	    ap_setup_client_block( r, REQUEST_CHUNKED_DECHUNK );
	    for (i = 0; i<=CLIENT_BLOCK_LENTH; i++) {  client_block[i]=0; }; 
	    ap_get_client_block( r, client_block, CLIENT_BLOCK_LENTH );
	    cou_decode_string( client_block);
	    ap_rputs( client_block, r);
    	    ap_rputs("</BODY>\n", r);
	    ap_rputs("</HTML>\n", r);
	}
	else if ( strncmp( r->path_info, "/ado/", 5) == 0 )
	{
	    r->content_type = "text/xml";
	    ap_send_http_header(r);
	    ap_rputs(
		cor_GetADOXMLSerialization( ctx,
		    cor_CreateOnSQLString( ctx, "SELECT * FROM co_sys_meta_class")
		),
		r
	    );
	}
	else if (strncmp( r->path_info, handled_urls[CALL_JS].branch, handled_urls[CALL_JS].length) == 0) 
	{
	    ctx->client_mode = CO_CLIENT_JS;
            ctx->send_data   = 1;
	    return co_handle_url_call(ctx, CALL_JS, co_send_javascript);
	}
	else if (strncmp( r->path_info, handled_urls[CALL_HTML].branch, handled_urls[CALL_HTML].length) == 0) 
	{
	    ctx->client_mode = CO_CLIENT_HTML;
            ctx->send_data   = 1;
	    return co_handle_url_call(ctx, CALL_HTML, co_send_html);
	}
	else if (strncmp( r->path_info, handled_urls[CALL_ASP].branch, handled_urls[CALL_ASP].length) == 0) 
	{
	    ctx->client_mode = CO_CLIENT_ASP;
            ctx->send_data   = 0;
	    return co_handle_url_call(ctx, CALL_ASP, co_send_active);
	}
	else if(strncmp( r->path_info, "/images/", 8) == 0)
	{
	    return send_file(ctx, ap_psprintf(r->pool,"%s/%s",cou_getCOImagePath(),r->path_info+8));

	}
	else
	{
	    return send_file(ctx, r->path_info);
	    /* return DECLINED; */
	}
	return OK;
}


/* The content handler */
static int coffice_handler(request_rec *r)
{
    int result;
    r->content_type = "text/html";      
//    r->content_type = "image/gif";      

    /* find more about timeouts: */
    ap_soft_timeout(ap_pstrdup(r->pool,"send example call trace"), r);

    if (r->header_only) {
	ap_send_http_header(r);
	/* already done */
        ap_kill_timeout(r);
        return OK;
    }

    if (coffice_error) {
	/* r->content_type = "message/http"; */
	send_error_info( r, coffice_error_msg );
	return OK;
    }

    {
	co_context *ctx;
	ctx=(co_context*)ap_get_module_config(r->request_config, &coffice_module);
	if (!ctx) {
	    //send_error_info( r, "Access forbidden: ask system administrator." );
    	/* somebody logins: create context of this connection */
    	ctx = ap_pcalloc( r->pool, sizeof(co_context) );
    	ctx->request = r;
    	ctx->dbconn = conn;

    	/* set authentication info */
    	ctx->id_user=0;
    	ctx->id_role=0;
    	ctx->language = "pl_PL";

    	/* init other context fields */
    	ctx->id_session=0;
    	ctx->jscnx = NULL;
    	ctx->putJS = false;
    	ctx->result_type = CO_NO_TYPE;
    	ctx->debug_level = 0;

	    //return OK;
	}
	/* prepare response and message lists */
	ctx->js_response = cos_CreateEmpty(ctx);
	ctx->messages = cos_CreateEmpty(ctx);

	/* do real work */
	result = send_pgdata(ctx);

/*
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, "Bytes in request pool=%d",
	ap_bytes_in_pool(r->pool)
    );
*/
	/* cleanup */
        if ( ctx->jscnx ) /* there are contexts remaining */
            cox_finishJSContext(ctx);
        if ( (ctx->result_type == CO_RECORDSET) && (ctx->result != NULL) )
		{
                if (!cor_IsCleared(ctx,ctx->result)) cor_Clear(ctx,ctx->result);
		}
    }
/*
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, "Bytes in request pool=%d",
	ap_bytes_in_pool(r->pool)
    );
*/
/*
    malloc_stats();
*/    
    ap_kill_timeout(r);
    ap_finalize_request_protocol(r);
    return result;
}

static int coffice_validate_user(request_rec *r)
{
    char *sent_pw, *real_pw;
    char *user;
    int res;
    //int id_preference;
    PGresult *pgres;
    co_context *ctx; /* here we create co context */

    if ((res = ap_get_basic_auth_pw(r, (const char **)&sent_pw)))
        return res;

    if (!(user = r->connection->user)) return AUTH_REQUIRED;

    /* somebody logins: create context of this connection */
    ctx = ap_pcalloc( r->pool, sizeof(co_context) );
    ctx->request = r;
    ctx->dbconn = conn;

    pgres = PQexec(conn, ap_pstrcat(r->pool, "SELECT id, passwd, id_role FROM co_sec_user WHERE deleted=0 AND login='", cou_buildEncodedSQL(ctx, user), "'", NULL));
    if (PQresultStatus(pgres) != PGRES_TUPLES_OK)
    {
        PQclear(pgres);
        return AUTH_REQUIRED;
    }
    if (PQntuples(pgres)==0) {
        PQclear(pgres);
        return FORBIDDEN;
    }
    if (PQntuples(pgres)!=1) {
        PQclear(pgres);
        return PRECONDITION_FAILED;
    }
    real_pw = PQgetvalue(pgres, 0, 1);

    if (strncmp(real_pw,"$1$", 3)) {
		/* malformed user.passwd or no passwd */
        PQclear(pgres);
        return FORBIDDEN;
    }

    if(strcmp(real_pw,crypt(sent_pw,real_pw))) {
        ap_log_rerror(APLOG_MARK, APLOG_INFO, r, "User %s: password mismatch",user);
        //log_reason (pg_errstr, r->uri, r);
        ap_note_basic_auth_failure (r);
        PQclear(pgres);
        return FORBIDDEN;
    }
    /* we have user found in database and the password is correct */

    /* set authentication info */
    ctx->id_user=atoi(PQgetvalue(pgres, 0, 0));
    ctx->id_role=atoi(PQgetvalue(pgres, 0, 2));

//    id_preference=atoi(PQgetvalue(pgres, 0, 3));
    PQclear(pgres);

    /* set locale */
    pgres = PQexec(conn, ap_psprintf(r->pool, "SELECT language FROM co_sec_preference WHERE deleted=0 AND id_user=%d ", ctx->id_user));
    if (PQresultStatus(pgres) != PGRES_TUPLES_OK)
    {
        PQclear(pgres);
        return PRECONDITION_FAILED;
    }
    if (PQntuples(pgres)<1) {
        PQclear(pgres);
        return PRECONDITION_FAILED;
    }

    ctx->language = ap_pstrdup(r->pool, PQgetvalue(pgres, 0, 0));

    /* init other context fields */
    ctx->id_session=0;
    ctx->jscnx = NULL;
    ctx->putJS = false;
    ctx->result_type = CO_NO_TYPE;
    ctx->debug_level = 0;
	
    /* pass context to request handler: use module's slot in request config */
    ap_set_module_config(r->request_config, &coffice_module, (void *)ctx);

    PQclear(pgres);
    return OK;

}

static const char *cmd_pgconn(cmd_parms *cmd, void *mconfig,
                                char *pgconn_config)
{
    ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, cmd->server, "cmd_pgconn: DB connection string set to '%s'", pgconn_config);
    coffice_pgconn = pgconn_config;
    return NULL;
}

static const char *cmd_comodulepath(cmd_parms *cmd, void *mconfig,
                                char *co_module_path)
{
    ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, cmd->server, "cmd_comodulepath: module path set to '%s'", co_module_path);
    coffice_module_path = co_module_path;
    return NULL;
}

static const char *cmd_corootpath(cmd_parms *cmd, void *mconfig,
                                char *path)
{
    ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, cmd->server, "cmd_corootpath: set to: %s", path);
    coffice_root_path = path;
    return NULL;
}

static const char *cmd_codatapath(cmd_parms *cmd, void *mconfig,
                                char *path)
{
    ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, cmd->server, "cmd_codatapath: set to: %s", path);
    coffice_data_path = path;
    return NULL;
}

static const char *cmd_cohtmllibpath(cmd_parms *cmd, void *mconfig,
                                char *path)
{
    ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, cmd->server, "cmd_cohtmllibpath: set to: %s", path);
    coffice_html_lib_path = path;
    return NULL;
}

static const char *cmd_cojslibpath(cmd_parms *cmd, void *mconfig,
                                char *path)
{
    ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, cmd->server, "cmd_cojslibpath: set to: %s", path);
    coffice_js_lib_path = path;
    return NULL;
}

static const char *cmd_coimagepath(cmd_parms *cmd, void *mconfig,
                                char *path)
{
    ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, cmd->server, "cmd_coimagepath: set to: %s", path);
    coffice_image_path = path;
    return NULL;
}

static const char *cmd_usejavascript(cmd_parms *cmd, void *mconfig, int js_set)
{
    if (js_set) ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, cmd->server, "cmd_usejavascript: JavaScript support turned on");
    coffice_usejs = js_set;
    return NULL;
}

/* comments on library/module handling see: mod_so.c */
static void unload_library(coffice_libinfo *libinfo)
{
    if (libinfo->dlh == NULL)
        return;

    /*dlclose (libinfo->dlh);*/
    ap_os_dso_unload(libinfo->dlh);

    libinfo->dlh = NULL;
    libinfo->name = NULL;
    libinfo->new_param = NULL;
    libinfo->set_param = NULL;
    libinfo->get_param = NULL;
}


static const char *cmd_coloadsubmodule(cmd_parms *cmd, void *dummy,
                               const char *name /* name of the submodule */)
{
	coffice_libinfo *libinfo;
        coffice_server_conf *sc;
        void *dlh;
	
	/* you can do: dlh = ap_os_dso_open(...) - it is more portable
	 but there is no need of exporting symbols (ap_os_dso_open uses RTLD_NOW | RTLD_GLOBAL)*/
	dlh = dlopen(
		(char*)ap_pstrcat(cmd->pool, coffice_module_path, "/co_", name, ".so", NULL),
		RTLD_NOW
	    );
        if (dlh) {
   		sc = (coffice_server_conf *)ap_get_module_config(cmd->server->module_config,
                                                &coffice_module);
  	
		libinfo = ap_push_array( sc->loaded_libs );
		libinfo->name = name;
		libinfo->dlh = dlh;
		libinfo->new_param = ap_os_dso_sym( dlh, 
		    ap_pstrcat( cmd->pool, "coj_", name, "_new_param", NULL )
		);
		libinfo->set_param = ap_os_dso_sym( dlh, 
		    ap_pstrcat( cmd->pool, "coj_", name, "_set_param", NULL )
		);
		libinfo->get_param = ap_os_dso_sym( dlh, 
		    ap_pstrcat( cmd->pool, "coj_", name, "_get_param", NULL )
		);
		ap_register_cleanup(cmd->pool, libinfo,
                     (void (*)(void*))unload_library, ap_null_cleanup);
       	}
	else {
	    ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, cmd->server, 
		"cmd_coloadsubmodule: cannot open module %s\nLOADING ERROR: dlopen failed\nREASON: %s",
		name,dlerror());
	}
	return NULL;
}

static void *coffice_sconf_create(pool *p, server_rec *s)
{
    coffice_server_conf *sc;

    sc = (coffice_server_conf *)ap_pcalloc(p, sizeof(coffice_server_conf));
    /* create array for maximum number of libinfo structures */
    sc->loaded_libs = ap_make_array(p, CO_SUBMODULE_LIMIT,
                                     sizeof(coffice_libinfo));

    return (void *)sc;
}

/*
 * module-initialiser
 */
static void coffice_init(server_rec *s, pool *p)
{

}

/*
 * All our process-initialiser does is init database connection.
 */
static void coffice_child_init(server_rec *s, pool *p)
{
	coffice_server_conf *sc;
	
	/* make a connection to the database */
	if (conn == NULL) {
	   conn = PQconnectdb(coffice_pgconn);
	};

	/* check to see that the backend connection was successfully made */
	if (PQstatus(conn) == CONNECTION_BAD)
	{
		ap_log_error(APLOG_MARK, APLOG_ERR, s, "mod_coffice: Connection to database '%s' failed.\n", coffice_pgconn);
		ap_log_error(APLOG_MARK, APLOG_ERR, s, "mod_coffice: %s", PQerrorMessage(conn));
		coffice_error = true;
		coffice_error_msg = "Database connection failed; module inactive.";
		PQfinish(conn);
		return; 
	}

	/* we have a valid connection: add it to the server config struct */
	sc = (coffice_server_conf *)ap_get_module_config(s->module_config,
                                    &coffice_module);
	sc->dbconn = conn;

	/* if javascript is required - init it and add to server conf, too */	
	if (coffice_usejs) {
	    cox_initJSRuntime(sc);
	    if (sc->jsrt == NULL) {
		ap_log_error(APLOG_MARK, APLOG_ERR, s, "Cannot create JSRuntime");
	    }
	    //cox_buildInheritanceScript( p, conn ); // called in compileJSScript
	    cox_compileJSScript(s, p, sc );
	}
}


static void coffice_child_exit(server_rec *s, pool *p)
{
        if (coffice_usejs) {
	    coffice_server_conf *sc;
	    sc = (coffice_server_conf *)ap_get_module_config(s->module_config,
                                                &coffice_module);
	    cox_finishJSRuntime(sc);
	}

	/* close the connection to the database and cleanup */
	PQfinish(conn);
}

/* 
 * List of directives specific to our module.
 */
static const command_rec coffice_cmds[] =
{
    {
        "PGConn",              /* directive name */
        cmd_pgconn,            /* config action routine */
        NULL,                   /* argument to include in call */
        OR_OPTIONS,             /* where available */
        TAKE1,                /* arguments */
        "PostgreSQL connection string - one argument (see pqlib docs)"
                                /* directive description */
    },
    {
        "COModulePath",
        cmd_comodulepath,
        NULL,
        OR_OPTIONS,
        TAKE1,
        "Full path to CO (sub)module directory - one argument"
    },
    {
        "COModuleLoad",
        cmd_coloadsubmodule,
        NULL,
        OR_OPTIONS,
        TAKE1,
        "Name of CO submodule/library. File co_<lib_name>.so should be placed in COModulePath"
    },
    {
        "CORootPath",
        cmd_corootpath,
        NULL,
        OR_OPTIONS,
        TAKE1,
        "Full path to dir containing ordinary CO Web files - one argument"
    },
    {
        "CODataPath",
        cmd_codatapath,
        NULL,
        OR_OPTIONS,
        TAKE1,
        "Full path to dir containing data files for modules - one argument"
    },
    {
        "COHtmlLibPath",
        cmd_cohtmllibpath,
        NULL,
        OR_OPTIONS,
        TAKE1,
        "Full path to dir containing CO files - one argument"
    },
    {
        "COJsLibPath",
        cmd_cojslibpath,
        NULL,
        OR_OPTIONS,
        TAKE1,
        "Full path to dir containing CO JavaScript library files - one argument"
    },
    {
	"COImagePath",
	cmd_coimagepath,
	NULL,
	OR_OPTIONS,
	TAKE1,
	"Full path to dir containing grahical elements - one argument"
    },
    {
        "COUseJavaScript",
        cmd_usejavascript,
        NULL,
        OR_OPTIONS,
        FLAG,
        "JavaScript engine is loaded into every server process if set (On - Off)"
    },
    {NULL}
};

/* Dispatch list of content handlers */
static const handler_rec coffice_handlers[] = { 
    { "coffice", coffice_handler }, 
    { NULL, NULL }
};

/* Dispatch list for API hooks */
module MODULE_VAR_EXPORT coffice_module = {
    STANDARD_MODULE_STUFF, 
    coffice_init,          /* module initializer                  */
    NULL,                  /* create per-dir    config structures */
    NULL,                  /* merge  per-dir    config structures */
    coffice_sconf_create,  /* create per-server config structures */
    NULL,                  /* merge  per-server config structures */
    coffice_cmds,          /* table of config file commands       */
    coffice_handlers,      /* [#8] MIME-typed-dispatched handlers */
    NULL,                  /* [#1] URI to filename translation    */
    coffice_validate_user, /* [#4] validate user id from request  */
    NULL,                  /* [#5] check if the user is ok _here_ */
    NULL,                  /* [#3] check access by host address   */
    NULL,                  /* [#6] determine MIME type            */
    NULL,                  /* [#7] pre-run fixups                 */
    NULL,                  /* [#9] log a transaction              */
    NULL,                  /* [#2] header parser                  */
    coffice_child_init,    /* child_init                          */
    coffice_child_exit,    /* child_exit                          */
    NULL                   /* [#0] post read-request              */
#ifdef EAPI
   ,NULL,                  /* EAPI: add_module                    */
    NULL,                  /* EAPI: remove_module                 */
    NULL,                  /* EAPI: rewrite_command               */
    NULL                   /* EAPI: new_connection                */
#endif
};

