#include "../../include/types.t"      
#include "../../include/defines.t"      

// CLASS CATALOG

class catalog

{ 
  area "sec";
  area "sys";
  area "usr";
 
#include "../../include/check_acl.t"
 
  inherits generic_acl;
  
  container_class 'catalog';
  container_column_name 'id';

  include_c "http_protocol.h";
  include_c "http_main.h";
  include_c "co_session.h";
      
 // PROPERTIES
      
  default_form "showCatalog";

#include "../../include/catalog_props.t"      

  prop id_parent
  {
    type CO_INTEGER;
    default 0;
    params '';
    representant 0;
    foreign 'catalog';
    foreign_prop 'id';
    position 350;
  };

// METHODS

    uses "generic";
    uses "generic_acl";
	uses "clipboard";

	method  cut
	{
		id_permission __PERMISSION_DELETE;
		sec_type __METHOD_SEC_TYPE_OBJECT;
 
		doc "Deletes objects and puts it into clipboard."
		{
			pl "Kasuje obiekt i wklada go do clipboardu";
		};

		__PARAMS_HEADER

		declare language_c
		{
			C "int c;";
			C "char * serialization;";
		};
 
		body language_c
		{
			C 
			%%
			// we do not want to start from @
			@co_call(par_area, par_class, "delete", par_area=>par_area, par_class=>par_class, par_id=>par_id, version=>-1);
			serialization = ap_psprintf(p, "%d", par_id);
			@co_call(par_area, '"clipboard"', "insertClipboardObject", par_area=>par_area, par_class=>'"clipboard"', par_id=>-1, serialization=>serialization, owner_area=>par_area, owner_class=>par_class);

			%%
        ;
     };
   }; 
 
	method  paste
	{
		id_permission __PERMISSION_INSERT_CHILD;
		sec_type __METHOD_SEC_TYPE_OBJECT;
 
		doc "Gets objects from clipboard, undeletes it ands puts into current dir."
		{
			pl "Pobiera obiekt z clipboardu, odkasowuje go i wkada do aktualnego katalogu";
		};

		__PARAMS_HEADER

		declare language_c
		{
			C "int id_from_clipboard;";
			C "char * serialization;";
		};
 
		body language_c
		{
			C 
			%%
			// we do not want to start from @
			@co_call(par_area, '"clipboard"', "retrieveClipboardObject", par_area=>par_area, par_class=>'"clipboard"', par_id=>-1, owner_area=>par_area, owner_class=>par_class);
			__GET_RETURNED_STRING(serialization)
			id_from_clipboard = atoi(serialization);

			@co_call(par_area, par_class, "undelete", par_area=>par_area, par_class=>par_class, par_id=>id_from_clipboard, version=>-1);
			@co_call(par_area, par_class, "moveItemByID", par_area=>par_area, par_class=>par_class, par_id=>id_from_clipboard, new_catalog=>par_id);

			%%
        ;
     };
   }; 
 
	method  moveItemByID
	{
		id_permission __PERMISSION_UPDATE;
		sec_type __METHOD_SEC_TYPE_OBJECT;
 
		doc "Moves item to other catalog. NOT IMPLEMENTED YET!"
		{
			pl "Przenosi element do innego katalogu. NIE ZAIMPLEMENTOWANA!";
		};

		__PARAMS_HEADER
		param new_catalog 
		{
			type CO_INTEGER;
			doc "Id of destination catalog"
			{
				pl "Identyfikator katalogu, w ktrym ma si znale obiekt";
			};
		};
 
		declare language_c
		{
			C "char * str_id;";
		};
 
		body language_c
		{
			C 
			%%
			// we do not want to start from @
			str_id = ap_psprintf(p, "%d", new_catalog);
			@co_call(par_area, par_class, "update", par_area=>par_area, par_class=>par_class, par_id=>par_id, version=>-1, par_prop=>'"id_parent"', new_value=>str_id);
			%%
			;
		};
	}; 
    
 
  method  getFullPath
  {
     id_permission __PERMISSION_GET;
     sec_type __METHOD_SEC_TYPE_OBJECT;
 
     doc "Retrieves full path for given object. Path is represented as string similar to file system path /root/sub_catalog1/sub_catalog2/item"
     {
         pl "Zwraca pelna sciezke dla danego obiektu. Sciezka jest lancuchem znakow, jest podobna do sciezek spotykanych w systemach plikow /root/sub_catalog1/sub_catalog2/item";
     };

     __PARAMS_HEADER_WITH_CALLBACK
     param id_root 
     {
       type CO_INTEGER;
       doc "This is left end of the path (local root). If not found, then path is constructed to the catalog that has id_parent==0"
       {
          pl "Po napotkaniu tego katalogu konstruowanie sciezki jest konczone. Jesli nie zostanie znalezione, to sciezka jest konstruowana az do znalezienia katalogu z id_parent==0";
       };
     };
 
     declare language_c
     {
        C "char *path;";
     };
 
     body language_c
     {
       C
       %%
	path = ap_pstrdup(p, "");
	__SQL_GET_STRING(path, "select getpath('%s','%s','name','/',%d,%d)",par_area, par_class, id_root,par_id)
        __ASSERT(ok, "Error during getting path for: par_area: '%s', par_class: '%s', id_root: %d, par_id: %d)",par_area, par_class, id_root,par_id)
	__CREATE_JS_FROM_STRING(path)
	__RETURN_STRING(path)

       %%
       ;
     };
  };


  method  getTree
  {
     id_permission __PERMISSION_GET;
     sec_type __METHOD_SEC_TYPE_OBJECT;

     doc "This method retrieves string with whole tree of objects"
     {
        pl "Ta metoda zwraca string z calym drzewem obiektow";
     };
 
     __PARAMS_HEADER_WITH_CALLBACK
     param prop_name 
     {
        type CO_STRING;
        doc "Name of the prop to be put in tree string"
        {
           pl "Nazwa prop, ktore bedzie wykorzystane do zbudowania stringu z drzewem";
        };
     };
     param deli 
     {
        type CO_STRING; 
        doc "String used to separate objects taken from one container -- 'leaves'"
        {
           pl "String uzywany do odseparowania obiektow z jednego kontenera -- 'lisci'";
        };
        optional 1; 
        default 'deli=ap_pstrdup(p,"|");';  
     };
     param group_deli 
     {
        type CO_STRING; 
        doc "String used to separate containers -- 'branches'"
        {
           pl "String uzywany do odseparowania kontenerow -- 'galezi'";
        };
        optional 1; 
        default 'group_deli=ap_pstrdup(p,"");';  
     };
 
     declare language_c
     {
        C 'char *tree=NULL;';
     };
 
     body language_c
     {
       C
       %%
	__SQL_GET_STRING(tree, "select gettree('%s'::text,'%s'::text,'%s'::text,'%s'::text,'%s'::text,%d::int)",
                                            par_area, par_class, prop_name, deli, group_deli, par_id)
	__CREATE_JS_FROM_STRING(tree)
	__RETURN_STRING(tree)
       %%
       ;
     };
  };

	method getParents
	{
		id_permission __PERMISSION_GET;
		sec_type __METHOD_SEC_TYPE_OBJECT;

     doc "Retrieves ids of all parents"
     {
        pl "Zwraca identyfikatory wszytskich przodkow";
     };

     __PARAMS_HEADER_WITH_CALLBACK
     param id_root 
     {
       type CO_INTEGER;
       doc "This is left end of the path (local root). If not found, then path is constructed to the catalog that has id_parent==0"
       {
          pl "Po napotkaniu tego katalogu konstruowanie sciezki jest konczone. Jesli nie zostanie znalezione, to sciezka jest konstruowana az do znalezienia katalogu z id_parent==0";
       };
     };

     declare language_c
     {
        C "char * parents;";
     };
 
     body language_c
     {
       C
       %%
	__SQL_GET_STRING(parents, "select getpath('%s','%s','id','/',%d,%d)",par_area, par_class, id_root,par_id)
        __ASSERT(ok, "Error during getting parents for: par_area: '%s', par_class: '%s', id_root: %d, par_id: %d)",par_area, par_class, id_root,par_id)
	__CREATE_JS_FROM_STRING(parents)
	__RETURN_STRING(parents)
      %%
       ;
     };
  };

	method getIdByPath
	{
		id_permission __PERMISSION_GET;
		sec_type __METHOD_SEC_TYPE_OBJECT;

		doc "Finds object by path. If not found returns -1."
		{
			pl "Szuka obiektu wskazanego przez sciezke. Jeslinie znajdzie o zwraca -1.";
		};

		__PARAMS_HEADER_WITH_CALLBACK

		param id_root 
		{
			type CO_INTEGER;
			doc "This is left end of the path (local root) when we start search."
			{
				pl "Katalog startowy, gdzie zaczynamy przeszukiwanie.";
			};
		};

		param path 
		{
			type CO_STRING;
			doc "Path, when searched object is p.a: 'catalog/sub_catalog/object_name'"
			{
				pl "Sciezka wskazujaca obiekt n.p.: 'catalog/sub_catalog/object_name'";
			};
		};


		declare language_c
		{
			C "int id_found=-1;";
			C "co_array_of_pointers * arrPath; ";
			C "int num_elems = 0 ; ";
			C "int i=0;";
			C "int start=0; ";
		};
 
		body language_c
		{
			C
			%%
				arrPath = coa_Create(cnx);
				num_elems = coa_AppendSplit(cnx, arrPath, path, "/");
				start = 0;
				if(path[0]=='/') start=1;
				for(i=start; i<num_elems; i++)
				{
					char * cur_name = coa_GetElement(cnx, arrPath, i);
					__SQL_GET_INTEGER(id_found, "SELECT id FROM co_%s_%s WHERE id_parent = %d AND name = '%s' AND deleted=0", par_area, par_class, id_root,cur_name)
					if(!ok) {id_found=-1; break; }
					id_root = id_found;
				}
				__CREATE_JS_FROM_INTEGER(id_found)
				__RETURN_INTEGER(id_found)
			%%
			;
		};
	}; // method getIdByPath


  method getDocumentTypes
  {
     id_permission __PERMISSION_GET;
     sec_type __METHOD_SEC_TYPE_CLASS;

     doc "Retrieves all document types alowed to be stored in container of current class"
     {
        pl "Zwraca liste typow dokumentow, ktore moga byc skladowane w kontenerze z biezacej klasy";
     };

     __PARAMS_HEADER_WITH_CALLBACK
 
     declare language_c
     {
        C "int c;";
     };
 
     body language_c
     {
       C
       %%
       __SEND_MESG_DEBUG("entering getDocumentTypes")
       @co_filterByProp(par_area,'"catalog_type"','"*"', '"parent_class"',par_class);
       __CREATE_JS 
       __RETURN_RESULT
       %%
       ;
     };
  };

  method insertCatalogItem
  {
     id_permission __PERMISSION_INSERT_CHILD;
     sec_type __METHOD_SEC_TYPE_OBJECT;

     doc "This method inserts object in container. It first creates object in class corresponding to id_type, then creates catalog item and sets relation between both"
     {
        pl "Ta metoda tworzy nowy obiekt w katalogu. Najpier tworzy nowy obiekt w klasie, ktora odpowiada id_type, a nastepnie tworzy element katalogu";
     };
 
     __PARAMS_HEADER_WITH_CALLBACK
 
     param name
     {
         type CO_STRING;
         doc "this is the name of the item";
     };
 
    param description
     {
         type CO_STRING;
         optional 1;
         default "description = ap_pstrdup(p, \"\");";
         doc "this is the description of the item";
     };

    param id_type
     {
         type CO_INTEGER;
         doc "this is the id of the type of the item being created";
     };


     param create_params
     {
         type CO_ARRAY_OF_STRINGS;
         doc "this is array of additional parameters that we can pass to target class to create object";
     };
         
     declare language_c
     {
        C "int id_new;";
        C "int i;";
        C "const char * str_id_type;"; 
        C "const char * str_id_user;"; 
        C "const char * str_id_item;";
        C "int    id_item;";
        C "const char * type_class_name;";

     };
 
     body language_c
     {
       C
       %%
       __SEND_MESG_DEBUG("entering insertCatalogItem");
	@co_call(par_area,par_class,"insertChild",par_area=>par_area,par_class=>par_class,par_id=>par_id,child_class=>par_class,child_prop=>'"id_parent"',foreign_prop=>'"id"');
       __GET_RETURNED_INTEGER(id_new)

       @co_call(par_area,par_class,"update",par_area=>par_area,par_class=>par_class,par_id=>id_new,par_prop=>'"name"',new_value=>name,version=>0);
       @co_call(par_area,par_class,"update",par_area=>par_area,par_class=>par_class,par_id=>id_new,par_prop=>'"description"',new_value=>description,version=>1);

       __SEND_MESG_DEBUG("decription added")
       
       str_id_type = ap_psprintf(p, "%d",id_type);
       __SEND_MESG_DEBUG("id_type prepared")
       @co_call(par_area,par_class,"update",par_area=>par_area,par_class=>par_class,par_id=>id_new,par_prop=>'"id_catalog_type"',new_value=>str_id_type,version=>2);
 
       __SEND_MESG_DEBUG("catalog_type added")
       
       str_id_user = ap_psprintf(p, "%d",cnx->id_user);
       @co_call(par_area,par_class,"update",par_area=>par_area,par_class=>par_class,par_id=>id_new,par_prop=>'"id_user"',new_value=>str_id_user,version=>3);

       __SEND_MESG_DEBUG("id_user added")
       
       @co_getProp(par_area, '"catalog_type"', id_type, '"class_name"');
       
       __SEND_MESG_DEBUG("catalog_type received")
       
       type_class_name = get_str;
       
       __SEND_MESG_DEBUG("substitution made")

/*       
       for(i=0; create_params[i] != NULL; i++)
       {
          cou_sendMessageX(cnx, create_params[i], CO_MSG_DEBUG);
       };
*/

       // we do not ask to create object if this is current catalog object -- par_class
       // we get id_item from function create and put into
       if(strcmp(type_class_name,par_class)!=0)
       {        
         @co_call(par_area, type_class_name, "create", par_area=>par_area, par_class=>type_class_name, par_id=>-1, create_params => create_params);
	 __GET_RETURNED_INTEGER( id_item )
         str_id_item = ap_psprintf(p, "%d",id_item);
         @co_call(par_area,par_class,"update",par_area=>par_area,par_class=>par_class,par_id=>id_new,par_prop=>'"id_item"',new_value=>str_id_item,version=>4);
         @co_call(par_area,type_class_name,"cloneACL",par_area=>par_area,par_class=>type_class_name,par_id=>id_item,area_parent=>par_area,class_parent=>par_class,id_parent=>id_new);
       }
       else
       {
         int NParams, i;
         str_id_item = ap_psprintf(p, "%d",id_new);
         @co_call(par_area,par_class,"update",par_area=>par_area,par_class=>par_class,par_id=>id_new,par_prop=>'"id_item"',new_value=>str_id_item,version=>4);
         NParams = coa_GetElementCount(cnx, create_params);
         for(i=0; i<NParams; i+=2)
         {
            char * prop_name = (char *) coa_GetElement(cnx, create_params, i);
            char * prop_val  = (char *) coa_GetElement(cnx, create_params, i+1);
            @co_call(par_area,par_class,"update",par_area=>par_area,par_class=>par_class,par_id=>id_new,par_prop=>prop_name,new_value=>prop_val,version=>-1);
         }

       }
       __RETURN_INTEGER(id_new)
       %%
       ;
     };
  };

  method unfoldCatalog
  {
     id_permission __PERMISSION_GET_RELATED;
     sec_type __METHOD_SEC_TYPE_OBJECT;

     doc "This method retrieves all items in current catalog that current user is allowed to see (__PERMISSION_GET)."
     {
        pl "Ta metoda zwraca wszytskie obiekty w danym katalogu, ktore moga byc ogladane (__PERMISSION_GET) przez biezacego uzytkownika";
     };
 
     __PARAMS_HEADER_WITH_CALLBACK

     param filter_type
     {
         type CO_INTEGER;
         doc " 0 - all items not deleted, 1 - only containers not deleted, 2 - only deleted items";
     };
 
     param sort_prop
     {
         type CO_STRING;
         optional 1;
         default "sort_prop = ap_psprintf(p, \"name\");";
         doc "this is the name of property to sort values";
     };
         
     declare language_c
     {
        C "int id_new;";
        C "char * str_id_type;"; 
     };
 
     body language_c
     {
       C
       %%
       // all not deleted
       if(filter_type == 0)
       {
         __CREATE_QUERY("SELECT DISTINCT c.*, t.class_name as type_class_name, t.container as type_container, u.name as creator "
                           "FROM co_%s_%s c, co_%s_catalog_type t , co_sec_user u , co_sec_acl a "
                           "WHERE c.id_catalog_type = t.id AND c.id_parent=%d AND u.id = c.id_user "
                           " AND a.id_permission = %d AND a.area = '%s' AND a.class_name = '%s' AND a.id_object = c.id AND a.id_role = %d"
                           " AND c.deleted = 0 "
                           "ORDER BY %s ",
                        par_area, par_class, par_area, par_id, __PERMISSION_GET, par_area, par_class, cnx->id_role, sort_prop)
       }
       // containers not deleted
       if(filter_type == 1) 
       {
         __CREATE_QUERY("SELECT DISTINCT c.*, t.class_name as type_class_name, t.container as type_container "
                           "FROM co_%s_%s c, co_%s_catalog_type t, co_sec_user u , co_sec_acl a "
                           "WHERE c.id_catalog_type = t.id AND c.id_parent=%d AND u.id = c.id_user "
                           " AND a.id_permission = %d AND a.area = '%s' AND a.class_name = '%s' AND a.id_object = c.id AND a.id_role = %d"

                           " AND c.deleted = 0 "
                           "AND t.container = 1 "                         
                           "ORDER BY %s ",
                       par_area, par_class, par_area, par_id, __PERMISSION_GET, par_area, par_class, cnx->id_role, sort_prop)
       }
       // all deleted
       if(filter_type == 2) 
       {
         __CREATE_QUERY("SELECT DISTINCT c.*, t.class_name as type_class_name, t.container as type_container "
                           "FROM co_%s_%s c, co_%s_catalog_type t, co_sec_user u , co_sec_acl a "
                           "WHERE c.id_catalog_type = t.id AND c.id_parent=%d AND u.id = c.id_user "
                           " AND c.deleted = 1 "
                           " AND a.id_permission = %d AND a.area = '%s' AND a.class_name = '%s' AND a.id_object = c.id  AND a.id_role = %d"
                           "ORDER BY %s ",
                       par_area, par_class, par_area, par_id, __PERMISSION_GET, par_area, par_class, cnx->id_role, sort_prop)
       }
  
      __GET_RESULT_AND_CREATE_JS
       %%
       ;
     };
  };

  method getItemData
  {
     id_permission __PERMISSION_GET;
     sec_type __METHOD_SEC_TYPE_OBJECT;

     doc "This method retrieves data for current item with catalog_type name and creator name"
     {
        pl "Ta metoda zwraca dane biezacego obiektu razem z nazwa katalogu i uzytkownika, ktory utworzyl obiekt";
     };
 
     __PARAMS_HEADER_WITH_CALLBACK
 
     body language_c
     {
       C
       %%
       __CREATE_QUERY("SELECT DISTINCT c.*, t.class_name as type_class_name, t.container as type_container, u.name as creator "
                           "FROM co_%s_%s c, co_%s_catalog_type t , co_sec_user u "
                           "WHERE c.id_catalog_type = t.id AND c.id=%d AND u.id = c.id_user " ,
                        par_area, par_class, par_area, par_id)
      __GET_RESULT_AND_CREATE_JS
       %%
       ;
     };
  };

  method delete
  {
     doc "This method is setting deleted flag to 1 in catalog (proxy) object and original object"
     {
           pl "Ta metoda ustawia flage deleted na 1 w obiekcie katalogu i w obiekcie wskazanym przez katalog";
     };

     id_permission __PERMISSION_DELETE;
     sec_type __METHOD_SEC_TYPE_OBJECT;
 
     __PARAMS_HEADER_WITH_VERSION

     declare language_c
     {
        C 'const char * target_class = "";';
        C 'int target_id = 0;';
     };

     body language_c
     {
       C
       %%
       // first we delete catalog (proxy) object
       @co_call(par_area, '"generic_acl"', "delete", par_area=>par_area, par_class=>par_class, par_id=>par_id, version=>version);

       // next we delete object pointed by catalog object
       @co_lcall(par_area, par_class, "getItemData", par_area=>par_area, par_class=>par_class, par_id=>par_id, callback=>'""');
       __GET_RETURNED_RECORDSET(corec)
       target_class = cor_GetValueByName(cnx, corec, 0, "type_class_name");

       // we do not delete again if object points to itself 
       if(strcmp(target_class, par_class) != 0)
       {
           target_id = atoi(cor_GetValueByName(cnx, corec, 0, "id_item"));
           @co_call(par_area, target_class, "delete", par_area=>par_area, par_class=>target_class, par_id=>target_id, version=>-1);
       }
      %%
       ;
     };
  };

  method undelete
  {
     doc "This method is setting deleted flag to 0 in catalog (proxy) object and original object"
     {
           pl "Ta metoda ustawia flage deleted na 0 w obiekcie katalogu i w obiekcie wskazanym przez katalog";
     };

     id_permission __PERMISSION_UNDELETE;
     sec_type __METHOD_SEC_TYPE_OBJECT;
 
     __PARAMS_HEADER_WITH_VERSION

     declare language_c
     {
        C 'const char * target_class = "";';
        C 'int target_id = 0;';
     };

     body language_c
     {
       C
       %%
       // first we undelete catalog (proxy) object
       @co_call(par_area, '"generic_acl"', "undelete", par_area=>par_area, par_class=>par_class, par_id=>par_id, version=>version);

       // next we undelete object pointed by catalog object
       @co_lcall(par_area, par_class, "getItemData", par_area=>par_area, par_class=>par_class, par_id=>par_id, callback=>'""');
       __GET_RETURNED_RECORDSET(corec)
       target_class = cor_GetValueByName(cnx, corec, 0, "type_class_name");
 
      // we do not delete again if object points to itself 
       if(strcmp(target_class, par_class) != 0)
       {
            target_id = atoi(cor_GetValueByName(cnx, corec, 0, "id_item"));
           @co_call(par_area, target_class, "undelete", par_area=>par_area, par_class=>target_class, par_id=>target_id, version=>-1);
       }
      %%
       ;
     };
  };

  method purge
  {
     doc "This method removes catalog object and original object permamently"
     {
           pl "Ta metoda w sposob trwaly usuwa obiekt katalogu i obiekt wskazany przez katalog";
     };

     id_permission __PERMISSION_PURGE;
     sec_type __METHOD_SEC_TYPE_OBJECT;
 
     __PARAMS_HEADER_WITH_VERSION

     declare language_c
     {
        C 'const char * target_class = "";';
        C 'int target_id = 0;';
     };

     body language_c
     {
       C
       %%
       // first we delete catalog (proxy) object
       @co_call(par_area, '"generic_acl"', "purge", par_area=>par_area, par_class=>par_class, par_id=>par_id, version=>version);

       // next we delete object pointed by catalog object
       @co_lcall(par_area, par_class, "getItemData", par_area=>par_area, par_class=>par_class, par_id=>par_id, callback=>'""');
       __GET_RETURNED_RECORDSET(corec)
       target_class = cor_GetValueByName(cnx, corec, 0, "type_class_name");
 
      // we do not delete again if object points to itself 
       if(strcmp(target_class, par_class) != 0)
       {
           target_id = atoi(cor_GetValueByName(cnx, corec, 0, "id_item"));
           @co_call(par_area, target_class, "purge", par_area=>par_area, par_class=>target_class, par_id=>target_id, version=>-1);
       }
      %%
       ;
     };
  };

  method permissionOn
  {
    doc "This method changes acls for catalog object and original object pointed by catalog"
    {
       pl "Ta metoda w zmienia acl'e dla obiektu katalogu i obiektu wskazanego przez katalog";
    };

    id_permission __PERMISSION_CHANGE_ACL;
    sec_type __METHOD_SEC_TYPE_OBJECT;
 
    __PARAMS_HEADER_WITH_CALLBACK

    param id_permission 
         { 
           type CO_INTEGER; 
           doc "this is the id of permission that we wish to set on for current object";
         };
  
    param id_role
         { 
           type CO_INTEGER; 
           doc "this is the id of the role that we want to grant permission";
         };
 
     declare language_c
     {
        C 'const char * target_class = "";';
        C 'int target_id = 0;';
     };

     body language_c
     {
       C
       %%
       // first we change acl for catalog (proxy) object
       @co_call(par_area,'"generic_acl"',"permissionOn",par_area=>par_area,par_class=>par_class,par_id=>par_id,callback=>callback,id_permission=>id_permission,id_role=>id_role);

       // next we change acl for object pointed by catalog object
       @co_lcall(par_area, par_class, "getItemData", par_area=>par_area, par_class=>par_class, par_id=>par_id, callback=>'""');
       __GET_RETURNED_RECORDSET(corec)
       target_class = cor_GetValueByName(cnx, corec, 0, "type_class_name");

       // we do not delete again if object points to itself 
       if(strcmp(target_class, par_class) != 0)
       {
           target_id = atoi(cor_GetValueByName(cnx, corec, 0, "id_item"));
           @co_lcall(par_area, target_class,"permissionOn",par_area=>par_area,par_class=>target_class,par_id=>target_id,callback=>'""',id_permission=>id_permission,id_role=>id_role);


       }
      %%
       ;
     };
  };

  method permissionOff
  {
     doc "This method changes acls for catalog object and original object pointed by catalog"
     {
           pl "Ta metoda w zmienia acl'e dla obiektu katalogu i obiektu wskazanego przez katalog";
     };

     id_permission __PERMISSION_CHANGE_ACL;
     sec_type __METHOD_SEC_TYPE_OBJECT;
 
    __PARAMS_HEADER_WITH_CALLBACK

    param id_permission 
         { 
           type CO_INTEGER; 
           doc "this is the id of permission that we wish to set on for current object";
         };
  
    param id_role
         { 
           type CO_INTEGER; 
           doc "this is the id of the role that we want to grant permission";
         };
 
     declare language_c
     {
        C 'const char * target_class = "";';
        C 'int target_id = 0;';
     };

     body language_c
     {
       C
       %%
       // first we change acl for catalog (proxy) object
       @co_call(par_area,'"generic_acl"',"permissionOff",par_area=>par_area,par_class=>par_class,par_id=>par_id,callback=>callback,id_permission=>id_permission,id_role=>id_role);

       // next we change acl for object pointed by catalog object
       @co_lcall(par_area, par_class, "getItemData", par_area=>par_area, par_class=>par_class, par_id=>par_id, callback=>'""');
       __GET_RETURNED_RECORDSET(corec)
       target_class = cor_GetValueByName(cnx, corec, 0, "type_class_name");

       // we do not delete again if object points to itself 
       if(strcmp(target_class, par_class) != 0)
       {
           target_id = atoi(cor_GetValueByName(cnx, corec, 0, "id_item"));
           @co_lcall(par_area, target_class,"permissionOff",par_area=>par_area,par_class=>target_class,par_id=>target_id,callback=>'""',id_permission=>id_permission,id_role=>id_role);

       }
      %%
       ;
     };
  };


   method showCatalog
   {
 
     id_permission __PERMISSION_GET;
     sec_type __METHOD_SEC_TYPE_OBJECT;
 
     __PARAMS_HEADER
     param root_name 
     {
        doc "in this parameter we pass the name of the catalog from which we start browsing";
        type CO_STRING;
        optional 1;
        default 'root_name = ap_pstrdup(p, "");';
     };
 
     declare language_c
      {
        C "FILE *fh;";
        C "int c;";
        C "const char* root_path;";
        C "long len;";
        C "int user_id=1;";
      };
 
     body language_c
      {
        C 
	%%
	root_path = cou_getCORootPath( cnx );
        __SEND_MESG_DEBUG(root_path)
        fh = ap_pfopen(p, ap_pstrcat( p, root_path, "/forms/", "catalog", "_head.html", NULL), "r");
        __ASSERT(fh, "Head file not found")
        __SEND_MESG_DEBUG("sending head")
        ap_rputs("<HEAD>\n", r);
        ap_kill_timeout(r);
 
        len = ap_send_fd( fh, r ); /* returns total of bytes sent */
        __SEND_MESG_DEBUG("Head sent: %ld bytes", len)
 
        ap_pfclose(p,fh);
 
 
        __SEND_MESG_DEBUG("sending script")
        ap_rputs("<SCRIPT>\n", r);
 
        ap_rprintf( r, "var  id_session = %d;\n", cou_assignNewSession( cnx ) );
        ap_rprintf( r, "var  par_id = %d;\n", par_id );
        ap_rprintf( r, "var  par_class=\'%s\';\n", par_class );
        ap_rprintf( r, "var  par_area = \'%s\';\n", par_area );
        ap_rprintf( r, "var  root_name = \'%s\';\n", root_name );
        ap_rprintf( r, "var  debug_level = %d;\n", cnx->debug_level );

        ap_rputs("</SCRIPT>\n", r);
        ap_rputs("</HEAD>\n", r);
        fh = ap_pfopen(p, ap_pstrcat( p, root_path, "/forms/", "catalog", "_body.html", NULL), "r");
        __ASSERT(fh,"Body file not found")
        __SEND_MESG_DEBUG("sending body")
        ap_rputs("<BODY>\n", r);
        len = ap_send_fd( fh, r ); /* returns total of bytes sent */
 
        if (r->connection->aborted) ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, "connection aborted");
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, "bytes_send=%ld", r->bytes_sent);
 
        ap_soft_timeout((char *) "finish request", r);
 
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, "Body sent: %ld bytes", len);
 
        ap_pfclose(p,fh);
        ap_rputs("</BODY>\n", r);
 
        ap_rflush(r);
	%%
	;
    };
  };                                                                           
};

//END OF CLASS CATALOG
