// TYPES FOR CO

#include "../../include/types.t"
#include "../../include/defines.t"

#define STANDARD_ID \
     default "0"; \
     params ''; \
     representant 1; \
     foreign ''; \
     foreign_prop ''; \
     position 200;

// CLASS expression_cell

class expression_cell
{
  inherits generic;

  uses "spreadsheet_cell";

#include "../spreadsheet_cell/spreadsheet_acl.t"


  container_class "spreadsheet";
  container_column_name "id_spreadsheet";

 // PROPERTIES
   area sys;

index scrid
{
 unique 1;
 column id_spreadsheet;
 column id_column;
 column id_row;
 column index_ref;
};

prop version
{
  type CO_INTEGER;
  default 0;
  params "READONLY=1";
  representant 1;
  foreign '';
  foreign_prop '';
  position 0;
};

prop deleted
{
  type CO_INTEGER;
  default 0;
  params "READONLY=1";
  representant 1;
  foreign '';
  foreign_prop '';
  position 0;
};

   prop id_spreadsheet
   {
     type CO_INTEGER;
     STANDARD_ID
   };

   prop id_column
   {
     type CO_INTEGER;
     STANDARD_ID
   };

   prop id_row
   {
     type CO_INTEGER;
     STANDARD_ID
   };

   prop index_ref
   {
     type CO_INTEGER;
     default "0";
     params '';
     representant 1;
     foreign '';
     foreign_prop '';
     position 200;
   };

   prop id_ref_spreadsheet
   {
     type CO_INTEGER;
     default "0";
     params '';
     representant 1;
     foreign '';
     foreign_prop '';
     position 200;
   };

   prop id_ref_row
   {
     type CO_INTEGER;
     default "0";
     params '';
     representant 1;
     foreign '';
     foreign_prop '';
     position 200;
   };

   prop id_ref_column
   {
     type CO_INTEGER;
     default "0";
     params '';
     representant 1;
     foreign '';
     foreign_prop '';
     position 200;
   };

   prop is_ref_row_hard
   {
     type CO_BOOL;
     default "false";
     params '';
     representant 1;
     foreign '';
     foreign_prop '';
     position 200;
   };

   prop is_ref_column_hard
   {
     type CO_BOOL;
     default "false";
     params '';
     representant 1;
     foreign '';
     foreign_prop '';
     position 200;
   };

   prop is_ref_row_named
   {
     type CO_BOOL;
     default "false";
     params '';
     representant 1;
     foreign '';
     foreign_prop '';
     position 200;
   };

   prop is_ref_column_named
   {
     type CO_BOOL;
     default "false";
     params '';
     representant 1;
     foreign '';
     foreign_prop '';
     position 200;
   };

// SQL FUNCTIONS
   function_sql fillV
   {
     doc "copy range references (gor given cell) down to given row";
     params '(int,int,int,int)';
     returns float4;
   S
   %%
   @"
DECLARE
    _id_spreadsheet ALIAS FOR \$1;
    _id_column ALIAS FOR \$2;
    _id_row ALIAS FOR \$3;
    _id_row_last ALIAS FOR \$4;
    _nr_row_first INTEGER;
    _nr_row_last INTEGER;
    _reference co_sys_expression_cell\%ROWTYPE;
BEGIN
    SELECT row_number INTO _nr_row_first FROM co_sys_spreadsheet_row
        WHERE
        id = _id_row
    ;
    SELECT row_number INTO _nr_row_last FROM co_sys_spreadsheet_row
        WHERE
        id = _id_row_last
    ;
    DELETE FROM co_sys_expression_cell
        WHERE EXISTS(
            SELECT * FROM co_sys_spreadsheet_row  AS ROW
                WHERE
                ROW.id = id_row AND
                ROW.row_number > _nr_row_first AND
                ROW.row_number <= _nr_row_last
            )
	    AND id_column = _id_column
    ;
    /* referred column id never change; ref. row id may change */
    INSERT INTO co_sys_expression_cell
        SELECT
        RANGE.version,
        RANGE.deleted,
        RANGE.id_spreadsheet,
        RANGE.id_column,
        ROW.id,
        RANGE.index_ref,
        RANGE.id_ref_spreadsheet,
        NEWREFROW.id,/* update as you go; depends on is_ref_row_hard */
        RANGE.id_ref_column,
        RANGE.is_ref_row_hard,
        RANGE.is_ref_column_hard,
        RANGE.is_ref_row_named,
        RANGE.is_ref_column_named

        FROM
            co_sys_expression_cell AS RANGE,
            co_sys_spreadsheet_row AS ROW,
            co_sys_spreadsheet_row AS REFROW,
            co_sys_spreadsheet_row AS NEWREFROW
        WHERE
            RANGE.id_spreadsheet = _id_spreadsheet AND
            RANGE.id_column = _id_column AND
            RANGE.id_row = _id_row AND
            ROW.spreadsheet_id = _id_spreadsheet AND
            ROW.row_number > _nr_row_first AND
            ROW.row_number <= _nr_row_last AND
            REFROW.id = RANGE.id_ref_row AND
            NEWREFROW.spreadsheet_id = _id_spreadsheet AND
            (
            (NEWREFROW.row_number = REFROW.row_number + (ROW.row_number - _nr_row_first)
                AND NOT RANGE.is_ref_row_hard) OR (
                NEWREFROW.id = REFROW.id AND RANGE.is_ref_row_hard)
            )
    ;
RETURN 0;

END;
    "
   %%
   ;
   };



 // METHODS
#include "../expression/expression_acl.t"

  method getReferencedValue
  {
    id_permission __PERMISSION_GET;

//    param id_spreadsheet { type CO_INTEGER; };
    __PARAMS_HEADER
    param id_column { type CO_INTEGER; };
    param id_row { type CO_INTEGER; };
    param index { type CO_INTEGER; };

    declare language_c
    {
      C 'int nTuple;';
      C 'int id_spreadsheet;';
    };

    body language_c
    {
     C
     %%
     id_spreadsheet = par_id;
          snprintf(SQLString, SQLBUFSIZE, "
SELECT value 
FROM co_sys_spreadsheet_cell AS CELL, co_sys_expression_cell AS REF
WHERE 
 REF.deleted = 0 AND REF.id_spreadsheet=%d AND
 REF.id_column=%d AND REF.id_row=%d AND REF.index_ref = %d AND
 CELL.row_id = REF.id_ref_row AND CELL.column_id = REF.id_ref_column AND
 CELL.deleted = 0 ;
",	id_spreadsheet, id_column, id_row, index);

cou_logError( cnx, CO_LOG_DEBUG, "getReferencedValue(%d,%d,%d,%d)",id_spreadsheet, id_column, id_row, index);
cou_logError( cnx, CO_LOG_DEBUG, "getReferencedValue 1st done");
          __SQL_GET_RESULT

nTuple = cor_GetTupleCount(cnx, corec);

cou_logError( cnx, CO_LOG_DEBUG, "getReferencedValue 2nd done, ntuples=%d",
	nTuple);

    if (nTuple == 0) {
	  cou_setReturnValue( cnx, ap_pstrdup(p, ""), CO_STRING );
    } else if (nTuple == 1) {
	  cou_setReturnValue( cnx, cor_GetCopyOfValue(cnx, corec,0,0), CO_STRING );
    } else {
	  cou_error( cnx, "ntuples not in 0..1 in getReferencedValue");
    }
cou_logError( cnx, CO_LOG_DEBUG, "getReferencedValue 3rd done");

     %%
     ;
    };
  };

  method setReferencedCell
  {
    id_permission __PERMISSION_UPDATE;

//    param id_spreadsheet { type CO_INTEGER; };
    __PARAMS_HEADER
    param id_column { type CO_INTEGER; };
    param id_row { type CO_INTEGER; };
    param index { type CO_INTEGER; };
    param id_ref_spreadsheet { type CO_INTEGER; };
    param id_ref_column { type CO_INTEGER; };
    param id_ref_row { type CO_INTEGER; };
    param ref_column_hard { type CO_INTEGER; };
    param ref_row_hard { type CO_INTEGER; };
    param ref_column_named { type CO_INTEGER; };
    param ref_row_named { type CO_INTEGER; };

    declare language_c
    {
     C
     %%
     int id_spreadsheet;
     %%
     ;
    };

    body language_c
    {
     C
     %%
     id_spreadsheet = par_id;
     cou_logError( cnx, CO_LOG_DEBUG, "setReferencedCell called");

        snprintf(SQLString, SQLBUFSIZE, "
UPDATE co_sys_expression_cell
     SET 
       id_ref_spreadsheet = %d ,
       id_ref_column = %d ,
       id_ref_row = %d ,
       is_ref_column_hard = (%d<>0) ,
       is_ref_row_hard = (%d<>0) ,
       is_ref_column_named = (%d<>0) ,
       is_ref_row_named = (%d<>0) ,
       deleted = 0
WHERE id_spreadsheet=%d AND id_column=%d AND id_row=%d AND index_ref = %d ;
", id_ref_spreadsheet, id_ref_column, id_ref_row, ref_column_hard, ref_row_hard,
 ref_column_named, ref_row_named,
 id_spreadsheet, id_column, id_row,index);

          __SQL_EXEC

cou_logError( cnx, CO_LOG_DEBUG, "setReferencedCell 1st done");

	  if (ok==0) { /* new reference */
		snprintf(SQLString, SQLBUFSIZE, "
INSERT INTO co_sys_expression_cell
 (id_spreadsheet,id_column,id_row, index_ref, id_ref_spreadsheet, id_ref_column, id_ref_row, is_ref_column_hard, is_ref_row_hard, is_ref_column_named, is_ref_row_named)
VALUES(%d,%d,%d,%d,%d,%d,%d,(%d<>0),(%d<>0),(%d<>0),(%d<>0)) ;
", id_spreadsheet,id_column,id_row,index,id_ref_spreadsheet,id_ref_column, id_ref_row, ref_column_hard, ref_row_hard,ref_column_named, ref_row_named);

cou_logError( cnx, CO_LOG_DEBUG, "setReferencedCell(%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d)",
id_spreadsheet,id_column,id_row,index,id_ref_spreadsheet,id_ref_column, id_ref_row, ref_column_hard, ref_row_hard,ref_column_named, ref_row_named
);
	        __SQL_EXEC

cou_logError( cnx, CO_LOG_DEBUG, "setReferencedCell 2nd done");

	  }

          {
             const char* class_names[] = {"spreadsheet","spreadsheet_column","spreadsheet_row",NULL};
             int object_ids[] = {id_spreadsheet,id_column,id_row};
            char prop_name[32]; /* 2*id(10dig) + 1*index+2*'_' */
            char prop_value[32]; /* 2*id + 2*bool+3*',' */
                snprintf(prop_name, 32, "p%d_%d_%d", id_column, id_row,index);
                snprintf(prop_value, 32, "%d,%d,%d,%d,%d,%d", id_ref_column, id_ref_row, ref_column_hard, ref_row_hard, ref_column_named, ref_row_named);

             cou_queueActionX( cnx, "expression_cell", class_names, object_ids,CO_OP_UPDATE, prop_name, prop_value, NULL );
          }

     %%
     ;
    };
  };

  method purgeExtraReferences
  {
    id_permission __PERMISSION_PURGE;
    doc "Remove range references for given cell and index>= index_from ";
    doc "by default - all references for given cell.";
    doc "Generates actions.";

//    param id_spreadsheet { type CO_INTEGER; };
    __PARAMS_HEADER
    param id_column { type CO_INTEGER; };
    param id_row { type CO_INTEGER; };
    param index_from { type CO_INTEGER; optional 1; default "index_from=0"; };

    declare language_c
    {
     C
     %%
     int id_spreadsheet;
     %%
     ;
    };

    body language_c
    {
     C
     %%
     id_spreadsheet = par_id;
          __CREATE_QUERY( "
DELETE FROM co_sys_expression_cell
WHERE id_spreadsheet=%d AND id_column=%d AND id_row=%d AND index_ref>=%d;
",
          id_spreadsheet,id_column,id_row, index_from)

          __SQL_EXEC

          {
            const char* class_names[] = {"spreadsheet","spreadsheet_column","spreadsheet_row",NULL};
            int object_ids[] = {id_spreadsheet,id_column,id_row};

            cou_queueActionX( cnx, "expression_cell", class_names, object_ids,
                CO_OP_DELETE, NULL );
          }

     %%
     ;
    };
  };

  method fillDownCellRef
  {
    id_permission 3;

    /* original cell */
//    param id_spreadsheet { type CO_INTEGER; };
    __PARAMS_HEADER
    param id_column { type CO_INTEGER; };
    param id_row { type CO_INTEGER; };
    /* fill down until row... */
    param id_last_row { type CO_INTEGER; };

    declare language_c
    {
      C
      %%
	int num_first_row=0, num_last_row=0;
	int ntuples, i;
        int id_spreadsheet;
      %%
      ;
    };

    body language_c
    {
     C
     %%
     id_spreadsheet = par_id;
        __CREATE_QUERY("
SELECT co_sys_expression_cell_fillV(%d,%d,%d,%d);
",      id_spreadsheet, id_column, id_row, id_last_row)

        __SQL_GET_RESULT
        __CLEAR_RESULT

// TODO: send actions
        __SQL_GET_INTEGER(num_first_row,
            "SELECT row_number FROM co_sys_spreadsheet_row WHERE id=%d",
            id_row )
        __SQL_GET_INTEGER(num_last_row,
            "SELECT row_number FROM co_sys_spreadsheet_row WHERE id=%d",
            id_last_row )

        __CREATE_QUERY("
SELECT 
id_row, index_ref,
id_ref_column, id_ref_row, is_ref_column_hard, is_ref_row_hard, is_ref_column_named, is_ref_row_named
FROM co_sys_expression_cell AS CELL, co_sys_spreadsheet_row AS ROW
WHERE CELL.id_spreadsheet=%d AND CELL.id_column=%d AND 
CELL.id_row = ROW.id AND ROW.row_number>%d AND ROW.row_number<=%d
",      id_spreadsheet, id_column, 
	num_first_row, num_last_row)

        __SQL_GET_RESULT
	ntuples = cor_GetTupleCount(cnx, corec);
        {
    	    const char* class_names[] = {"spreadsheet","spreadsheet_column","spreadsheet_row",NULL};
            int object_ids[] = {id_spreadsheet,id_column,id_row};
	    int id_row_current;
            char prop_name[32]; /* 2*id(10dig) + 1*index+2*'_' */
            char prop_value[32]; /* 2*id + 2*bool+3*',' */
	    for (i=0; i<ntuples; i++) {
		id_row_current = atoi(cor_GetValue(cnx, corec, i, 0));
		object_ids[2] = id_row_current;
        	snprintf(prop_name, 32, "p%d_%d_%s", id_column, id_row_current, cor_GetValue(cnx, corec, i, 1));
        	snprintf(prop_value, 32, "%s,%s,%s,%s,%s,%s", 
		cor_GetValue(cnx, corec, i, 2),cor_GetValue(cnx, corec, i, 3),cor_GetValue(cnx, corec, i, 4),
		cor_GetValue(cnx, corec, i, 5),cor_GetValue(cnx, corec, i, 6),cor_GetValue(cnx, corec, i, 7));

        	cou_queueActionX( cnx, "expression_cell", class_names, object_ids,CO_OP_UPDATE, prop_name, prop_value, NULL );
	    }
        }
        __CLEAR_RESULT

     %%
     ;
    };
  };

};