// TYPES FOR CO

#include "../../include/types.t"
#include "../../include/defines.t"

#define STANDARD_ID \
     default "0"; \
     params ''; \
     representant 1; \
     foreign ''; \
     foreign_prop ''; \
     position 200;

// CLASS expression

class expression_range
{
  inherits generic;

  uses "spreadsheet_cell";

#include "../spreadsheet_cell/spreadsheet_acl.t"

  container_class "spreadsheet";
  container_column_name "id_spreadsheet";

  area sys;

index scrid
{
 unique 1;
 column id_spreadsheet;
 column id_column;
 column id_row;
 column index_ref;
};

 // PROPERTIES
prop version
{
  type CO_INTEGER;
  default 0;
  params "READONLY=1";
  representant 1;
  foreign '';
  foreign_prop '';
  position 0;
};

prop deleted
{
  type CO_INTEGER;
  default 0;
  params "READONLY=1";
  representant 1;
  foreign '';
  foreign_prop '';
  position 0;
};

   prop id_spreadsheet
   {
     type CO_INTEGER;
     STANDARD_ID
     identifier id_spreadsheet;
   };

   prop id_column
   {
     type CO_INTEGER;
     STANDARD_ID
     identifier id_column;
   };

   prop id_row
   {
     type CO_INTEGER;
     STANDARD_ID
     identifier id_row;
   };

   prop index_ref
   {
     type CO_INTEGER;
     default "0";
     params '';
     representant 1;
     foreign '';
     foreign_prop '';
     position 200;
   };

   prop function
   {
     type CO_INTEGER;
     default "0";
     params '';
     representant 1;
     foreign '';
     foreign_prop '';
     position 200;
   };

   prop id_ref_spreadsheet
   {
     type CO_INTEGER;
     default "0";
     params '';
     representant 1;
     foreign '';
     foreign_prop '';
     position 200;
   };

   prop id_ref_row
   {
     type CO_INTEGER;
     default "0";
     params '';
     representant 1;
     foreign '';
     foreign_prop '';
     position 200;
   };

   prop id_ref_column
   {
     type CO_INTEGER;
     default "0";
     params '';
     representant 1;
     foreign '';
     foreign_prop '';
     position 200;
   };

   prop is_ref_row_hard
   {
     type CO_BOOL;
     default "false";
     params '';
     representant 1;
     foreign '';
     foreign_prop '';
     position 200;
   };

   prop is_ref_column_hard
   {
     type CO_BOOL;
     default "false";
     params '';
     representant 1;
     foreign '';
     foreign_prop '';
     position 200;
   };

   prop is_ref_row_named
   {
     type CO_BOOL;
     default "false";
     params '';
     representant 1;
     foreign '';
     foreign_prop '';
     position 200;
   };

   prop is_ref_column_named
   {
     type CO_BOOL;
     default "false";
     params '';
     representant 1;
     foreign '';
     foreign_prop '';
     position 200;
   };

   prop id_ref_row2
   {
     type CO_INTEGER;
     default "0";
     params '';
     representant 1;
     foreign '';
     foreign_prop '';
     position 200;
   };

   prop id_ref_column2
   {
     type CO_INTEGER;
     default "0";
     params '';
     representant 1;
     foreign '';
     foreign_prop '';
     position 200;
   };

   prop is_ref_row2_hard
   {
     type CO_BOOL;
     default "false";
     params '';
     representant 1;
     foreign '';
     foreign_prop '';
     position 200;
   };

   prop is_ref_column2_hard
   {
     type CO_BOOL;
     default "false";
     params '';
     representant 1;
     foreign '';
     foreign_prop '';
     position 200;
   };

   prop is_ref_row2_named
   {
     type CO_BOOL;
     default "false";
     params '';
     representant 1;
     foreign '';
     foreign_prop '';
     position 200;
   };

   prop is_ref_column2_named
   {
     type CO_BOOL;
     default "false";
     params '';
     representant 1;
     foreign '';
     foreign_prop '';
     position 200;
   };


 // METHODS
   function_sql getNVal
   {
     doc "get range-dependent numeric value referenced in the expression";
     params '(int,int,int,int)';
     returns float4;
   S
   %%
   @"
DECLARE
    _id_spreadsheet ALIAS FOR \$1;
    _id_column ALIAS FOR \$2;
    _id_row ALIAS FOR \$3;
    _index ALIAS FOR \$4;
    _reference co_sys_expression_range\%ROWTYPE;
    _coordinates RECORD;
    _coordinates2 RECORD;
    _result float4;
BEGIN
    SELECT * INTO _reference FROM co_sys_expression_range
        WHERE
        id_spreadsheet = _id_spreadsheet AND
        id_column = _id_column AND
        id_row = _id_row AND
        index_ref = _index
    ;
    SELECT COL.column_number, ROW.row_number INTO _coordinates
    FROM co_sys_spreadsheet_column AS COL,co_sys_spreadsheet_row AS ROW
        WHERE
        COL.id = _reference.id_ref_column AND
        ROW.id = _reference.id_ref_row
    ;
    -- for sum only
    IF NOT FOUND THEN
	RETURN 0;
    END IF;

    SELECT COL.column_number, ROW.row_number INTO _coordinates2
    FROM co_sys_spreadsheet_column AS COL,co_sys_spreadsheet_row AS ROW
        WHERE
        COL.id = _reference.id_ref_column2 AND
        ROW.id = _reference.id_ref_row2
    ;
    -- for sum only
    IF NOT FOUND THEN
	RETURN 0;
    END IF;

    SELECT SUM(CELL.value_num) INTO _result
    FROM co_sys_spreadsheet_cell AS CELL,
        co_sys_spreadsheet_column AS COL,
        co_sys_spreadsheet_row AS ROW
    WHERE
        COL.spreadsheet_id = _reference.id_ref_spreadsheet AND
        COL.column_number >= _coordinates.column_number AND
        COL.column_number <= _coordinates2.column_number AND
        ROW.spreadsheet_id = _reference.id_ref_spreadsheet AND
        ROW.row_number >= _coordinates.row_number AND
        ROW.row_number <= _coordinates2.row_number AND
        CELL.column_id = COL.id AND
        CELL.row_id = ROW.id
    ;
RETURN _result;

END;
    "
   %%
   ;
   };

   function_sql replRow
   {
     doc "replace reference to given id_row with another id_new_row";
     params '(int,int,int,int)';
     returns int;
   S
   %%
   @"
DECLARE
    _id_spreadsheet ALIAS FOR \$1;
    _id_row ALIAS FOR \$2;
    _id_prev_row ALIAS FOR \$3;
    _id_next_row ALIAS FOR \$4;
BEGIN
    UPDATE co_sys_expression_range
    SET id_ref_row = _id_next_row
        WHERE
        id_spreadsheet = _id_spreadsheet AND
	id_ref_row = _id_row
    ;

    UPDATE co_sys_expression_range
    SET id_ref_row2 = _id_prev_row
        WHERE
        id_spreadsheet = _id_spreadsheet AND
	id_ref_row2 = _id_row
    ;
RETURN 0;

END;
    "
   %%
   ;
   };

   function_sql fillV
   {
     doc "copy range references (gor given cell) down to given row";
     params '(int,int,int,int)';
     returns float4;
   S
   %%
   @"
DECLARE
    _id_spreadsheet ALIAS FOR \$1;
    _id_column ALIAS FOR \$2;
    _id_row ALIAS FOR \$3;
    _id_row_last ALIAS FOR \$4;
    _nr_row_first INTEGER;
    _nr_row_last INTEGER;
    _reference co_sys_expression_range\%ROWTYPE;
BEGIN
    SELECT row_number INTO _nr_row_first FROM co_sys_spreadsheet_row
        WHERE
        id = _id_row
    ;
    SELECT row_number INTO _nr_row_last FROM co_sys_spreadsheet_row
        WHERE
        id = _id_row_last
    ;
    DELETE FROM co_sys_expression_range
        WHERE EXISTS(
            SELECT * FROM co_sys_spreadsheet_row  AS ROW
                WHERE
                ROW.id = id_row AND
                ROW.row_number > _nr_row_first AND
                ROW.row_number <= _nr_row_last
            )
    ;
    /* referred column id never change; ref. row id may change */
    INSERT INTO co_sys_expression_range
        SELECT
        RANGE.version,
        RANGE.deleted,
        RANGE.id_spreadsheet,
        RANGE.id_column,
        ROW.id,
        RANGE.index_ref,
        RANGE.function,
        RANGE.id_ref_spreadsheet,
        NEWREFROW.id,/* update as you go; depends on is_ref_row_hard */
        RANGE.id_ref_column,
        RANGE.is_ref_row_hard,
        RANGE.is_ref_column_hard,
        RANGE.is_ref_row_named,
        RANGE.is_ref_column_named,
        NEWREFROW2.id,/* update as you go; depends on is_ref_row2_hard */
        RANGE.id_ref_column2,
        RANGE.is_ref_row2_hard,
        RANGE.is_ref_column2_hard,
        RANGE.is_ref_row2_named,
        RANGE.is_ref_column2_named

        FROM
            co_sys_expression_range AS RANGE,
            co_sys_spreadsheet_row AS ROW,
            co_sys_spreadsheet_row AS REFROW,
            co_sys_spreadsheet_row AS REFROW2,
            co_sys_spreadsheet_row AS NEWREFROW,
            co_sys_spreadsheet_row AS NEWREFROW2
        WHERE
            RANGE.id_spreadsheet = _id_spreadsheet AND
            RANGE.id_column = _id_column AND
            RANGE.id_row = _id_row AND
            ROW.spreadsheet_id = _id_spreadsheet AND
            ROW.row_number > _nr_row_first AND
            ROW.row_number <= _nr_row_last AND
            REFROW.id = RANGE.id_ref_row AND
            REFROW2.id = RANGE.id_ref_row2 AND
            NEWREFROW.spreadsheet_id = _id_spreadsheet AND
            (
            (NEWREFROW.row_number = REFROW.row_number + (ROW.row_number - _nr_row_first)
                AND NOT RANGE.is_ref_row_hard) OR (
                NEWREFROW.id = REFROW.id AND RANGE.is_ref_row_hard)
            ) AND
            NEWREFROW2.spreadsheet_id = _id_spreadsheet AND
            (
            (NEWREFROW2.row_number = REFROW2.row_number + (ROW.row_number - _nr_row_first)
                AND NOT RANGE.is_ref_row2_hard) OR (
                NEWREFROW2.id = REFROW2.id AND RANGE.is_ref_row2_hard)
            )
    ;
RETURN 0;

END;
    "
   %%
   ;
   };


#include "../expression/expression_acl.t"

  method getReferencedValue
  {
    id_permission __PERMISSION_GET;

//    param id_spreadsheet { type CO_INTEGER; };
    __PARAMS_HEADER
    param id_column { type CO_INTEGER; };
    param id_row { type CO_INTEGER; };
    param index { type CO_INTEGER; };

    declare language_c
    {
       C 'int nTuples;';
       C 'int id_spreadsheet;';
    };

    body language_c
    {
     C
     %%
     id_spreadsheet = par_id;
	snprintf(SQLString, SQLBUFSIZE, "
SELECT co_sys_expression_range_getNVal(%d,%d,%d,%d);
",      id_spreadsheet, id_column, id_row, index);

cou_logError( cnx, CO_LOG_DEBUG, "expression_range::getReferencedValue(%d,%d,%d,%d)",id_spreadsheet, id_column, id_row, index);
cou_logError( cnx, CO_LOG_DEBUG, "getReferencedValue 1st done");
	__SQL_GET_RESULT

	nTuples = cor_GetTupleCount(cnx, corec);

cou_logError( cnx, CO_LOG_DEBUG, "getReferencedValue 2nd done, ntuples=%d",
        nTuples);

	if (nTuples != 1)
    	    cou_error( cnx, "ntuples <> 1 in getReferencedValue");

	cou_setReturnValue( cnx, cor_GetCopyOfValue(cnx, corec,0,0), CO_STRING );

cou_logError( cnx, CO_LOG_DEBUG, "getReferencedValue 3rd done");
	__CLEAR_RESULT

     %%
     ;
    };
  };

  method setReferencedRange
  {
    id_permission __PERMISSION_UPDATE;

//    param id_spreadsheet { type CO_INTEGER; };
    __PARAMS_HEADER
    param id_column { type CO_INTEGER; };
    param id_row { type CO_INTEGER; };
    param index { type CO_INTEGER; };
    param function { type CO_INTEGER; };
    param id_ref_spreadsheet { type CO_INTEGER; };
    param id_ref_column { type CO_INTEGER; };
    param id_ref_row { type CO_INTEGER; };
    param ref_column_hard { type CO_INTEGER; optional 1; default "ref_column_hard=0"; };
    param ref_row_hard { type CO_INTEGER; optional 1; default "ref_row_hard=0"; };
    param ref_column_named { type CO_INTEGER; optional 1; default "ref_column_named=0"; };
    param ref_row_named { type CO_INTEGER; optional 1; default "ref_row_named=0"; };
    param id_ref_column2 { type CO_INTEGER; };
    param id_ref_row2 { type CO_INTEGER; };
    param ref_column2_hard { type CO_INTEGER; optional 1; default "ref_column2_hard=0"; };
    param ref_row2_hard { type CO_INTEGER; optional 1; default "ref_row2_hard=0"; };
    param ref_column2_named { type CO_INTEGER; optional 1; default "ref_column2_named=0"; };
    param ref_row2_named { type CO_INTEGER; optional 1; default "ref_row2_named=0"; };

    declare language_c
    {
     C
     %%
     int id_spreadsheet;
     %%
     ;
    };

    body language_c
    {
     C
     %%
     id_spreadsheet = par_id;
cou_logError( cnx, CO_LOG_DEBUG, "setReferencedRange called");

        snprintf(SQLString, SQLBUFSIZE, "
UPDATE co_sys_expression_range
     SET index_ref = %d ,
       function = %d,
       id_ref_spreadsheet = %d ,
       id_ref_column = %d ,
       id_ref_row = %d ,
       is_ref_column_hard = (%d<>0) ,
       is_ref_row_hard = (%d<>0) ,
       is_ref_column_named = (%d<>0) ,
       is_ref_row_named = (%d<>0) ,
       id_ref_column2 = %d ,
       id_ref_row2 = %d ,
       is_ref_column2_hard = (%d<>0) ,
       is_ref_row2_hard = (%d<>0) ,
       is_ref_column2_named = (%d<>0) ,
       is_ref_row2_named = (%d<>0) 
WHERE id_spreadsheet=%d AND id_column=%d AND id_row=%d
",index, function, id_ref_spreadsheet,
 id_ref_column, id_ref_row, ref_column_hard, ref_row_hard, ref_column_named, ref_row_named,
 id_ref_column2, id_ref_row2, ref_column2_hard, ref_row2_hard, ref_column2_named, ref_row2_named,
 id_spreadsheet, id_column, id_row);

          __SQL_EXEC

cou_logError( cnx, CO_LOG_DEBUG, "setReferencedRange 1st done");

	  if (!ok) { /* new reference */
		snprintf(SQLString, SQLBUFSIZE, "
INSERT INTO co_sys_expression_range
 (id_spreadsheet,id_column,id_row, index_ref, function,
id_ref_spreadsheet,
id_ref_column, id_ref_row, is_ref_column_hard, is_ref_row_hard, is_ref_column_named, is_ref_row_named,
id_ref_column2, id_ref_row2, is_ref_column2_hard, is_ref_row2_hard, is_ref_column2_named, is_ref_row2_named) 
VALUES(%d,%d,%d,%d,%d,
%d,
%d,%d,(%d<>0),(%d<>0),(%d<>0),(%d<>0),
%d,%d,(%d<>0),(%d<>0),(%d<>0),(%d<>0))
", id_spreadsheet,id_column, id_row, index, function,
id_ref_spreadsheet,
id_ref_column, id_ref_row, ref_column_hard, ref_row_hard, ref_column_named, ref_row_named,
id_ref_column2, id_ref_row2, ref_column2_hard, ref_row2_hard, ref_column2_named, ref_row2_named);

	        __SQL_EXEC

cou_logError( cnx, CO_LOG_DEBUG, "setReferencedRange 2nd done");

	  }
          {
            const char* class_names[] = {"spreadsheet","spreadsheet_column","spreadsheet_row",NULL};
            int object_ids[] = {id_spreadsheet,id_column,id_row};
            char prop_name[32]; /* 2*id(10dig) + 1*index+2*'_' */
            char prop_value[64]; /* 2*id + 2*bool+3*',' */

            snprintf(prop_name, 32, "p%d_%d_%d", id_column, id_row,index);
            snprintf(prop_value, 64, "%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d",
	    id_ref_column, id_ref_row, ref_column_hard, ref_row_hard,ref_column_named, ref_row_named,
	    id_ref_column2, id_ref_row2, ref_column2_hard, ref_row2_hard, ref_column2_named, ref_row2_named,
	    function);

            cou_queueActionX( cnx, "expression_range", class_names, object_ids, CO_OP_UPDATE, prop_name, prop_value, NULL );
          }

     %%
     ;
    };
  };

  method purgeExtraReferences
  {
    id_permission __PERMISSION_PURGE;
    doc "Remove range references for given cell and index>= index_from ";
    doc "by default - all references for given cell.";
    doc "Generates actions.";

//    param id_spreadsheet { type CO_INTEGER; };
    __PARAMS_HEADER
    param id_column { type CO_INTEGER; };
    param id_row { type CO_INTEGER; };
    param index_from { type CO_INTEGER; optional 1; default "index_from=0"; };

    declare language_c
    {
     C
     %%
     int id_spreadsheet;
     %%
     ;
    };

    body language_c
    {
     C
     %%
     id_spreadsheet = par_id;
          __CREATE_QUERY( "
DELETE FROM co_sys_expression_range
WHERE id_spreadsheet=%d AND id_column=%d AND id_row=%d AND index_ref>=%d;
",
          id_spreadsheet,id_column,id_row, index_from)

          __SQL_EXEC

          {
            const char* class_names[] = {"spreadsheet","spreadsheet_column","spreadsheet_row",NULL};
            int object_ids[] = {id_spreadsheet,id_column,id_row};

            cou_queueActionX( cnx, "expression_range", class_names, object_ids,
                CO_OP_DELETE, NULL );
          }

     %%
     ;
    };
  };

  method resetReferencedRow
  {
    doc "all references to the row with id=id_row reset to row whose number is num_row";
    doc "WARNING: This method assumes that only one row may be deleted at a time.";

    id_permission 3;

    /* acl owner */
//    param id_spreadsheet { type CO_INTEGER; };
    __PARAMS_HEADER
    /* which row was removed */
    param id_row { type CO_INTEGER; };
    /* use row whose number is... */
    param num_row { type CO_INTEGER; };

    declare language_c
    {
      C "int id_row_prev=0;";
      C "int id_row_next=0;";
      C "int ntuples;";
      C "int id_spreadsheet;";
    };

    body language_c
    {
     C
     %%
     id_spreadsheet = par_id;
          __SQL_GET_INTEGER(id_row_prev,
"SELECT id FROM co_sys_spreadsheet_row 
WHERE spreadsheet_id=%d 
 AND row_number=%d
;",
	id_spreadsheet, num_row-1)

          __SQL_GET_INTEGER(id_row_next,
"SELECT id FROM co_sys_spreadsheet_row 
WHERE spreadsheet_id=%d 
 AND row_number=%d
;",
	id_spreadsheet, num_row)

          __CREATE_QUERY("
SELECT co_sys_expression_range_replRow(%d,%d,%d,%d);
",      id_spreadsheet, id_row, id_row_prev, id_row_next)

          __SQL_GET_RESULT
          __CLEAR_RESULT

// TODO: send actions
// TODO: evaluate dependent expressions
// find all expressions depending on this row
// evaluate them
// send actions for them

     %%
     ;
    };
  };


  method fillDownRangeRef
  {
    id_permission 3;

    /* original cell */
//    param id_spreadsheet { type CO_INTEGER; };
    __PARAMS_HEADER
    param id_column { type CO_INTEGER; };
    param id_row { type CO_INTEGER; };
    /* fill down until row... */
    param id_last_row { type CO_INTEGER; };

    declare language_c
    {
      C
      %%
        int num_first_row=0, num_last_row=0; /* checked later */
        int ntuples, i;
        int id_spreadsheet;
      %%
      ;
    };

    body language_c
    {
     C
     %%
     id_spreadsheet = par_id;
          __CREATE_QUERY("
SELECT co_sys_expression_range_fillV(%d,%d,%d,%d);
",      id_spreadsheet, id_column, id_row, id_last_row)

          __SQL_GET_RESULT
          __CLEAR_RESULT

// TODO: send actions
        __SQL_GET_INTEGER(num_first_row,
            "SELECT row_number FROM co_sys_spreadsheet_row WHERE id=%d",
            id_row )
	__ASSERT(num_first_row>0, "row with id=%d not found", id_row)
        __SQL_GET_INTEGER(num_last_row,
            "SELECT row_number FROM co_sys_spreadsheet_row WHERE id=%d",
            id_last_row )
	__ASSERT(num_last_row>0, "row with id=%d not found", id_last_row)

        __CREATE_QUERY("
SELECT
id_row, index_ref,
id_ref_column, id_ref_row, is_ref_column_hard, is_ref_row_hard,
id_ref_column2, id_ref_row2, is_ref_column2_hard, is_ref_row2_hard,
function
FROM co_sys_expression_range AS RANGE, co_sys_spreadsheet_row AS ROW
WHERE RANGE.id_spreadsheet=%d AND RANGE.id_column=%d AND
RANGE.id_row = ROW.id AND ROW.row_number>%d AND ROW.row_number<=%d
",      id_spreadsheet, id_column,
        num_first_row, num_last_row)

        __SQL_GET_RESULT
        ntuples = cor_GetTupleCount(cnx, corec);
        {
            const char* class_names[] = {"spreadsheet","spreadsheet_column","spreadsheet_row",NULL};
            int object_ids[] = {id_spreadsheet,id_column,id_row};
            int id_row_current;
            char prop_name[32]; /* 2*id(10dig) + 1*index+2*'_' */
            char prop_value[32]; /* 2*id + 2*bool+3*',' */
            for (i=0; i<ntuples; i++) {
                id_row_current = atoi(cor_GetValue(cnx, corec, i, 0));
                object_ids[2] = id_row_current;
                snprintf(prop_name, 32, "p%d_%d_%s", id_column, id_row_current,
		    cor_GetValue(cnx, corec, i, 1));
                snprintf(prop_value, 32, "%s,%s,%s,%s,%s,%s,%s,%s,%s",
                cor_GetValue(cnx, corec, i, 2),cor_GetValue(cnx, corec, i, 3),cor_GetValue(cnx, corec, i, 4),
                cor_GetValue(cnx, corec, i, 5),cor_GetValue(cnx, corec, i, 6),cor_GetValue(cnx, corec, i, 7),
                cor_GetValue(cnx, corec, i, 8),cor_GetValue(cnx, corec, i, 9),cor_GetValue(cnx, corec, i, 10));

                cou_queueActionX( cnx, "expression_range", class_names, object_ids,CO_OP_UPDATE, prop_name, prop_value, NULL );
            }
        }
        __CLEAR_RESULT

     %%
     ;
    };
  };


};