
#include "types.t"
#include "defines.t"

class generic_iacl
{
   area sec;
   area sys;
   area usr;

   container_class '';
   container_column_name '';   
    
   inherits generic;

#include "../../include/check_acl.t"

#include "generic_iacl_props.t" 

   include_c "../coffice/co_broker.h";

   method check_acl
   {
     doc "This is very important method that checks acl. Whole security system depends on this method!";
     id_permission 0;

     __PARAMS_HEADER

     declare language_c
     {
       C
       %%
        const char **container_info;
	const char *container_class;
	const char *container_column_name;
	int container_id = -1;
       %%
       ;
     };

     body language_c
     {
       C
       %%
       cou_logError(cnx, CO_LOG_DEBUG, 
               "check_acl in area '%s': class='%s' id_permission=%d par_id=%d \n", par_area, par_class, cnx->id_permission, par_id
		               );  
       // First we take definitions for this class
       // we can not use other methods becouse this would fire security checks and
       // create the infinite loop
       
        // THIS MUST BE OPTIMISED!! 
        container_info = cou_getContainerInfo( cnx, par_area, par_class);
        container_class = container_info[0];
	container_column_name = container_info[1];
	// END TO OPTIMISATION---OC
	
       // get container element id
	if (par_id != -1) {
	    __SQL_GET_INTEGER(container_id,
		"SELECT %s FROM co_%s_%s WHERE id=%d",
		container_column_name, par_area, par_class, par_id)
	}
	else {
	    int retval = 0;
	    __RETURN_INTEGER(retval);
	    return 0;
	}
//	else {
//	    container_id = -1; /* so that selects with id=-1 are possible */
//	}

       // call check_acl for container
       @co_call(par_area,container_class,"check_acl",par_area=>par_area,par_class=>container_class,par_id=>container_id);
       %%
      ;
     };
   };      

   method check_acl_class
   {
     doc "This is very method that checks acl using container class.";
     doc "You will want to overwrite this method if you have non-standard ACL checking.";
     id_permission 0;

     __PARAMS_HEADER_CLASS

     declare language_c
     {
       C
       %%
        const char **container_info;
        const char *container_class;
       %%
       ;
     };

     body language_c
     {
       C
       %%
       __SEND_MESG_DEBUG("check_acl_class in area '%s': class='%s' id_permission=%d ", par_area, par_class, cnx->id_permission)

        container_info = cou_getContainerInfo( cnx, par_area, par_class);
        container_class = container_info[0];

        @co_call(par_area, container_class, "check_acl_class", \
            par_area=>par_area,par_class=>container_class );

       %%
	;
      };
    }; // method

  method insertChild
  {
     doc "This method is overloaded of grneric::insertChild. In this method we add copying of acl from parent";
     id_permission __PERMISSION_INSERT_CHILD;
     
    __PARAMS_HEADER    
    param child_class 
         { 
           type CO_STRING; 
           doc "this is the name of class tu insert record in";
         };
    param child_prop 
        { 
           type CO_STRING;
           doc "this is the property name in child class that holds id of parent";
        };
    param foreign_prop 
       { 
           type CO_STRING; 
           doc "this is the prop in container (this:) class that holds id to copy to child class";
       };
 
     declare language_c
     {
       C "int security_type;" ;
       C "int id_new;"; 
     };  
       
     body language_c
     {
       C
       %%
         // we switch off security for this one method! be carefull!
         cnx->check_security = 0;
         @co_call(par_area,'"generic"',"insertChild",par_area=>par_area,par_class=>par_class,par_id=>par_id,child_class=>child_class,child_prop=>child_prop, foreign_prop=>foreign_prop);
         cnx->check_security = 1;
         // security switched on again

         id_new = *(int *) cou_getReturnValue(cnx, CO_INTEGER, &err);
         __CREATE_QUERY("INSERT INTO co_sec_acl 
                         SELECT NEXTVAL('co_sec_acl_sqid'), version, deleted, id_permission, %d ,id_role, area, '%s'
                         FROM co_sec_acl
                         WHERE area = '%s' AND class_name = '%s' AND id_object = %d ",
                             id_new, child_class, par_area, par_class, par_id );
         __SQL_EXEC
         cou_setReturnValue(cnx, &id_new, CO_INTEGER);         
       %%
       ;
     };
  };		   

  method cloneACL
  {
     doc "This method clones all ACLs from given object (used p.e. by catalog)";
     id_permission __PERMISSION_INSERT_CHILD;
     
    __PARAMS_HEADER

    param area_parent
         {
           type CO_STRING; 
           doc "this is the name of class to insert record in";
         };
 
    param class_parent 
         { 
           type CO_STRING; 
           doc "this is the name of class tu insert record in";
         };
    param id_parent 
        { 
           type CO_INTEGER;
           doc "this is the property name in child class that holds id of parent";
        };

     body language_c
     {
       C
       %%
         __CREATE_QUERY("INSERT INTO co_sec_acl 
                         SELECT NEXTVAL('co_sec_acl_sqid'), version, deleted, id_permission, %d ,id_role, '%s', '%s'
                         FROM co_sec_acl
                         WHERE area = '%s' AND class_name = '%s' AND id_object = %d ",
                         par_id, par_area, par_class, area_parent, class_parent, id_parent);
         __SQL_EXEC
       %%
       ;
     };
 };// end of cloneACL

 method permissionOn
  {
     doc "This method sets one ACL On";
     id_permission __PERMISSION_CHANGE_ACL;

    __PARAMS_HEADER    

    param id_permission 
         { 
           type CO_INTEGER; 
           doc "this is the id of permission that we wish to set on for current object";
         };

    param id_role
         { 
           type CO_INTEGER; 
           doc "this is the id of the role that we want to grant permission";
         };

     body language_c
     {
       C
       %%
         // WE SHOULD CHECK WHETER WE CAN CONTROL THIS ROLE !!!!

         __CREATE_QUERY("INSERT INTO co_sec_acl (id_permission, id_role, area, class_name, id_object)
                         VALUES ( %d, %d, '%s', '%s', %d)",
                           id_permission, id_role, par_area, par_class, par_id);
 
         __SQL_EXEC
       %%
       ;
     };
 }; // end of permissionOn


  method create
  {
     doc "this method creates empty object and sets ACL to allow changing of acls to active role";

     id_permission __PERMISSION_CREATE;

     __PARAMS_HEADER_WITH_CALLBACK

     param create_params
     {
         type CO_ARRAY_OF_STRINGS;
         doc "this is array of additional parameters that we can pass to target class to create object";
     };

   declare language_c
    {
      C 
      %%
        int new_id=0;
	int i;
	int nFields;
        char* puta_classes[2]={NULL,NULL};
	int puta_ids[2]={0,0}; // for queueActionX 
	char **props;
      %%
      ;	
    };  

     body language_c
     {
         C 
         %%
      __CREATE_QUERY("INSERT INTO co_%s_%s (deleted) VALUES (0)", 
                           par_area, par_class)

      __SQL_EXEC

      __CREATE_QUERY("SELECT last_value FROM co_%s_%s_sqid", 
                           par_area, par_class)
      __SQL_GET_RESULT
      new_id= atoi(cor_GetValue(cnx, corec, 0, 0));
      __CLEAR_RESULT

      puta_classes[0] = (char *) par_class;
      puta_ids[0] = new_id;

      cou_queueActionX( cnx, par_class, (const char **) puta_classes, puta_ids, CO_OP_CREATE, NULL);

      // getting current values

      __CREATE_QUERY("SELECT * FROM co_%s_%s WHERE id = %d",
                           par_area, par_class, new_id)

      __SQL_GET_RESULT

      puta_classes[0] = (char *) par_class;
      puta_ids[0] = new_id;

      nFields = cor_GetTupleCount(cnx, corec);
      props = ap_palloc(p, 2 * nFields * sizeof(char*) + 1);
      
      for (i = 0; i < nFields; i++)
      {
	 props[2*i]   = ap_pstrdup(p, cor_GetFieldName(cnx, corec, i));
	 props[2*i+1] = ap_pstrdup(p, cor_GetValue(cnx, corec, 0, i));
      } 
      props[2*i]=NULL;
      
      //cou_sendMessage(cnx, "we have props");
      
      __CLEAR_RESULT
      cou_queueActionXX( cnx, par_class, (const char **) puta_classes, puta_ids, CO_OP_UPDATE, (const char **) props);
      cou_setReturnValue(cnx, &new_id, CO_INTEGER);

      //setting ACL for current role
      __CREATE_QUERY("INSERT INTO co_sec_acl (id_permission, id_role, area, class_name, id_object)
                      VALUES ( __PERMISSION_CREATE, %d, '%s', '%s', %d)",
                      cnx->id_role, par_area, par_class, new_id);

         %%
         ;
     };
  }; // end of create

 
};  // end of class generic_iacl