#include "../../include/types.t"
#include "../../include/defines.t"

// CLASS PROFILER

class profiler
{
  area sec;
  area sys;
  area usr;

  inherits 0;

  container_class '';
  container_column_name '';

#include "../../include/check_acl.t" 

 // PROPERTIES

  sequence sqid;

  prop id
  {
   type CO_INTEGER;
   default SEQUENCE { sq_name sqid; };
   params "READONLY=1";
   representant 1;
   foreign '';
   foreign_prop '';
   position 0;
  };
  
  prop begin_sec
  {
   type CO_INTEGER;
   default "NULL";
   params 'READONLY=1;';
   representant 1;
   foreign 'cevent_type';
   foreign_prop 'id';
   position 40;
  };

  prop begin_usec
  {
   type CO_INTEGER;
   default "NULL";
   params 'READONLY=1;';
   representant 1;
   foreign 'cevent_type';
   foreign_prop 'id';
   position 40;
  };

  prop end_sec
  {
   type CO_INTEGER;
   default "NULL";
   params 'READONLY=1;';
   representant 1;
   foreign 'cevent_type';
   foreign_prop 'id';
   position 41;
  };

  prop end_usec
  {
   type CO_INTEGER;
   default "NULL";
   params 'READONLY=1;';
   representant 1;
   foreign 'cevent_type';
   foreign_prop 'id';
   position 41;
  };

  prop id_meta_method
  {
   type CO_INTEGER;
   default "NULL";
   params '';
   representant 1;
   foreign 'meta_method';
   foreign_prop 'id';
   position 50;
  };

 // METHODS:

  method setStart
  {
    doc "creates profiler entry with an exact start time (microseconds accuracy)";
    doc "returns id of created entry";
    id_permission __PERMISSION_CREATE;

    __PARAMS_HEADER
    
    param par_method { type CO_STRING; };
    
    declare language_c
    {
      C
      %%
      int id_newEntry;
      struct timeval tv;
      %%
      ;
    };
    
    body language_c
    {
      C
      %%
      gettimeofday(&tv,NULL);
      __CREATE_QUERY("insert into co_%s_profiler(id_meta_method,begin_sec,begin_usec)
                      values(
		      (select id from co_%s_meta_method where method_name='%s' and class_name='%s'),
		      %ld,
		      %ld
		      )",par_area,par_area,par_method,par_class,tv.tv_sec,tv.tv_usec)
      __SQL_EXEC
      
      __SQL_GET_INTEGER(id_newEntry,"select last_value from co_%s_profiler_sqid",par_area)
      
      cou_setReturnValue(cnx, &id_newEntry, CO_INTEGER);
      %%
      ;
    };
  
  };
  
  method setEnd
  {
    doc "writes end time to the entry of given id";
    id_permission __PERMISSION_UPDATE;
  
    __PARAMS_HEADER

    param id_entry { type CO_INTEGER; };

    declare language_c
    {
      C
      %%
      struct timeval tv;
      %%
      ;
    };
  
    body language_c
    {
      C
      %%
      gettimeofday(&tv,NULL);
      __CREATE_QUERY("update co_%s_profiler set end_sec=%ld, end_usec=%ld
                      where id=%d",par_area,tv.tv_sec,tv.tv_usec,id_entry)
      __SQL_EXEC	
      %%
      ;
    };
  };

};

//END OF CLASS profiler