// TYPES FOR CO

#include "types.t"
#include "defines.t"

// CLASS spreadsheet_cell

class spreadsheet_cell
{
  area sys;

#include "spreadsheet_acl.t"
#include "generic_props.t"

  uses "clipboard";
  uses "expression";
  uses "spreadsheet";

  index cr
  {
    unique 1;
    column column_id;
    column row_id;
  };

  prop column_id
  {
   type CO_INTEGER;
   default 0;
   params '';
   representant 1;
   foreign 'spreadsheet_column';
   foreign_prop 'id';
   position 44;
  };

  prop row_id
  {
   type CO_INTEGER;
   default 0;
   params '';
   representant 1;
   foreign 'spreadsheet_row';
   foreign_prop 'id';
   position 44;
  };

  prop column_number
  {
   f_body
   %%
   select column_number from co_sys_spreadsheet_column where id=($1.column_id);
   %%
   ;
   type CO_INTEGER;
   default 0;
   params '';
   representant 1;
   foreign '';
   foreign_prop '';
   position 44;
  };
  
  prop row_number
  {
   f_body
   %%
   select row_number from co_sys_spreadsheet_row where id=($1.row_id);
   %%
   ;
   type CO_INTEGER;
   default 0;
   params '';
   representant 1;
   foreign '';
   foreign_prop '';
   position 44;
  };

  prop type
  {
   type CO_STRING;
   default "'STR'";
   params '';
   representant 1;
   foreign '';
   foreign_prop '';
   position 44;
  };

  prop value
  {
   type CO_STRING;
   default "''";
   params '';
   representant 1;
   foreign '';
   foreign_prop '';
   position 44;
  };

  prop value_num
  {
   type CO_NUMERIC;
   default "0";
   params '';
   representant 1;
   foreign '';
   foreign_prop '';
   position 44;
  };

  prop value_date
  {
   type CO_DATE;
   default "'1990-01-01'::date";
   params '';
   representant 1;
   foreign '';
   foreign_prop '';
   position 44;
  };

  prop align
  {
   type CO_STRING;
   default "''";
   params '';
   representant 1;
   foreign '';
   foreign_prop '';
   position 44;
  };

  prop fontsize
  {
   type CO_STRING;
   default "''";
   params '';
   representant 1;
   foreign '';
   foreign_prop '';
   position 44;
  };

  prop textstyle
  {
   type CO_STRING;
   default "''";
   params '';
   representant 1;
   foreign '';
   foreign_prop '';
   position 44;
  };

  prop textcolor
  {
   type CO_STRING;
   default "''";
   params '';
   representant 1;
   foreign '';
   foreign_prop '';
   position 44;
  };

  prop fontfamily
  {
   type CO_STRING;
   default "''";
   params '';
   representant 1;
   foreign '';
   foreign_prop '';
   position 44;
  };

  inherits 0;

  container_class 'column';
  container_column_name 'column_id';

 // PROPERTIES

 // METHODS

  method getAttribute
  {
    doc "returns attribute of a cell";
    id_permission __PERMISSION_GET;
    
    __PARAMS_HEADER
    
    param column_id { type CO_INTEGER; };
    param row_id { type CO_INTEGER; };
    param attribute_name { type CO_STRING; };
    
    declare language_c
    {
      C
      %%
      const char *attribute_value = "";
      %%
      ;
    };
  
    body language_c
    {
      C
      %%
      __SQL_GET_STRING(attribute_value, "select %s from cov_%s_spreadsheet_cell where column_id=%d and row_id=%d",
                       attribute_name, par_area, column_id, row_id)
      __RETURN_STRING(attribute_value)
      %%
      ;
    };
  
  };
 
  method setAttribute
  {
    doc "changes cell attribute";
    id_permission __PERMISSION_UPDATE;

    __PARAMS_HEADER

    param column_id { type CO_INTEGER; };
    param row_id { type CO_INTEGER; };
    param attribute_name { type CO_STRING; };
    param attribute_value { type CO_STRING; };

    declare language_c
      {
        C
	%%
	const char* puta_classes[] = { "spreadsheet", "spreadsheet_column", "spreadsheet_row", NULL}; // for queueActionX
	int puta_ids[] = { 0, 0, 0}; // for queueActionX
	co_typein_evalDependent cparamsEval; /* eval dependent expressions */
	int bEvalDependent = 1; /* should eval all dependent expressions */
	char* end_ptr;
	double double_value;
        const char *sql_attribute_value;
	%%
	;
      };

    body language_c
      {
        C
	%%
	sql_attribute_value = cou_buildEncodedSQL(cnx, attribute_value);
	/* value_noeval is "virtual attribute" which does not cause */
	/* evaluation of dependent expressions. Used in evalDependent only */
	if (strcmp(attribute_name,"value_noeval")==0) {
	    attribute_name="value";
	    bEvalDependent=0; /* do not evaluate dependent expressions */
	}

	if (!strcmp(attribute_name,"value"))
	{
	  @co_call(par_area,'"spreadsheet"',"onEvent", \
	    par_area=>par_area, par_class=>'"spreadsheet"', par_id=>par_id, \
	    column_id=>column_id,row_id=>row_id,event_name=>'"onbeforevaluechange"');
	}
	
	/* first, just set the attribute (as string) */
	__CREATE_QUERY("UPDATE co_%s_spreadsheet_cell SET %s='%s' WHERE column_id=%d AND row_id=%d", par_area, attribute_name, sql_attribute_value, column_id, row_id);
    	__SQL_EXEC
    	if (!ok)
    	{
    	    __CREATE_QUERY("INSERT INTO co_%s_spreadsheet_cell(column_id,row_id,%s) VALUES(%d,%d,'%s')", par_area, attribute_name, column_id, row_id,  sql_attribute_value);
    	    __SQL_EXEC
    	}
	/* value is a special case: store numeric value in a separate field */
	/* and (perhaps) eval dependent expressions */
	if (!strcmp(attribute_name,"value")) {
	    /* useful constants */
	    double infinity;
	    infinity = atof("Inf");

	    /* set numeric value ?and date? see strtod() */
	    errno=0; /* clean up */
	    double_value = strtod(attribute_value, &end_ptr); /* or just use atof()*/
	    if (errno!=0) double_value=0;
	    /* more scrict test if value is numeric */
	    while (isspace(*end_ptr)) end_ptr++; 
	    if ((*end_ptr)!='\0') double_value=0; /* numeric only if whitespace till end */
	    if (double_value==infinity) cou_error(cnx, "Infinity detected");
	    //__SEND_MESG_DEBUG("value_num=%f errno=%d", double_value, errno )
//	    double_value = atof(attribute_value);
//	    if (double_value != 0)
//	    {
	      __CREATE_QUERY(
		  "UPDATE co_%s_spreadsheet_cell SET value_num='%f' "
		  "WHERE column_id=%d AND row_id=%d"
		  , par_area, double_value, column_id, row_id
	      )
    	      __SQL_EXEC
//	    }
	    if (bEvalDependent)
            {
		@co_call(par_area,'"expression"',"evalDependent",par_area=>par_area,par_class=>'"expression"',par_id=>par_id,id_column=>column_id,id_row=>row_id);
	    }
	}

	puta_ids[0] = par_id; puta_ids[1] = column_id; puta_ids[2] = row_id;
	cou_queueActionX( cnx, "spreadsheet_cell", puta_classes, puta_ids, CO_OP_UPDATE, attribute_name, attribute_value, NULL);

	if (!strcmp(attribute_name,"value"))
	{
	  @co_call(par_area,'"spreadsheet"',"onEvent", \
	    par_area=>par_area, par_class=>'"spreadsheet"', par_id=>par_id, \
	    column_id=>column_id,row_id=>row_id,event_name=>'"onaftervaluechange"');
	}
	
	%%
	;
      };
  };

  method fromClipboard
  {
    doc "retrieves cell data from clipboard";
    id_permission __PERMISSION_UPDATE;
    
    __PARAMS_HEADER

    param column_id { type CO_INTEGER; };
    param row_id { type CO_INTEGER; };

    declare language_c
    {
      C
      %%
      char *cell_value;
      %%
      ;
    };
    
    body language_c
    {
      C
      %%
	/****/
      @co_call(par_area,'"clipboard"',"retrieveClipboardObject", \
	par_area=>par_area, par_class=>'"clipboard"', par_id=>-1, \
	owner_area=>par_area, owner_class=>'"spreadsheet_cell"' );

      __GET_RETURNED_STRING(cell_value)

	  /* empty cell_value (i.e. null or empty string) means there is no object stored */
      if (cell_value && !(*cell_value=='\0')) {
        @co_call(par_area,'"spreadsheet_cell"',"setAttribute", \
	    par_area=>par_area, par_class=>'"spreadsheet_cell"', par_id=>par_id, \
	    row_id=>row_id, column_id=>column_id, \
	    attribute_value=>cell_value, \
	    attribute_name=>'"value"' );
	    return 0;
      }

      /* try expression */
        @co_call(par_area,'"expression"',"setExpressionFromClipboard", \
	    par_area=>par_area, \
	    par_class=>'"expression"', \
	    par_id=>par_id, \
	    id_row=>row_id, id_column=>column_id );

      %%
      ;
    };
  };

  method toClipboard
  {
    doc "saves cell data to clipboard";
    id_permission __PERMISSION_GET;

    __PARAMS_HEADER

    param column_id { type CO_INTEGER; };
    param row_id { type CO_INTEGER; };

    body language_c
    {
      C
      %%
      __CREATE_QUERY("
	SELECT value from co_%s_spreadsheet_cell
	WHERE column_id=%d AND row_id=%d", par_area, column_id, row_id)
      __SQL_GET_RESULT

      @co_call(par_area,'"clipboard"',"insertClipboardObject",serialization=>'cor_GetCopyOfValue(cnx, corec,0,0)', \
               par_id=>-1,par_area=>par_area,par_class=>'"clipboard"', owner_area=>par_area, owner_class=>'"spreadsheet_cell"');

      __CLEAR_RESULT

      %%
      ;
    };
  };

  method onClientEvent
  {
    doc "probably unused now";
    id_permission __PERMISSION_EVALUATE;

    __PARAMS_HEADER

    param column_id { type CO_INTEGER; };
    param row_id { type CO_INTEGER; };
    param event_name { type CO_STRING; };

    declare language_c
    {
      C
      %%
      struct co_typein_evalScript *in_params;
      char * js_temp;
      %%
      ;
    };

    body language_c
      {
        C
	%%
	__SQL_GET_STRING(js_temp,"
	SELECT script FROM co_%s_spreadsheet_cevents E,
			   co_%s_cevent_type ET
	WHERE (E.event_type_id=ET.id)
	AND (E.spreadsheet_id=%d OR E.column_id=%d OR E.row_id=%d)
	AND (ET.type='%s')
	ORDER BY E.row_id DESC,E.column_id DESC",par_area,par_area,par_id,column_id,row_id,event_name)

	// no script to execute
	%%
	;
      };
  };

  method cell_onBeforeValueChange
  {
    doc "for future use";
    id_permission __PERMISSION_EVALUATE;

    __PARAMS_HEADER
    param column_id { type CO_INTEGER; };
    param row_id { type CO_INTEGER; };
//  must have the same props as setAttribute method
    param attribute_name { type CO_STRING; };
    param attribute_value { type CO_STRING; };

    body language_c
    {
      C
      %%
      %%
      ;
    };
  };

};

//END OF CLASS spreadsheet_cell