#include "types.t"
#include "defines.t"

// CLASS text_document

class text_document
{
  area "sys";

#include "check_acl.t"

#include "generic_props.t"

  uses "file";

  include_c "http_protocol.h";
  include_c "co_session.h";
  include_c "co_mod.h";
  include_c "../coffice/mod_coffice.h";



/* mapping files into memory */
include_c "stdlib.h";
include_c "unistd.h";
include_c "sys/mman.h";
include_c "fcntl.h";

  default_form "viewTextDocument";

  prop name
  {
   type CO_STRING;
   default "''";
   params '';
   representant 0;
   foreign '';
   foreign_prop '';
   position 44;
  };

 prop file_version
  {
   type CO_INTEGER;
   default 0;
   params '';
   representant 0;
   foreign '';
   foreign_prop '';
   position 100;
  };


  inherits file;

  container_class '';
  container_column_name '';

  // PROPERTIES

  // METHODS

  method getContentType
  {
    doc "returns content-type which will be put into http header";
    id_permission __PERMISSION_GET;
    __PARAMS_HEADER

    body language_c
    {
      C
      %%
      cou_setReturnValue( cnx,
          ap_pstrdup(p,"text/plain"),
	  CO_STRING
      );
      %%
      ;
    };
  };

 local_method loc_getPathForVersion
 {
    C
    %%
	const char* loc_getPathForVersion( pool* p, const char* par_area, const char* par_class, int par_id, int version)
	{
    	    return ap_psprintf( p, "%s/%s/%d.%d", par_area, par_class, par_id, version);
	}
    %%
    ;
  };

  method getPath
  {
    doc "returns partial path to text object data file";
    id_permission __PERMISSION_GET;
    __PARAMS_HEADER
    
    declare language_c
    {
     C
     %%
     int version_number = 0;
     %%
     ; 
    };
    
    body language_c
    {
     C
     %%
     __SQL_GET_INTEGER(version_number, "SELECT file_version FROM co_%s_%s WHERE id=%d AND deleted=0", par_area, par_class, par_id)
     if (!ok) cou_error(cnx, "Cannot get version number");
     else
     cou_setReturnValue( cnx,
	 (char *) loc_getPathForVersion( p, par_area, par_class, par_id, version_number),
         CO_STRING
     );
     %%
     ;
    };
  };

  method create
  {
    doc "creates new object";
    id_permission __PERMISSION_CREATE;
    sec_type __METHOD_SEC_TYPE_CLASS;

#include "../../include/create_header.t"

    declare language_c
    {
      C
      %%
      int new_id;
      const char *name;
      char *filename;
      int fd;
      int security_state;
      const char * new_text;
      %%
      ;
    };

    body language_c
    {
      C
      %%
      if (!strcmp( coa_GetStringElement(cnx,create_params, 0),"name"))
      {
        name = cou_buildEncodedSQL(cnx, coa_GetStringElement(cnx, create_params, 1));
      }
      else name = "";
    
	  new_text = "new_text";	  
      if(coa_GetElementCount(cnx, create_params) > 3)
      {
	    if (!strcmp( coa_GetStringElement(cnx,create_params, 2),"new_text"))
    	{
        	new_text = coa_GetStringElement(cnx, create_params, 3);
		}
	  }

      @co_call(par_area,'"generic_acl"',"create",par_area=>par_area,par_class=>par_class,par_id=>-1,dont_set_acl=>dont_set_acl);

      __GET_RETURNED_INTEGER(new_id)

// should be done by update method
/*
      __CREATE_QUERY("UPDATE co_%s_%s SET version=0, name='%s' WHERE id=%d", 
	    par_area, par_class, name, new_id)
      __SQL_EXEC
*/
      security_state = cnx->check_security;
      cnx->check_security = 0;
      @co_call(par_area,par_class,'writeFileFromString',par_id=>new_id,par_area=>par_area,par_class=>par_class,text=>new_text);
      cnx->check_security = security_state;

      __RETURN_INTEGER(new_id)
      %%
      ;
    };
  };

  method viewTextDocument
 // sends text document in HTML in view mode
  {
    doc "generate client view interface";
    id_permission __PERMISSION_GET;

    __PARAMS_HEADER

    declare language_c
    {
      C
      %%
      FILE *fh;
      const char *filename;
      int fd;
      int document_version = 0;
      const char *filedata;
      %%
      ;
    };

    body language_c
    {
      C
      %%
      __SEND_MESG_DEBUG("viewTD par_id=%d",par_id)
      __SQL_GET_INTEGER(document_version,"SELECT file_version FROM co_%s_%s WHERE id=%d", par_area, par_class, par_id)

      if (!ok) // object does not exist
      {
       ap_rputs("<BODY>\n<H1>requested object does not exist</H1>\n</BODY>",r);
       return 0;
      }

      //ap_rprintf(r,"<HEAD>\n<SCRIPT>\nvar par_id=%d\nvar version_number=%d\n</SCRIPT>\n", par_id, document_version);
      filename = ap_psprintf(p, "%s/text_document/viewTextDocument.up", cou_getCORootPath(cnx));
      __PUT_ASP_FROM_FILE(filename)

      @co_call(par_area,par_class,'getFileAsString',par_id=>par_id,par_area=>par_area,par_class=>par_class);
      filedata = (const char *) cou_getReturnValue(cnx, CO_STRING, &err);

      ap_rputs( cou_buildHTMLEncodedTextDocument(cnx, filedata), r );

      filename = ap_psprintf(p, "%s/text_document/viewTextDocument.down", cou_getCORootPath(cnx));
      __PUT_HTML_FROM_FILE(filename)

      %%
      ;
    };
  };


	method writeTextAndEvalAsASP
	{
    	doc "write text of document and treat text_document as asp file and evaluate it";
	    id_permission __PERMISSION_EVALUATE;

    	__PARAMS_HEADER

		param text
		{
			type CO_STRING;
			doc "this is new text to be written for curent text_document";
		};
		
		body language_c
		{
			C
			%%
			// we dont want to start from @
			@co_call(par_area, par_class, "writeFileFromString",  par_area=>par_area, par_class=>par_class, par_id=>par_id, text=>text);
			@co_call(par_area, par_class, "evalTextDocumentAsASP",par_area=>par_area, par_class=>par_class, par_id=>par_id);
			%%
			;
		};
	};

 
  method evalTextDocumentAsASP
 // evaluates text document as it were an ASP file
  {
    doc "treat text_document as asp file and evaluate it";
    id_permission __PERMISSION_EVALUATE;

    __PARAMS_HEADER
	param args
	{
		doc "replacement parameters for the ASP page";
		type CO_STRING;
		optional 1;
		default 'args=NULL';
	};
		

    declare language_c
    {
      C
      %%
      FILE *fh;
      char *filename;
      int document_version;
	  char *method_args;
      %%
      ;
    };

    body language_c
    {
      C
      %%
      __SQL_GET_INTEGER(document_version,"SELECT file_version FROM co_%s_%s WHERE id=%d", par_area, par_class, par_id)

 //     ap_rputs("<HEAD>\n<META HTTP_EQUIV=\"Expires\" CONTENT=\"0\">\n</HEAD>\n",r);
      
      @co_call(par_area,par_class,'getPath',par_id=>par_id,par_area=>par_area,par_class=>par_class);
      filename = ap_psprintf(p, "%s/%s", cou_getCODataPath(), (char *) cou_getReturnValue( cnx, CO_STRING, &err));
      
      __SEND_MESG_DEBUG("file with script = %s", filename)
	  if (args) {
		/* fake replacement arguments */
		method_args = cnx->method_args;
		cnx->method_args = args;
		co_send_activeFromFile(cnx, filename);
		cnx->method_args = method_args;
	  } else {
		co_send_activeFromFile(cnx, filename);
	  }
      %%
      ;
    };
  };

  method getTextDocument
  {
    doc "get a copy of text_document content";
    id_permission __PERMISSION_GET;

    __PARAMS_HEADER

    declare language_c
    {
      C
      %%
      char *filename;
      int document_version;

      int fd;
      void *map;
      struct stat stat_buf;
      size_t length;
      char *out_buf;
      %%
      ;
    };

    body language_c
    {
      C
      %%
//
      @co_call(par_area,par_class,'getPath',par_id=>par_id,par_area=>par_area,par_class=>par_class);
      filename = ap_psprintf(p, "%s/%s", cou_getCODataPath(), (char *) cou_getReturnValue( cnx, CO_STRING, &err));

      __SEND_MESG_DEBUG("file with data = %s", filename)

      fd = open( filename, O_RDONLY);
      if (fd==-1) {
        __SEND_MESG_DEBUG("couldn't open file %s", filename)
	return -1; /* can't open file */
      }
      if (fstat(fd, &stat_buf)==-1) {
          __SEND_MESG_DEBUG("couldn't state file %s", filename)
  	  return -1; /* can't stat file */
      }
      length = stat_buf.st_size+1;
      out_buf = ap_pstrdup(p, (map=mmap( 0, length, PROT_READ, MAP_SHARED, fd, 0)));
      munmap(map, length);
      close(fd);

//    if (ap_pclosef(p, fd) == -1) cou_logError(cnx, CO_LOG_DEBUG, "couldn't close file %s", filename);

      
      __RETURN_STRING(out_buf)
      %%
      ;
    };
  };

  method browseTextDocumentVersions
 // displays version browser
  {
    doc "generates HTML version-browser";
    id_permission __PERMISSION_GET_HISTORY;

    __PARAMS_HEADER

    param version_number { type CO_INTEGER; };

    declare language_c
    {
      C
      %%
      FILE *fh;
      char *filename;
      int document_version;
      struct stat file_stat;
      %%
      ;
    };

    body language_c
    {
      C
      %%
      __CREATE_QUERY("SELECT file_version, name FROM co_%s_%s WHERE id=%d", par_area, par_class, par_id)
      __SQL_GET_RESULT
      document_version = atoi(cor_GetValue(cnx, corec,0,0));
      if (version_number == -1) version_number = document_version;
      ap_rprintf(r,"<HEAD>\n<SCRIPT>\nvar par_id=%d\nvar version_number=%d\n</SCRIPT>\n</HEAD>\n",par_id,version_number);
      ap_rputs("<BODY bgcolor=black>\n",r);
     
      ap_rputs("<TABLE>\n<TR>\n",r);
      ap_rputs("<TD><TEXTAREA rows=30 cols=64 READONLY=TRUE>\n",r);

//      @co_call(par_area,par_class,'getPath',par_id=>par_id,par_area=>par_area,par_class=>par_class);
//      filename = ap_psprintf(p, "%s/%s", cou_getCODataPath(), cou_getReturnValue( cnx, CO_STRING, &err));
      filename = ap_psprintf(p, "%s/%s", cou_getCODataPath(), loc_getPathForVersion( p, par_area, par_class, par_id, version_number));

      __PUT_HTML_FROM_FILE(filename)
      ap_rputs("</TEXTAREA>\n</TD>\n",r);
      
      ap_rputs("<TD valign=top>",r);
      ap_rputs("<TABLE BORDER=1 bgcolor=white>\n",r);
      ap_rprintf(r,"<TR>\n<TD bgcolor=yellow>document name</TD>\n<TD align=center><FONT color=green>%s</FONT></TD>\n",cor_GetValue(cnx, corec,0,1));

      __CLEAR_RESULT

      stat(filename,&file_stat);

      ap_rprintf(r,"<TR>\n<TD bgcolor=cyan>document length</TD>\n<TD align=center>%d bytes</TD>\n",(int) file_stat.st_size);
      ap_rprintf(r,"<TR>\n<TD bgcolor=cyan>last access</TD>\n<TD><FONT color=blue>%s</FONT></TD>\n</TR>\n",ctime(&(file_stat.st_atime)));
      ap_rprintf(r,"<TR>\n<TD bgcolor=cyan>last modification</TD>\n<TD><FONT color=blue>%s</FONT></TD>\n</TR>\n",ctime(&(file_stat.st_mtime)));
      ap_rprintf(r,"<TR>\n<TD bgcolor=cyan>last change</TD>\n<TD><FONT color=blue>%s</FONT></TD>\n</TR>\n",ctime(&(file_stat.st_ctime)));
      ap_rputs("<TR>\n<TD COLSPAN=2 align=center bgcolor=green>\n<FONT color=white>version ",r);
      if(version_number>1) ap_rprintf(r,"<INPUT type=button value='<<' onclick='location.href=\"%s?browseTextDocumentVersions&par_id=%d&version_number=%d\"'>\n",par_class,par_id,version_number-1);
      ap_rprintf(r,"<INPUT type=text size=3 value='%d' onchange='location.href=\"%s?browseTextDocumentVersions&par_id=%d&version_number=\"+this.value'>\n",version_number, par_class, par_id);
      if(version_number<document_version) ap_rprintf(r,"<INPUT type=button value='>>' onclick='location.href=\"%s?browseTextDocumentVersions&par_id=%d&version_number=%d\"'>\n",par_class,par_id,version_number+1);
      ap_rprintf(r," out of %d</FONT>",document_version);
      ap_rputs("</TD>\n</TR>\n</TABLE>\n",r);

      ap_rputs("</TD>\n</TR>\n</TABLE>\n",r);
      
      ap_rputs("</BODY>",r);

      %%
      ;
    };
  };

  method showTextDocumentVersion
 // sends version of text document in edit mode
  {
    doc "show particular version of text_document as HTML";
    id_permission __PERMISSION_GET;

    __PARAMS_HEADER

    param version_number { type CO_INTEGER; };

    declare language_c
    {
      C
      %%
      FILE *fh;
      char *filename;
      int document_version = 0;
      %%
      ;
    };

    body language_c
    {
      C
      %%
      __SQL_GET_INTEGER(document_version, "SELECT file_version FROM co_%s_%s WHERE id=%d", par_area, par_class, par_id)

      if (version_number>document_version) version_number = document_version;

      filename = ap_psprintf(p, "%s/%s", cou_getCODataPath(), loc_getPathForVersion( p, par_area, par_class, par_id, version_number));
      __PUT_HTML_FROM_FILE(filename)

      %%
      ;
    };
  };

  method editTextDocument
 // sends text document in HTML in edit mode
  {
    doc "generates client interface to edit document";
    id_permission __PERMISSION_UPDATE;

    __PARAMS_HEADER

    declare language_c
    {
      C
      %%
      FILE *fh;
      const char *filename;
      int document_version;
      const char *filedata;
      %%
      ;
    };

    body language_c
    {
      C
      %%
      __SQL_GET_INTEGER(document_version, "SELECT file_version FROM co_%s_%s WHERE id=%d", par_area, par_class, par_id)

      if (!ok) // object does not exist
      {
       cou_error(cnx, "Document not found");
       ap_rputs("<BODY>\n<H1>requested object does not exist</H1>\n</BODY>",r);
       return 0;
      }

      //ap_rprintf(r,"<HEAD>\n<SCRIPT>\nvar par_id=%d\nvar version_number=%d\n</SCRIPT>\n",par_id,document_version);
      filename = ap_psprintf(p, "%s/text_document/editTextDocument.up", cou_getCORootPath(cnx));
      __PUT_ASP_FROM_FILE(filename)

      @co_call(par_area,par_class,'getFileAsString',par_id=>par_id,par_area=>par_area,par_class=>par_class);
      filedata = (const char *) cou_getReturnValue(cnx, CO_STRING, &err);

      ap_rputs( cou_buildTEXTAREAEncodedTextDocument(cnx, filedata), r );

      filename = ap_psprintf(p, "%s/text_document/editTextDocument.down", cou_getCORootPath(cnx));
      __PUT_HTML_FROM_FILE(filename)

      %%
      ;
    };
  };

  method saveFormToTextDocument
 // obtains form and puts its contents into TextDocument
  {
    
    doc "saves HTML form to text_document file";
    id_permission __PERMISSION_UPDATE;

    __PARAMS_HEADER

    param text_document_data { type CO_STRING; };
    param text_document_action_param { type CO_STRING; };
    param active_path { type CO_STRING; };

    declare language_c
    {
      C
      %%
      char *filename;
      int fd; //file descriptor
      int document_version;
      %%
      ;
    };

    body language_c
    {
      C
      %%
// serious work is done by file class
	@co_call(par_area,par_class,'writeFileFromString',par_id=>par_id,par_area=>par_area,par_class=>par_class,text=>text_document_data);

	if (!strcmp(text_document_action_param,"saveAndExit")) ap_rprintf(r, "<BODY onload='window.close();window.history.go(-1);'></BODY>");
	if (!strcmp(text_document_action_param,"justSave")) 
	    ap_rprintf(r, 
		"<BODY onload='location.href=\"/co/html/%s/%s?editTextDocument&par_area=%s&par_class=%s&par_id=%d&active_path=%s\"'></BODY>",
		par_area, par_class, par_area, par_class, par_id, active_path
		);

      %%
      ;
    };
  };
};

//END OF CLASS text_document
